#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <check.h>

#include "test_dff.h"

static void setup (void);
static void teardown (void);
static Suite* onedff_suite (void);

static int device_handle = -1;

static void setup(void)
{
        //! Set up connection to FPGA device driver.
        //! Create device with major number 240 and minor 0.
        device_handle = open(SWREG_DEVICE, O_RDWR);

        write_fpga(device_handle, "00000001", "00000000");
        write_fpga_zeros(device_handle);
}


static void teardown (void)
{
        close(device_handle);
}


START_TEST (test_create)
{
        char rbuffer[4];

        fail_unless (device_handle != -1, "Could not connect to FPGA device driver! device_handle=%d", device_handle);
        //! Write 1 to flip-flop and 1 to global clock buffer CE.
        write_fpga(device_handle, "10000001", "00000000");
        //! Flush the input HW/SW registers and read the buffer.
        write_fpga_zeros(device_handle);
        read_fpga(device_handle, rbuffer);
        fail_unless(rbuffer[2] == (char)bin2dec("10000000"), "Could not set flip-flop value!");
}
END_TEST


//! Test that the register works as expected when global clock buffer's CE-port is enabled.
START_TEST (test_clock_enabled)
{
        char rbuffer[4];

        read_fpga(device_handle, rbuffer);
        fail_unless(rbuffer[2] == (char)bin2dec("00000000"), "Initial values not correct.");

        char reg0[8];
        int i;
        //! Reset input register.
        for (i=0;i<8;i++)
                reg0[i] = '0';
        reg0[8] = '\0';

        //! Always set bit 7 to '1' and set each bit to '1' in sequence. 
        //! Switch first bit for each time.
        reg0[7] = '1';
        int firstbit;

        for (firstbit = 0; firstbit < 2; firstbit++) {
                if (firstbit == 0)
                        reg0[0] = '0';
                else
                        reg0[0] = '1';
                for (i=1;i<7;i++) {
                        reg0[i] = '1';
                        write_fpga(device_handle, reg0, "00000000");
                        write_fpga_zeros(device_handle);
                        read_fpga(device_handle, rbuffer);
                        if (firstbit == 0)
                                fail_unless(rbuffer[2] == (char)bin2dec("00000000"), "Register is not set!");
                        else if (i == 1) { // firstbit == 1 and reset
                                //! Register is reset.
                                fail_unless(rbuffer[2] == (char)bin2dec("00000000"), "Register is not reset!");
                                //! Turn of reset and put register back to initial state.
                                reg0[i] = '0';
                                write_fpga(device_handle, "10000001", "00000000");
                                write_fpga_zeros(device_handle);
                        }
                        else // firstbit == 1 and no reset
                                fail_unless(rbuffer[2] == (char)bin2dec("10000000"), "Register is not set! i=%X, reg0=%s, rbuffer[2]=%X", i, reg0, rbuffer[2]);
                }
        }
}
END_TEST


//! Test that the register works as expected when global clock buffer's CE-port is disabled.
START_TEST (test_clock_disabled)
{
        char rbuffer[4];

        read_fpga(device_handle, rbuffer);
        fail_unless(rbuffer[2] == (char)bin2dec("00000000"), "Initial values not correct.");

        char reg0[8];
        int i;
        //! Reset input register.
        for (i=0;i<8;i++)
                reg0[i] = '0';
        //! This is a standard text string. Must null-terminate last char.
        reg0[8] = '\0';

        //! Always set bit 7 to '0' and set each bit to '1' in sequence. 
        //! Switch first bit for each time.
        reg0[7] = '0';
        int firstbit;

        for (firstbit = 0; firstbit < 2; firstbit++) {
                if (firstbit == 0)
                        reg0[0] = '0';
                else {
                        reg0[0] = '1';
                        write_fpga(device_handle, "10000000", "00000000");
                        write_fpga_zeros(device_handle);
                }
                for (i=1;i<7;i++) {
                        reg0[i] = '1';
                        write_fpga(device_handle, reg0, "00000000");
                        write_fpga_zeros(device_handle);
                        read_fpga(device_handle, rbuffer);
                        if (firstbit == 0)
                                fail_unless(rbuffer[2] == (char)bin2dec("00000000"), "Register is not set!");
                        else if (i == 1) { // firstbit == 1 and reset
                                //! Register is reset.
                                fail_unless(rbuffer[2] == (char)bin2dec("00000000"), "Register is not reset!");
                                //! Turn of reset and put register back to initial state.
                                reg0[i] = '0';
                                write_fpga(device_handle, "10000001", "00000000");
                                write_fpga_zeros(device_handle);
                        }
                        else // firstbit == 1 and reset
                                fail_unless(rbuffer[2] == (char)bin2dec("10000000"), "Register is not set! i=%X, reg0=%s, rbuffer[2]=%X", i, reg0, rbuffer[2]);
                }
        }
}
END_TEST


static Suite* onedff_suite (void)
{
        Suite *s = suite_create ("onedff");

        TCase* tc_core = tcase_create ("Core");
        tcase_add_checked_fixture (tc_core, setup, teardown);
        tcase_add_test (tc_core, test_create);
        tcase_add_test (tc_core, test_clock_enabled);
        tcase_add_test (tc_core, test_clock_disabled);
        //! Disable timeout because multiple tests (and icap_write) takes long time.
        tcase_set_timeout(tc_core, 0);
        suite_add_tcase (s, tc_core);

        return s;
}

int test_onedff (void)
{
        int number_failed;
        Suite *s = onedff_suite ();
        SRunner *sr = srunner_create (s);
        srunner_set_log (sr, TEST_LOG_ONEDFF);
        srunner_run_all (sr, CK_VERBOSE);

        number_failed = srunner_ntests_failed (sr);
        srunner_free (sr);
        //free(s);
        return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

