#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <check.h>

#include "test_dff.h"

static void setup (void);
static void teardown (void);
static Suite* threedffs_suite (void);

static int device_handle = -1;

static void setup(void)
{
        //! Set up connection to FPGA device driver.
        device_handle = open(SWREG_DEVICE, O_RDWR);

        write_fpga(device_handle, "00000001", "00000000");
        write_fpga_zeros(device_handle);
}


static void teardown (void)
{
        close(device_handle);
}


START_TEST (test_create)
{
        char rbuffer[4];

        fail_unless (device_handle != -1, "Could not connect to FPGA device driver! device_handle=%d", device_handle);
        //! Write 1 to flip-flop and 1 to global clock buffer CE.
        write_fpga(device_handle, "10000001", "00000000");
        //! Flush the input HW/SW registers and read the buffer.
        write_fpga_zeros(device_handle);
        read_fpga(device_handle, rbuffer);
        fail_unless(rbuffer[2] == (char)bin2dec("10000000"), "Could not set flip-flop value!");
}
END_TEST


//! Test that the register works as expected when global clock buffer's CE-port is enabled.
START_TEST (test_clock_enabled)
{
        char rbuffer[4];

        read_fpga(device_handle, rbuffer);
        fail_unless(rbuffer[2] == (char)bin2dec("00000000"), "Initial values not correct.");

        //! Register value to be written.
        char reg0[9] = "00000000";
        //! Register value initially written to register.
        char initstr[9] = "00000000";
        //! Expected register value after write.
        char goldenstr[9] = "00000000";
        reg0[8] = '\0';
        initstr[8] = '\0';
        goldenstr[8] = '\0';
        int initvalue;
        int i;
        for (initvalue=0;initvalue<8;initvalue++) {
                if (initvalue == 0)
                        strcpy(initstr, "00000000");
                else if (initvalue == 1)
                        strcpy(initstr, "00100000");
                else if (initvalue == 2)
                        strcpy(initstr, "01000000");
                else if (initvalue == 3)
                        strcpy(initstr, "01100000");
                else if (initvalue == 4)
                        strcpy(initstr, "10000000");
                else if (initvalue == 5)
                        strcpy(initstr, "10100000");
                else if (initvalue == 6)
                        strcpy(initstr, "11000000");
                else if (initvalue == 7)
                        strcpy(initstr, "11100000");
                
                initstr[7] = '1';
                //printf("Writing: %s\n", initstr);
                write_fpga(device_handle, initstr, "00000000");
                write_fpga_zeros(device_handle);
                initstr[7] = '0';
                //! Always set bit 7 to '1'/enable the clock signal.
                strcpy(reg0, "00000001");
                strcpy(goldenstr, "00000000");
                for (i=0;i<7;i++) {
                        //! Try to switch each bit that was set initially.
                        if (initstr[i] == '1')
                                reg0[i] = '0';
                        else
                                reg0[i] = '1';

                        //! Make sure goldenstr follows the change.
                        if (i < 3) {
                                strcpy(goldenstr, reg0);
                                goldenstr[7] = '0';
                        }

                        write_fpga(device_handle, reg0, "00000000");
                        write_fpga_zeros(device_handle);
                        read_fpga(device_handle, rbuffer);
                        if (i == 3) {
                                //! Register is reset.
                                fail_unless(rbuffer[2] == (char)bin2dec("00000000"), "Register is not reset!");
                                //! Turn of reset and put register back to initstr.
                                reg0[i] = '0';
                                initstr[7] = '1';
                                write_fpga(device_handle, initstr, "00000000");
                                write_fpga_zeros(device_handle);
                                initstr[7] = '0';
                        } else //! Output register must follow the same change as goldenstr.
                                fail_unless(rbuffer[2] == (char)bin2dec(goldenstr), "Register did not change!");
                }
        }
}
END_TEST


//! Test that the register works as expected when global clock buffer's CE-port is disabled.
START_TEST (test_clock_disabled)
{
        char rbuffer[4];

        read_fpga(device_handle, rbuffer);
        fail_unless(rbuffer[2] == (char)bin2dec("00000000"), "Initial values not correct.");

        //! Register value to be written.
        char reg0[9] = "00000000";
        //! Register value initially written to register.
        char initstr[9] = "00000000";
        //! Expected register value after write.
        char goldenstr[9] = "00000000";
        reg0[8] = '\0';
        initstr[8] = '\0';
        goldenstr[8] = '\0';
        int initvalue;
        int i;
        for (initvalue=0;initvalue<8;initvalue++) {
                if (initvalue == 0)
                        strcpy(initstr, "00000000");
                else if (initvalue == 1)
                        strcpy(initstr, "00100000");
                else if (initvalue == 2)
                        strcpy(initstr, "01000000");
                else if (initvalue == 3)
                        strcpy(initstr, "01100000");
                else if (initvalue == 4)
                        strcpy(initstr, "10000000");
                else if (initvalue == 5)
                        strcpy(initstr, "10100000");
                else if (initvalue == 6)
                        strcpy(initstr, "11000000");
                else if (initvalue == 7)
                        strcpy(initstr, "11100000");
                
                initstr[7] = '1';
                write_fpga(device_handle, initstr, "00000000");
                write_fpga_zeros(device_handle);
                initstr[7] = '0';
                //! Always set bit 7 to '0'/disable the clock signal.
                strcpy(reg0, "00000000");
                strcpy(goldenstr, initstr);
                for (i=0;i<7;i++) {
                        //! Try to switch each bit that was set initially.
                        if (initstr[i] == '1')
                                reg0[i] = '0';
                        else
                                reg0[i] = '1';

                        write_fpga(device_handle, reg0, "00000000");
                        write_fpga_zeros(device_handle);
                        read_fpga(device_handle, rbuffer);
                        if (i == 3) {
                                //! Register is reset.
                                fail_unless(rbuffer[2] == (char)bin2dec("00000000"), "Register is not reset!");
                                //! Turn of reset and put register back to initstr.
                                reg0[i] = '0';
                                initstr[7] = '1';
                                write_fpga(device_handle, initstr, "00000000");
                                write_fpga_zeros(device_handle);
                                initstr[7] = '0';
                        } else
                                fail_unless(rbuffer[2] == (char)bin2dec(goldenstr), "Register did change!");
                }
        }
}
END_TEST


static Suite* threedffs_suite (void)
{
        Suite *s = suite_create ("threedffs");

        TCase *tc_core = tcase_create ("Core");
        tcase_add_checked_fixture (tc_core, setup, teardown);
        tcase_add_test (tc_core, test_create);
        tcase_add_test (tc_core, test_clock_enabled);
        tcase_add_test (tc_core, test_clock_disabled);
        //! Disable timeout because multiple tests (and icap_write) takes long time.
        tcase_set_timeout(tc_core, 0);
        suite_add_tcase (s, tc_core);

        return s;
}

int test_threedffs (void)
{
        int number_failed;
        Suite *s = threedffs_suite ();
        SRunner *sr = srunner_create (s);
        srunner_set_log (sr, TEST_LOG_THREEDFFS);
        srunner_run_all (sr, CK_VERBOSE);

        number_failed = srunner_ntests_failed (sr);
        srunner_free (sr);
        //free(s);
        return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

