#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <hlog.h>
#include <hmem.h>

#include "hdaemon.h"
#include "hdsched.h"
#include "hdmsg.h"
#include "hdplacer.h"
#include "hdtimer.h"


//! Initialize this process as a daemon (Linux actually has it's own interface for this).
static int init_daemon_();
static void signal_handler_(int sig);


static void signal_handler_(int sig)
{
        switch (sig) {
                case SIGINT:
                case SIGTERM:
                        hdmsg_notify(HDME_QUIT);
                        hdmsg_synchronize();
                        break;
        }
}

static int init_daemon_()
{
        // Fork off the parent process.
        // Exit the parent process if success.
        pid_t pid;
        pid = fork();
        if (pid < 0) {
                return -1;
        } else if (pid > 0)
                exit(EXIT_SUCCESS);

        // Change the file mode mask (should not inherit umask from parent).
        umask(0);

        // Create a new SID for the child process.
        pid_t sid;
        sid = setsid();
        if (sid < 0) {
                return -1;
        }

        // Change the current working directory.
        if ((chdir("/")) < 0) {
                return -1;
        }

        // Close out the standard file descriptors.
        close(STDIN_FILENO);
        close(STDOUT_FILENO);
        close(STDERR_FILENO);

        hlog_write(HLOG_NORMAL, "Daemon initialized.\n");

        return 0;
}


int main( int argc, const char* argv[] )
{
        // Connect the different signals (that the daemon can receive) to signal_handler_ function.
        signal(SIGINT, signal_handler_);
        signal(SIGTERM, signal_handler_);

        if (hlog_init(DAEMON_LOG) < 0) {
                printf("Error. Could not init log file. Daemon is stopping.\n");
                exit(EXIT_FAILURE);
        }

        // Init the daemon.
    //    if (init_daemon_() < 0)
      //          exit(EXIT_FAILURE); 

        // Initialize a dynamic memory module with 256 lines.
        //dmemInit(256);

        static pthread_t scheduler_thread;
        static pthread_t placer_thread;
        static pthread_t message_server_thread;
        static pthread_t timer_thread;

        // Takes care of scheduling processes.
        pthread_create(&scheduler_thread, NULL, hdsched_main, (void*)NULL);
        // Takes care of placing processes on the hardware.
        pthread_create(&placer_thread, NULL, hdplacer_main, (void*)NULL);
        // Generates timer interrupts.
        pthread_create(&timer_thread, NULL, hdtimer_main, (void*)NULL);
        // Takes care of receiving and processing messages from client programs.
        // This thread is also the master thread and responsible for exiting all other threads.
        pthread_create(&message_server_thread, NULL, hdmsg_main, (void*)NULL);

        // Wait for each thread to finish before exiting the daemon.
        pthread_join(timer_thread, NULL);
        pthread_join(scheduler_thread, NULL);
        pthread_join(placer_thread, NULL);
        pthread_join(message_server_thread, NULL);

        hlog_close();

        exit(EXIT_SUCCESS);
}






