#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <pthread.h>

#include <hlog.h>
#include <hstructures.h>
#include <hplacer.h>
#include <hprocess.h>
#include <hmsg.h>
#include <hmqueue.h>
#include <hevent.h>

#include "hdplacer.h"
#include "hdmsg.h"
#include "hdsched.h"
#include "../platform_defines.h"


static void cleanup_(int exit_code);

/*! \brief Handle an incoming event.
 *
 *  This functions locks a mutex and waits on a
 *  condition variable.
 */
static int wait_event_();
static struct hprocess* get_current_process_();
static int set_current_process_(struct hprocess* process);

static struct hprocess* current_process = NULL;
//! The event handler for this thread.
static struct hevent* event_handler = NULL;
//! If this is true, this thread should ask the server to quit all threads when exiting.
static int notify_server_quit = 1;
static const char* bitfilespath = "../bitfiles";
static const char* icapdevice = "/var/icap";

static int wait_event_()
{
        enum hdpevent incoming_event = (enum hdmevent)hevent_wait(event_handler);

        struct hprocess* next_process = hdsched_get_next_process();
        struct hprocess* current_process = get_current_process_();
        int replace_current = 0;

        // An incoming event is handled here.
        switch (incoming_event) {
                case HDPE_QUIT:
                        // Do not ask server to quit (the server asked this thread to quit).
                        notify_server_quit = 0;
                        // Quit the thread.
                        cleanup_(EXIT_SUCCESS);
                        break;
                case HDPE_TIMER:
                        // Timer interrupt.
                        if (current_process != next_process)
                                replace_current = 1;
                        break;
        }

        // Replace process if current and next are unequal.
        if (replace_current) {
                printf("hdplacer_main: Replacing process. old-pid=%d. new-pid=%d\n", hprocess_get_pid(current_process), hprocess_get_pid(next_process));
                // Not ready yet.
                //hplacer_interrupt_process(current_process);
                //hplacer_load_process(next_process);
                set_current_process_(next_process);
                hdsched_reschedule(current_process, HDSE_RESCHED_TIMER);
        } else
                printf("hdplacer_main: Keeping current process on FPGA. pid=%d\n", hprocess_get_pid(next_process));

        return 0;
}


static struct hprocess* get_current_process_()
{
        return current_process;
}


static int set_current_process_(struct hprocess* process)
{
        current_process = process;

        return 0;
}


static void cleanup_(int exit_code)
{	
        // Only notify server if not done already.
        if (notify_server_quit) {
                notify_server_quit = 0;
                hdmsg_notify(HDME_QUIT);
        }

        hdmsg_synchronize();
        pthread_exit(NULL);
}


int hdplacer_notify(enum hdpevent event)
{
        hevent_notify(event_handler, (int)event);

        return 0;
}


void* hdplacer_main()
{
        // Create event handler.
        event_handler = hevent_create(HEVENT_BLOCKING);
        if (event_handler == NULL) {
                notify_server_quit = 0;
                cleanup_(EXIT_FAILURE);
        }

        hplacer_set_bitfilespath(bitfilespath);
        hplacer_set_icapdevice(icapdevice);

        // Sleep a bit to initialize scheduler.
        sleep(1);
        // Service loop for placer.
        while (1) {
                // Ask scheduler to find a new process to place.
                hdsched_notify(HDSE_SHORT);
                wait_event_();
        }

        hevent_remove(event_handler);
        cleanup_(EXIT_SUCCESS);
}






