#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pthread.h>

#include <hlog.h>
#include <hstructures.h>
#include <hprocess.h>
#include <hsqueue.h>
#include <hsqlist.h>
#include <hplacer.h>
#include <hlist.h>
#include <hevent.h>

#include "hdsched.h"
#include "hdmsg.h"


static void cleanup_(int exit_code);

//! Suspend the thread until a new event occurs.
static int wait_event_();

/*! \brief Performs the (long-term) scheduling of a new process.
 *
 *  This function should decide if the given process should be accepted
 *  by the system or not.
 *
 *  @param process Points to the first process in the HPS_NEW-queue.
 *
 *  @return Negative if process not accepted. 0 or positive value if it was.
 *  @retval HDSR_TOO_MANY Too many processes in system.
 *  @retval HDSR_NO_BITFILE Bitfile does not exists.
 *  @retval HDSR_READY Process added to a ready queue.
 */
static int schedule_longterm_(struct hprocess* process);
static int schedule_shortterm_();
static int increment_process_number_();
static int decrement_process_number_();

/*! \brief Get first queue with given priority or create a queue and insert it at correct place in list.
 *
 *  @param list The list of queues.
 *  @param priority The priority to be compared.
 *  @return The exisiting queue or a new queue (at correct place in list) on success. NULL on failure.
 */
static struct hsqueue* create_queue_by_priority_(int list, int priority);

/*! \brief Set next process to be placed on the FPGA.
 *
 *  The process can be fetched by another thread at the same
 *  time and is therefore secured with a mutex.
 *
 *  @param process The process.
 *  @return 0 on success. Negative on failure.
 */
int set_next_process_(struct hprocess* process);


//! Mutex for access to the queue of new processes.
static pthread_mutex_t new_hsqueue_mutex = PTHREAD_MUTEX_INITIALIZER;
//! Ask the server to quit all threads.
static int notify_server_quit = 1;
//! Number of (accepted) processes in the system and it's mutex.
static int processes_num = 0;
static pthread_mutex_t processes_num_mutex = PTHREAD_MUTEX_INITIALIZER;
//! Next process to be placed on the FPGA and it's mutex.
static struct hprocess* next_process = NULL;
static pthread_mutex_t next_process_mutex = PTHREAD_MUTEX_INITIALIZER;
//! Queue for processes that need rescheduling and it's mutex.
static struct hlist* resched_queue = NULL;
static pthread_mutex_t resched_queue_mutex = PTHREAD_MUTEX_INITIALIZER;
//! Mutex for HPS_READY-queue.
static pthread_mutex_t ready_queue_mutex = PTHREAD_MUTEX_INITIALIZER;
//! The event handler for this thread.
static struct hevent* event_handler = NULL;


static int increment_process_number_()
{
        int ret = -1;
        pthread_mutex_lock( &processes_num_mutex );
        processes_num += 1;
        ret = processes_num;
        pthread_mutex_unlock( &processes_num_mutex );

        return ret;
}


static int decrement_process_number_()
{
        int ret = -1;
        pthread_mutex_lock( &processes_num_mutex );
        if (processes_num < 1)
                processes_num = 0;
        else
                processes_num -= 1;

        ret = processes_num;
        pthread_mutex_lock( &processes_num_mutex );

        return ret;
}


static void cleanup_(int exit_code)
{	
        // Only notify server if not done already.
        if (notify_server_quit) {
                notify_server_quit = 0;
                hdmsg_notify(HDME_QUIT);
        }

        hdmsg_synchronize();
        pthread_exit(NULL);
}


static struct hsqueue* create_queue_by_priority_(int list, int priority)
{
        struct hsqueue* queue = hsqlist_get_last_queue(list);

        // Walk through the queues from the last (lowest priority) to
        // the first (highest priority).
        // Return the first queue that has larger or equal priority to
        // the given priority.
        while (hsqueue_get_prev(queue) != NULL) {
                if ( hprocess_get_priority ( hsqueue_get_first_process(queue) ) >= priority)
                        return queue;
                queue = hsqueue_get_prev(queue);
        }

       if (hprocess_get_priority( hsqueue_get_first_process(queue) ) < hprocess_base_priority()) {
                // All queues have less priority (or there is no queues).
                // Create new queue and put it in front of list.
                queue = hsqueue_create();
                hsqlist_insert_queue_before(HPS_READY, hsqlist_get_first_queue(HPS_READY), queue);
        } else if (hprocess_get_priority( hsqueue_get_first_process(queue) ) > hprocess_base_priority()) {
                // The queue found have larger priority than the given priority.
                // Create new queue and put it behind that queue.
                struct hsqueue* new_queue = hsqueue_create();
                if (hsqueue_get_next(queue) == NULL)
                        hsqlist_add_queue(HPS_READY, new_queue);
                else
                        hsqlist_insert_queue_before(HPS_READY, hsqueue_get_next(queue), new_queue);
                queue = new_queue;
        } // Else: The queue found has the given priority.

        return queue;
}


static int schedule_shortterm_()
{
        // Get the HPS_READY-queue with highest priority.
        pthread_mutex_lock(&ready_queue_mutex);
        struct hsqueue* ready_queue = hsqlist_get_first_queue(HPS_READY);
        struct hprocess* process = hsqueue_dequeue_process(ready_queue);
        pthread_mutex_unlock(&ready_queue_mutex);

        if (process != NULL) {
                set_next_process_(process);
        }

        return 0;
}


static int schedule_longterm_(struct hprocess* process)
{
        if ( hdsched_processes_number() + 1 > hprocess_max_processes() )
               return HDSR_TOO_MANY; 

        // Check if bitfile is regular file.
        struct stat statbuffer;
        char* filename_full = hplacer_create_full_bitfilename(process);
        hlog_write(HLOG_DEBUG, "schedule_longterm_: Filename for process: ");
        hlog_write_text(filename_full);
        hlog_write_text("\n");

        stat(filename_full, &statbuffer);
        free(filename_full);
        if ( !S_ISREG(statbuffer.st_mode) )
                return HDSR_NO_BITFILE;

        // Process can be added to the HPS_READY-queue with the base
        // priority.
        hprocess_set_priority(process, hprocess_base_priority());
        pthread_mutex_lock(&ready_queue_mutex);
        struct hsqueue* queue = create_queue_by_priority_( HPS_READY, hprocess_base_priority() );
        hsqueue_enqueue_process(queue, process);
        pthread_mutex_unlock(&ready_queue_mutex);
        increment_process_number_();

        return HDSR_READY;
}


static int wait_event_()
{
        enum hdsevent incoming_event = (enum hdsevent)hevent_wait(event_handler);

        // An incoming event is handled here.
        switch (incoming_event) {
                case HDSE_QUIT: {
                                        // Do not ask server to quit (the server asked this thread to quit).
                                        notify_server_quit = 0;
                                        cleanup_(EXIT_FAILURE);
                                        break;
                                }
                case HDSE_NEW_PROCESS: {
                                               struct hprocess* new_process = hdsched_get_new_process();
                                               int sched_long = schedule_longterm_(new_process);
                                               hlog_write(HLOG_DEBUG, "wait_event_: (scheduler) A new process has been added.\n");
                                               hlog_write(HLOG_DEBUG, "Return value from long-term scheduler: ");
                                               hlog_write_integer(sched_long);
                                               hlog_write_text(".\n");

                                               // Scheduler refuses the new process.
                                               // TODO: Should notify the message server and let it send a response to the client.
                                               if (sched_long < 0)
                                                       hprocess_remove(new_process);

                                               break;
                                       }
                case HDSE_SHORT:
                                       schedule_shortterm_();
                                       break;
                case HDSE_RESCHED_TIMER: {
                                                 struct hprocess* resched_process = hlist_dequeue(resched_queue);
                                                 if (resched_process != NULL) {
                                                         struct hsqueue* ready_queue = create_queue_by_priority_(HPS_READY, hprocess_get_priority(resched_process));
                                                         hsqueue_enqueue_process(ready_queue, resched_process);
                                                 }
                                                 break;
                                         }
                default:
                                         break;
        }

        return 0;
}


int set_next_process_(struct hprocess* process)
{
        // Lock the mutex to get a stable value of the process.
        pthread_mutex_lock( &next_process_mutex );
        next_process = process;
        pthread_mutex_unlock( &next_process_mutex );

        return 0;
}


int hdsched_reschedule(struct hprocess* process, enum hdsevent event)
{
        pthread_mutex_lock( &resched_queue_mutex );
        hlist_enqueue(resched_queue, process);
        pthread_mutex_unlock( &resched_queue_mutex );

        printf("Rescheduled process. pid=%d\n", hprocess_get_pid(process));

        hdsched_notify(event);

        return 0;
}


int hdsched_add_new_process(struct hprocess* process)
{
        //! TODO: This function always adds the new process to the first of the HPS_NEW-queues.
        //! A future version of the scheduler might want to add it to another queue based on
        //! the process' initial priority.
        pthread_mutex_lock( &new_hsqueue_mutex );
        hsqueue_enqueue_process(hsqlist_get_first_queue(HPS_NEW), process);
        pthread_mutex_unlock( &new_hsqueue_mutex );

        return 0;
}


struct hprocess* hdsched_get_new_process()
{
        struct hprocess* new_process = NULL;

        pthread_mutex_lock( &new_hsqueue_mutex );
        new_process = hsqueue_dequeue_process( hsqlist_get_first_queue(HPS_NEW) );
        pthread_mutex_unlock( &new_hsqueue_mutex );

        return new_process;
}


struct hprocess* hdsched_get_next_process()
{
        struct hprocess* ret = NULL;
        // Lock the mutex to get a stable value of the process.
        pthread_mutex_lock( &next_process_mutex );
        ret = next_process;
        pthread_mutex_unlock( &next_process_mutex );

        return ret;
}


int hdsched_processes_number()
{
        int ret = -1;
        pthread_mutex_lock( &processes_num_mutex );
        ret = processes_num;
        pthread_mutex_unlock( &processes_num_mutex );

        return ret;
}


int hdsched_notify(enum hdsevent event)
{
        hevent_notify(event_handler, (int)event);

        return 0;
}


void* hdsched_main()
{
        hsqlist_create(HPS_READY);
        struct hsqueue* new_hsqueue = hsqueue_create();
        hsqlist_create(HPS_NEW);
        hsqlist_add_queue(HPS_NEW, new_hsqueue);

        // Create event handler.
        event_handler = hevent_create(HEVENT_BLOCKING);
        if (event_handler == NULL) {
                notify_server_quit = 0;
                cleanup_(EXIT_FAILURE);
        }

        // Create queue for processes that need rescheduling.
        resched_queue = hlist_create();

        // Service loop for scheduler.
        while (1) {
                printf("hdsched_main: Waiting\n");
                // Suspend the scheduler until a new event.
                wait_event_(); 
        }


        hevent_remove(event_handler);
        cleanup_(EXIT_SUCCESS);
}






