/*! \file hwos.h
 *  \brief [Hardware OS Scheduler] Top level interface for scheduler.
 *
 *  This defines the interface at top-level for the scheduler. This is
 *  typically only functions to notify the scheduler on events etc.
 *
 *  This module pretty much works as a server.
 *
 *  Original author (2011): Sindre Hansen
 */

#include <hprocess.h>

#ifndef HDSCHED_H
#define HDSCHED_H


enum hdsevent {
       HDSE_NEW_PROCESS                   //! A new process has arrived to the HWOS.
       ,HDSE_QUIT                         //! Quit the placer thread.
       ,HDSE_SHORT                        //! Perform short-term scheduling.
       ,HDSE_RESCHED_TIMER                //! Reschedule process (was timer interrupted).
       ,HDSE_RESCHED_PRIORITY             //! Reschedule process (was replaced by higher priority process).
       ,HDSE_RESCHED_IO                   //! Reschedule process (needed unavailable IO).
};

enum hdsreturn {
       HDSR_NO_BITFILE = -2                 //! Bitfile does not exist.
       ,HDSR_TOO_MANY = -1                  //! Too many processes in system already.
       ,HDSR_READY = 0                      //! Process has been added to a ready queue.
};

/*! \brief Main function for scheduler thread.
 */
void* hdsched_main();

/*! \brief Reschedule the given process.
 *
 *  This function is called when the given an
 *  event has occured for the given process.
 *
 *  @param process The process to be rescheduled.
 *  @return 0 on success. Negative on failure.
 */
int hdsched_reschedule(struct hprocess* process, enum hdsevent event);

/*! \brief Notify the scheduler on new event.
 *
 *  @param event The notified event.
 */
int hdsched_notify(enum hdsevent event);

/*! \brief Return the number of accepted processes in the scheduler.
 *
 *  This is the number of all processes that has been accepted.
 *  It does not include processes in the HPS_NEW-queues.
 *
 *  This number can be modifed by the scheduler at the same
 *  time and is therefore secured with a mutex.
 *
 *  @return Number of processes.
 */
int hdsched_processes_number();

/*! \brief Add a new process to one of the HPS_NEW-queues.
 *
 *  The HPS_NEW-queues can be accessed by another thread at the same
 *  time and is therefore secured with a mutex.
 *
 *  @param process The process to be added.
 *  @return 0 on success. Negative on failure.
 */
int hdsched_add_new_process(struct hprocess* process);

/*! \brief Get process in front of the HPS_NEW-queue with highest priority.
 *
 *  The HPS_NEW-queues can be modifed by another thread at the same
 *  time and is therefore secured with a mutex.
 *
 *  @return The process.
 */
struct hprocess* hdsched_get_new_process();


/*! \brief Get next process to be placed on the FPGA.
 *
 *  The process can be set by the scheduler at the same
 *  time and is therefore secured with a mutex.
 *
 *  @return The process.
 */
struct hprocess* hdsched_get_next_process();

#endif

