/*! \file hdev.h
 *  \brief [Hardware OS Device Driver Interface] Keep track of device drivers in the HWOS.
 *
 *  This functionality is not really used in this version of the HWOS.
 *  The module can function as a base for further development. It is probably a good idea
 *  to perform the actual creation of a device driver in hdev_add (now it only adds a
 *  description to a list).
 *
 * Based on devicelist by Vegard Endresen.
 * Sindre Hansen (2011): Rewritten.
 *      - Changed names and public interface.
 *      - Hidden the device structure.
 *      - Encapsulated variables and data structures.
 */


#include "hstructures.h"

#ifndef HDEV_H
#define HDEV_H

//! Longest allowed path for device file.
#define HDEV_PATH_SIZE 40

/*! \brief Size of the device list.
 *
 *  @return Number of elements in the device list.
 */
int hdev_size();

/*! \brief Get device by major number.
 *
 *  @return Number of elements in the device list.
 */
struct hdev* hdev_get_by_major(int major);

/*! \brief Add a device description.
 *
 *  This adds a description of a character device
 *  with a path and a major number.
 *
 *  @param major The major number of the device.
 *  @param device_path Path of the device.
 *  @return A pointer to a struct representing the device on success. NULL on failure.
 */
struct hdev* hdev_add(int major, char* device_path);

/*! \brief Remove a device from the HWOS.
 *
 *  @param device Pointer to the struct representing the device.
 *  @return 0 on success. Negative on failure.
 */
int hdev_remove(struct hdev* device);

/*! \brief Get major number for device.
 *
 *  @param device Pointer to the device.
 *  @return Major number (positive integer) on success. Negative on failure.
 */
int hdev_get_major(struct hdev* device);

/*! \brief Remove device by major number.
 *
 *  @param major Major number.
 *  @return 0 on success. Negative on failure.
 */
int hdev_remove_by_major(int major);

/*! \brief Remove all devices in the system.
 *
 *  @return 0 on success. Negative on failure.
 */
int hdev_remove_all();

/*! \brief Get first device in list.
 *
 *  @return Pointer to the device on success. NULL on failure.
 */
struct hdev* hdev_get_first();

/*! \brief Get last device in list.
 *
 *  @return Pointer to the device on success. NULL on failure.
 */
struct hdev* hdev_get_last();

/*! \brief Get next device in list.
 *
 *  @param device Pointer to the device that has the next field.
 *  @return Pointer to the device on success. NULL on failure.
 */
struct hdev* hdev_get_next(struct hdev* device);

/*! \brief Get previous device in list.
 *
 *  @param device Pointer to the device that has the previous field.
 *  @return Pointer to the device on success. NULL on failure.
 */
struct hdev* hdev_get_prev(struct hdev* device);


#endif

