/*! \file hevent.h
 *  \brief [Hardware OS Event Interface] Notify and handle events between threads.
 *
 *  This module contains functionality for handling events between threads.
 *  Events can be handled both blocking (asynchronously) or non-blocking (synchronously).
 *  The underlying implemenation use System V message queues.
 *
 *  Author (2011): Sindre Hansen
 *
 */

#ifndef HEVENT_H
#define HEVENT_H

#define HEVENT_BLOCKING 1
#define HEVENT_NONBLOCKING 0

#include "hstructures.h"

/*! \brief Creates an event handler.
 *
 *  @param async Set to true if hevent_wait should block while waiting for events.
 *  @return The event handler object.
 */
struct hevent* hevent_create(int blocking);

/*! \brief Notify an event.
 *
 *  @param event Event handler.
 *  @param notification The actual event/notification. Must be an integer.
 *  @return 0 on success. Negative on failure.
 */
int hevent_notify(struct hevent* event, int notification);

/*! \brief Wait on event.
 *
 *  Wait may be performed in a blocking or
 *  non-blocking manner (see hevent_create).
 *
 *  @param event Event handler.
 *  @return The event on success. -1 on failure.
 */
int hevent_wait(struct hevent* event);

/*! \brief Remove an event handler.
 *
 *  @param event Event handler.
 *  @return 0 on success. -1 on failure.
 */
int hevent_remove(struct hevent* event);


#endif

