#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#include "hicap.h"
#include "hlog.h"


static int clbcol_to_abscol_(int col);
static int realrow_to_bitstreamrow_(int row);
static int write_header_(int device, uint32_t row, uint32_t column, uint32_t frame, int frames);
static int write_frames_(int device, uint32_t frames, char* filename,uint32_t frameOffset);
static int write_postconfig_(int device);

//! Array for converting CLB column to absolute column in the device currently open. 
static int* clbcol_to_abscol_current = NULL;
/*! Array for converting real row address to bitstream row address.
 *  - Each row has vertical width of 16 CLBs.
 *  - Real row address starts from the top and increments to the bottom.
 *  - Bitstream row address is NOT in a sequence from 0 to max number of rows.
 */
static int* realrow_to_bitstreamrow_current = NULL;
//! Number of rows (each row is 16 CLBs in vertical width) on the device currently open.
static int clbrows_current = -1;
//! Number of CLB columns on the device currently open.
static int clbcols_current = -1;


static int clbcol_to_abscol_(int col)
{
        if (clbcol_to_abscol_current == NULL || clbcols_current == -1)
                return -1;
        else if (col > -1 && col < clbcols_current)
                return clbcol_to_abscol_current[col];
        else
                return -1;
}


static int realrow_to_bitstreamrow_(int row)
{
        if (realrow_to_bitstreamrow_current == NULL || clbrows_current == -1)
                return -1;
        else if (row > -1 && row < clbrows_current)
                return realrow_to_bitstreamrow_current[row];
        else
                return -1;
}

/*
 *  Setup hwicap to write out frames.
 */
static int write_header_(int device, uint32_t row, uint32_t column, uint32_t frame, int frames)
{
        uint32_t data[19];
        int ret=0;
        uint32_t address;
        uint32_t packet;
        uint32_t headerWords;
        uint32_t totalWords;

        // Dummy, synch  
        data[0] = DUMMY_PACKET;
        data[1] = SYNC_PACKET;
        data[2] = NOP_PACKET;    //Flush
        data[3] = NOP_PACKET;    //Flush

        // Reset CRC
        data[4] = CMD_PACKET;
        data[5] = RCRC_CMD;
        data[6] = NOP_PACKET;    //Flush
        data[7] = NOP_PACKET;    //Flush

        // Id register
        data[8] = IDCODE_PACKET;
        data[9] = V4FX12ID;      //Device ID code;

        // Top/bottom bit, block type, row address, column, frame
        address = (0<<22) | (0<<19) | (row<<14) | (column<<6) | frame;

        // Setup FAR
        data[10] = FAR_PACKET;
        data[11] = address;       // address (top, CLB, row, column, minor)

        // Setup CMD write configuration.
        data[12] = CMD_PACKET;
        data[13] = WCFG_CMD;      // Type1 write  words to FDRI
        data[14] = NOP_PACKET;    //Flush

        totalWords = ((frames+1)*WORDS_PR_FRAME);     //Frames + dumy frame times
        //words pr frame.
        if(totalWords < XHI_TYPE_1_PACKET_MAX_WORDS){
                // Create Type 1 packet
                packet = (1<<TYPE1) | (1<<OP_WRITE) | (FDRI<<REG_ADDR) | totalWords;
                data[15] = packet;

                headerWords = 16;
        }else{
                // Create Type 2 packet
                packet = (1<<TYPE1) | (1<<OP_WRITE) | (FDRI<<REG_ADDR);
                data[15] = packet;

                packet = (1<<TYPE2) | (1<<OP_WRITE) | totalWords;
                data[16] = packet;

                headerWords = 17;
        }

        // Write out header.
        ret = write(device, data, sizeof(uint32_t)*headerWords);   // write out data
        // to device.
        return ret;
}


static int write_frames_(int device, uint32_t frames, char* filename, uint32_t frameOffset)
{
        int ret=0;
        void *frameBuffer;
        FILE *inFile;

        // allocate memmory for frames.
        frameBuffer = malloc(sizeof(uint32_t)*WORDS_PR_FRAME*(frames+1));
        if(frameBuffer == NULL){
                hlog_write(HLOG_ERROR, "write_frames_: (hicap) allocating frame buffer\n");
                return -1;   // ERROR
        }

        // open a file to read frames from. 
        inFile = fopen(filename, "r");    // Read frames from this file.
        if(inFile == NULL){
                hlog_write(HLOG_ERROR, "write_frames_: (hicap) opening file\n");
                free(frameBuffer);
                return -1;    // ERROR
        }

        // Go to correct place in file.
        if(fseek(inFile, frameOffset*WORDS_PR_FRAME*sizeof(uint32_t), SEEK_SET) < 0){
                hlog_write(HLOG_ERROR, "write_frames_: (hicap) setting filepointer offset\n");
                free(frameBuffer);
                fclose(inFile);
                return -1; // ERROR;
        }

        // Read frames from file and insert in buffer.
        if(fread(frameBuffer, WORDS_PR_FRAME*sizeof(uint32_t), (frames) , inFile) < 0){
                hlog_write(HLOG_ERROR, "write_frames_: (hicap) reading from file\n");
                ret = -1;
        }else{
                // Write frames + PAD frame.
                // As the memory is not initialized the PAD frame here can be anything.
                // Is this a BUG? should not mather as it is just for flushing the framebuffer.
                if(write(device, frameBuffer, sizeof(uint32_t)*WORDS_PR_FRAME*(frames+1)) < 0){
                        hlog_write(HLOG_ERROR, "write_frames_: (hicap) writing frames to device\n");
                        ret = -1;
                }
        }

        // Clean up.
        free(frameBuffer);
        fclose(inFile);

        return ret;  
}


static int write_postconfig_(int device)
{
        uint32_t data[5];
        int ret=0;

        // CRC
        data[0] = CMD_PACKET;
        data[1] = XHI_DISABLED_AUTO_CRC;

        // Desync
        data[2] = DESYNC_CMD;
        data[3] = NOP_PACKET;
        data[4] = NOP_PACKET;

        ret = write(device, data, 4*5);   // write;

        return ret;
}


int hicap_get_status(int icap_handle)
{
        uint32_t data[7];
        uint32_t status;

        // Dummy, synch  
        data[0] = DUMMY_PACKET;
        data[1] = SYNC_PACKET;
        data[2] = NOP_PACKET;    //Flush
        data[3] = NOP_PACKET;    //Flush

        // Read out status, one word.
        data[4] = (1<<TYPE1) | (1<<OP_READ) | (STAT<<REG_ADDR) | 1; 
        data[5] = NOP_PACKET;
        data[6] = NOP_PACKET;

        if(write(icap_handle, data, sizeof(uint32_t)*7) == 0){
                hlog_write(HLOG_ERROR, "hicap_get_status: writing get status\n");
                return 0; //error
        }

        // Get status.
        if(read(icap_handle, &status, sizeof(uint32_t)) == 0){
                hlog_write(HLOG_ERROR, "hicap_get_status: reading status\n");
                return 0; //error
        }

        // Desync
        data[0] = DESYNC_CMD;
        data[1] = NOP_PACKET;
        data[2] = NOP_PACKET;

        if(write(icap_handle, data, sizeof(uint32_t)*3) == 0){   // write;
                hlog_write(HLOG_ERROR, "hicap_get_status: desyncing get status\n");
                return 0; //error
        }

        return status;
}


int hicap_open(char* icap_device, char* device_type)
{
        clbcol_to_abscol_current = NULL;
        clbcols_current = -1;
        realrow_to_bitstreamrow_current = NULL;
        clbrows_current = -1;

        if (strcmp((const char*)device_type, "XC4VFX12") == 0) {
                clbcol_to_abscol_current = calloc(1, HC_CLBCOLS_XC4VFX12);
                clbcols_current = HC_CLBCOLS_XC4VFX12;
                clbrows_current = HC_CLBMINORROWS_XC4VFX12 / 16;
                realrow_to_bitstreamrow_current = calloc(1, clbrows_current);
                
                // Mapping found in Sverre Hamre's master thesis.
                int colmapping[HC_CLBCOLS_XC4VFX12] = {1,2,3,4,5,6,7,8,9,10,11,12,15,16,17,18,20,21,22,23,24,25,26,27};
                // Mapping found in CLBRead and in Vegard Endresen's master thesis.
                int rowmapping[HC_CLBMINORROWS_XC4VFX12 / 16] = {1, 0, 2, 3};
                int i;
                for (i = 0; i<HC_CLBCOLS_XC4VFX12; i++)
                        clbcol_to_abscol_current[i] = colmapping[i];
                for (i = 0; i<clbrows_current; i++)
                        realrow_to_bitstreamrow_current[i] = rowmapping[i];
        }

        return open(icap_device, O_RDWR);
}


int hicap_close(int icap_handle)
{
        clbcols_current = -1;
        clbrows_current = -1;
        if (clbcol_to_abscol_current != NULL)
                free(clbcol_to_abscol_current);

        if (realrow_to_bitstreamrow_current != NULL)
                free(realrow_to_bitstreamrow_current);

        if (close(icap_handle) != 0) {
                hlog_write(HLOG_ERROR, "hicap_close: closing driver\n");
                return -1;
        } else
                return 0;
}


int hicap_write(int icap_handle, char* infilename, int real_row, int clb_column, int frames_offset, int frames)
{
        if (write_header_(icap_handle, realrow_to_bitstreamrow_(real_row), clbcol_to_abscol_(clb_column), frames_offset, frames) < 0) {
                hlog_write(HLOG_ERROR, "hicap_write: Writing header\n");
                return -1;
        }

        if (write_frames_(icap_handle, frames, infilename, frames_offset) < 0) {
                hlog_write(HLOG_ERROR, "hicap_write: Writing frames\n");
                return -2;
        }

        if (write_postconfig_(icap_handle) < 0) {
                hlog_write(HLOG_ERROR, "hicap_write: Writing postconfig\n");
                return -3;
        }

        return 0;
}

