/*! \file hicap.h
 *  \brief [Internal Configuration Port Interface] Interface for partial reconfiguration on the FPGA.
 *
 *  This module makes it possible to write a partial bitstream to the ICAP on the FPGA. Based on
 *  the important work done by Sverre Hamre in his thesis.
 *
 *  Original author (2009): Sverre Hamre
 *  Modified (2011): Sindre Hansen
 *      - Customized interface for the HWOS.
 *      - Added conversion from CLB column to absolute column.
 *
 */

#ifndef HICAP_H
#define HICAP_H

//! Number of CLB columns on the XC4VFX12 (UG070, page 184).
#define HC_CLBCOLS_XC4VFX12           24
/*! Number of CLB minor rows on the XC4VFX12 (UG070, page 184).
 *  A minor row is defined here as a row that is 1 CLB in vertical width and
 *  spanning the entire device horizontally.
 */
#define HC_CLBMINORROWS_XC4VFX12      64


// TODO: These defines should be given the HC_-prefix (Sindre Hansen).

#define DUMMY_PACKET            0xFFFFFFFFUL
#define SYNC_PACKET             0xAA995566UL
#define NOP_PACKET              0X20000000UL
#define CMD_PACKET              0x30008001UL
#define FAR_PACKET              0x30002001UL
#define IDCODE_PACKET           0x30018001UL
#define STAT_PACKET             0x2800E001UL


#define WCFG_CMD                1    // Write configuration comand.
#define LFRM_CMD                3    // Write last frame command.
#define RCFG_CMD                4   
#define RCRC_CMD                7
#define DESYNC_CMD              14


/* Configuration registers */
#define FDRI                    2   //Frame data register input
#define STAT                    7   //Status register

/* Constant to use for CRC check when CRC has been disabled */
#define XHI_DISABLED_AUTO_CRC       0x0000DEFCUL
#define XHI_TYPE_1_PACKET_MAX_WORDS 1024

#define WORDS_PR_FRAME              41

#define TYPE2                       30  //Bitposition for type 2 header
#define TYPE1                       29  //Bitposition for type 1 header

#define REG_ADDR                    13  //Bitposition for register address in type 1 header

#define OP_READ                     27  //Bitposition for read op
#define OP_WRITE                    28  //Bitposition for write op

/* Device specific ID code */
#define V4FX12ID                    0x21E58093;  // Value read out from the virtex fx12 used.


/*! \brief Open the ICAP-device.
 *  
 *  @param icap_device Path to the device in the system. Example: /dev/icap
 *  @param device_type Type of the Virtex-4 device. Only "XC4VFX12" is supported at this time.
 *  @return A file descriptor for the device. -1 on failure.
 */
int hicap_open(char* icap_device, char* device_type);

/*! \brief Close the ICAP-device.
 *  
 *  @param icap_handle The handle/file descriptor to the device as returned by hicap_open.
 *  @return 0 on success. Negative on failure
 */
int hicap_close(int icap_handle);

/*! \brief Write a bitstream to ICAP.
 *  
 *  Each row spans the entire FPGA horizontally and is 16 CLBs in vertical width.
 *  The real row address start from 0 at the top of the device and increments downwards.
 *
 *  @param icap_handle The handle/file descriptor to the device as returned by hicap_open.
 *  @param infilename The filename for the bitfile.
 *  @param real_row Specify a row from 0 to [max rows - 1]. See definition of row in description.
 *  @param clb_column Specify a CLB column from 0 to [max CLB columns - 1] to start from.
 *  @param frames_offset Specify an offset of frames within the given address (normally 0).
 *  @param frames The number of frames to be written. 22 frames are 8 CLBs + 1 HCLK + 8 CLBs.
 *  @return 0 on success. Negative on failure
 */
int hicap_write(int icap_handle, char* infilename, int real_row, int clb_column, int frames_offset, int frames);

/*! \brief Get status of the ICAP-device.
 *  
 *  Refer to the work of Sverre Hamre to find out what this function should
 *  return.
 *
 *  @param icap_handle Handle for icap device. Returned by hicap_open.
 *  @return 0 on failure. Undefined on success?
 */
int hicap_get_status(int icap_handle);


#endif


