#include <stdio.h>
#include <stdlib.h>

#include "hlist.h"
#include "hsqlist.h"
#include "hlog.h"
#include "hprocess.h"
#include "hsignal.h"


//! One instance of this structure points to one list of elements.
struct hlist {
        int elements_num;                //! Number of elements in the list.
        struct hlelement* first;         //! Points to the first element in the list.
        struct hlelement* last;          //! Points to the last element in the list.
};

//! A double linked list structure. The data_element field can point to any data structure.
struct hlelement {
        struct hlelement* next;
        struct hlelement* prev;
        void* data_element;
};

//! A double linked list structure. The data_element field can point to any data structure.
//! This structure is not a part of a list.
struct hlorphan {
        struct hlorphan* next;
        struct hlorphan* prev;
        void* data_element;
};


static int set_first_(struct hlist* list, struct hlelement* element);
static int set_last_(struct hlist* list, struct hlelement* element);
static int increment_size_(struct hlist* list);
static int decrement_size_(struct hlist* list);
static int set_size_(struct hlist* list, int number);

static int element_set_prev_(struct hlelement* element, struct hlelement* element_prev);
static int element_set_next_(struct hlelement* element, struct hlelement* element_next);


static int element_set_prev_(struct hlelement* element, struct hlelement* element_prev)
{
        if (element == NULL)
                return -1;

        element->prev = element_prev;
        return 0;
}


static int element_set_next_(struct hlelement* element, struct hlelement* element_next)
{
        if (element == NULL)
                return -1;

        element->next = element_next;
        return 0;
}


static int set_first_(struct hlist* list, struct hlelement* element)
{
        if (list == NULL)
                return -1;

        list->first = element;
        return 0;
}


static int set_last_(struct hlist* list, struct hlelement* element)
{
        if (list == NULL)
                return -1;

        list->last = element;
        return 0;
}


static int increment_size_(struct hlist* list)
{
        if (list == NULL)
                return -1;
        
        list->elements_num += 1;
        
        return list->elements_num;
}


static int decrement_size_(struct hlist* list)
{
        if (list == NULL)
                return -1;

        if (list->elements_num < 1) {
                list->elements_num = 0;
                return 0;
        }
        
        list->elements_num = list->elements_num - 1;
        
        return list->elements_num;
}


static int set_size_(struct hlist* list, int number)
{
        if (list == NULL)
                return -1;

        list->elements_num = number;

        return list->elements_num;
}


int hlist_size(struct hlist* list)
{
        if (list == NULL)
                return -1;

        return list->elements_num;
}


void* hlelement_get_data(struct hlelement* element)
{
        if (element == NULL)
                return NULL;

        return element->data_element;
}


int hlelement_set_data(struct hlelement* element, void* data)
{
        if (element == NULL)
                return -1;

        element->data_element = data;

        return 0;
}


struct hlelement* hlelement_get_prev(struct hlelement* element)
{
        if (element == NULL)
                return NULL;

        return element->prev;
}


struct hlelement* hlelement_get_next(struct hlelement* element)
{
        if (element == NULL)
                return NULL;

        return element->next;
}


struct hlist* hlist_create()
{
        struct hlist* list = calloc(1, sizeof(*list));
        set_first_(list, NULL);
        set_last_(list, NULL);
        set_size_(list, 0);

        return list;
}


struct hlelement* hlist_enqueue(struct hlist* list, void* data_element)
{
        if (list == NULL)
                return NULL;

        struct hlelement* new_element = calloc(1, sizeof(*new_element));
        new_element->data_element = data_element;
        struct hlelement* last_element = hlist_get_last(list);

        // Update element structure.
        element_set_next_(last_element, new_element );
        element_set_prev_(new_element, last_element );
        element_set_next_(new_element, NULL );

        if (last_element == NULL) {
               set_first_(list, new_element);
        }

        set_last_(list, new_element);

        increment_size_(list);

        return new_element;
}


struct hlelement* hlist_insert_after(struct hlist* list, struct hlelement* element, void* data_element)
{
        if (list == NULL || element == NULL)
                return NULL;

        struct hlelement* new_element = calloc(1, sizeof(*new_element));
        new_element->data_element = data_element;

        // Update element structure.
        element_set_prev_(new_element, element);
        element_set_next_(new_element, hlelement_get_next(element));
        element_set_prev_(hlelement_get_next(element), new_element);
        element_set_next_(element, new_element);

        if (hlist_get_last(list) == element)
                set_last_(list, new_element);

        increment_size_(list);

        return new_element;
}


struct hlelement* hlist_insert_before(struct hlist* list, struct hlelement* element, void* data_element)
{
        if (list == NULL || element == NULL)
                return NULL;

        struct hlelement* new_element = calloc(1, sizeof(*new_element));
        new_element->data_element = data_element;

        // Update element structure.
        element_set_prev_(new_element, hlelement_get_prev(element) );
        element_set_next_(hlelement_get_prev(element), new_element);
        element_set_next_(new_element, element);
        element_set_prev_(element, new_element);

        if (hlist_get_first(list) == element)
                set_first_(list, new_element);

        increment_size_(list);

        return new_element;
}


void* hlelement_remove(struct hlelement* element, struct hlist* list)
{
        if (element == NULL)
                return NULL;
        
        void* data_element = hlelement_get_data(element);
        hlelement_set_data(element, NULL);

        if (element == hlist_get_first(list))
                set_first_(list, hlelement_get_next(element));

        if (element == hlist_get_last(list))
                set_last_(list, hlelement_get_prev(element));

        // Update element structure and remove element from memory.
        element_set_next_(hlelement_get_prev(element), hlelement_get_next(element) );
        element_set_prev_(hlelement_get_next(element), hlelement_get_prev(element) );
        free(element);

        decrement_size_(list);

        return data_element;
}


struct hlorphan* hlorphan_create(void* data_element)
{
        struct hlorphan* new_element = calloc(1, sizeof(*new_element));
        new_element->data_element = data_element;

        return new_element;
}


struct hlorphan* hlorphan_get_next(struct hlorphan* orphan)
{
        if (orphan == NULL)
                return NULL;

        return orphan->next;
}


int hlorphan_set_next(struct hlorphan* orphan, struct hlorphan* next)
{
        if (orphan == NULL)
                return -1;

        orphan->next = next;

        return 0;
}


struct hlorphan* hlorphan_get_prev(struct hlorphan* orphan)
{
        if (orphan == NULL)
                return NULL;

        return orphan->prev;
}


int hlorphan_set_prev(struct hlorphan* orphan, struct hlorphan* prev)
{
        if (orphan == NULL)
                return -1;

        orphan->prev = prev;

        return 0;
}


void* hlorphan_get_data(struct hlorphan* orphan)
{
        if (orphan == NULL)
                return NULL;

        return orphan->data_element;
}


void* hlist_dequeue(struct hlist* list)
{
        if (list == NULL)
                return NULL;

        // Delist is the same as removing the first element in the list.
        return hlelement_remove(hlist_get_first(list), list);
}


int hlist_remove(struct hlist* list)
{
        if (list == NULL)
                return -1;

        // Remove all the elements in the list and free the list from memory.
        struct hlelement* element = hlist_get_first(list);
        struct hlelement* next_element = NULL;
        while (element != NULL) {
                next_element = hlelement_get_next(element);
                hlelement_remove(element, list);
                element = next_element;
        }
        free(list);

        return 0;
}


struct hlelement* hlist_get_first(struct hlist* list)
{
        if (list == NULL)
                return NULL;

        return list->first; 
}


struct hlelement* hlist_get_last(struct hlist* list)
{
        if (list == NULL)
                return NULL;

        return list->last;
}

