/*! \file hlist.h
 *  \brief [Hardware OS Double Linked List Interface] General functionality for double linked lists.
 *
 *  General purpose double linked list structure. Each list (hlist) has elements (hlelement).
 *  Each element has data pointed to by void-pointers.
 *
 *  Author (2011): Sindre Hansen
 *
 */

#ifndef HLIST_H
#define HLIST_H

#include "hstructures.h"

/*! \brief Get size of elements in list.
 *
 *  @param list Pointer to list.
 *  @return Number of elements in list.
 */
int hlist_size(struct hlist* list);

/*! \brief Get data for the given element.
 *
 *  @param element Pointer to element.
 *  @return Pointer to the element's data.
 */
void* hlelement_get_data(struct hlelement* element);

/*! \brief Set data for the given element.
 *
 *  @param element Pointer to element.
 *  @param data Pointer to the data.
 *  @return 0 on success. Negative on failure.
 */
int hlelement_set_data(struct hlelement* element, void* data);

/*! \brief Get previous element in list.
 *
 *  @param element Pointer to element that has the previous field.
 *  @return Pointer to the previous element on success. NULL on failure.
 */
struct hlelement* hlelement_get_prev(struct hlelement* element);

/*! \brief Get next element in list.
 *
 *  @param element Pointer to element that has the next field.
 *  @return Pointer to the next element on success. NULL on failure.
 */
struct hlelement* hlelement_get_next(struct hlelement* element);

/*! \brief Create a new list.
 *
 *  @return Pointer to the new list structure on success. NULL on failure.
 */
struct hlist* hlist_create();

/*! \brief Enqueue a new element in a list.
 *
 *  @param list Pointer to the given list.
 *  @param data_element Pointer to the data that should be connected to the element.
 *  @return Pointer to the new element on success. NULL on failure.
 */
struct hlelement* hlist_enqueue(struct hlist* list, void* data_element);

/*! \brief Insert a new element after another in the given list.
 *
 *  @param list Pointer to the given list.
 *  @param element Element that the new element should be placed after.
 *  @param data_element Pointer to the data that should be connected to the element.
 *  @return Pointer to the new element on success. NULL on failure.
 */
struct hlelement* hlist_insert_after(struct hlist* list, struct hlelement* element, void* data_element);

/*! \brief Insert a new element before another in the given list.
 *
 *  @param list Pointer to the given list.
 *  @param element Element that the new element should be placed before.
 *  @param data_element Pointer to the data that should be connected to the element.
 *  @return Pointer to the new element on success. NULL on failure.
 */
struct hlelement* hlist_insert_before(struct hlist* list, struct hlelement* element, void* data_element);

/*! \brief Remove a given element from the given list.
 *
 *  @param list Pointer to the given list.
 *  @param element Pointer to the given element.
 *  @return Pointer to the elements data on success. NULL on failure.
 */
void* hlelement_remove(struct hlelement* element, struct hlist* list);

/*! \brief Create an list element that does not belong to any list.
 *
 *  @param data_element Pointer to the data that should be connected to the element.
 *  @return Pointer to the new element on success. NULL on failure.
 */
struct hlorphan* hlorphan_create(void* data_element);

/*! \brief Get next element for the given element.
 *
 *  @param orphan Pointer to the element that has the next field.
 *  @return Pointer to the element on success. NULL on failure.
 */
struct hlorphan* hlorphan_get_next(struct hlorphan* orphan);

/*! \brief Set next field for the given element.
 *
 *  @param orphan Pointer to the element that has the next field.
 *  @param next Pointer to the next element.
 *  @return 0 on success. Negative on failure.
 */
int hlorphan_set_next(struct hlorphan* orphan, struct hlorphan* next);

/*! \brief Get previous element for the given element.
 *
 *  @param orphan Pointer to the element that has the previous field.
 *  @return Pointer to the element on success. NULL on failure.
 */
struct hlorphan* hlorphan_get_prev(struct hlorphan* orphan);

/*! \brief Set previous field for the given element.
 *
 *  @param orphan Pointer to the element that has the previous field.
 *  @param prev Pointer to the previous element.
 *  @return 0 on success. Negative on failure.
 */
int hlorphan_set_prev(struct hlorphan* orphan, struct hlorphan* prev);

/*! \brief Get pointer to data for the given element.
 *
 *  @param orphan Pointer to the element that has the data field.
 *  @return Pointer to the data on success. NULL on failure.
 */
void* hlorphan_get_data(struct hlorphan* orphan);

/*! \brief Remove the first element (dequeue) from the list.
 *
 *  @param list Pointer to the given list.
 *  @return Pointer to the element's data on success. NULL on failure.
 */
void* hlist_dequeue(struct hlist* list);

/*! \brief Remove the given list and it's elements.
 *
 *  Note that this will NOT remove the data elements
 *  connected to each list element.
 *
 *  @param list Pointer to the given list.
 *  @return 0 on success. Negative on failure.
 */
int hlist_remove(struct hlist* list);

/*! \brief Get first element of the given list.
 *
 *  @param list Pointer to the given list.
 *  @return Pointer to the first element on success. NULL on failure.
 */
struct hlelement* hlist_get_first(struct hlist* list);

/*! \brief Get last element of the given list.
 *
 *  @param list Pointer to the given list.
 *  @return Pointer to the last element on success. NULL on failure.
 */
struct hlelement* hlist_get_last(struct hlist* list);

#endif

