/****************************************************************************
*File			: queuelist.c
*Author			: Vegard Endresen
*Description	: Hashmap<int,int> for (message queue key, message queue id)

Modified by Sindre Hansen as described in .h-file.
*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <sys/msg.h>

#include "hmqueue.h"
#include "hlog.h"


struct hmqueue
{	
	struct hmqueue *next;
	struct hmqueue *prev;
	int key; /*Message queue key, used under creation*/	
	int id;  /*Message queue id, used for sending*/
};


//! Number of message queues.
int queue_cnt = 0;
//! First and last element in list.
struct hmqueue* firstq = NULL;
struct hmqueue* lastq = NULL;

static int set_prev_(struct hmqueue* element, struct hmqueue* element_prev);
static int set_next_(struct hmqueue* element, struct hmqueue* element_next);
static int create_mqueue_(int key, int msgflg);
static int set_id_(struct hmqueue* element, int id);
static int set_key_(struct hmqueue* element, int key);


static int set_prev_(struct hmqueue* element, struct hmqueue* element_prev)
{
        if (element == NULL)
                return -1;

        element->prev = element_prev;
        return 0;
}


static int set_next_(struct hmqueue* element, struct hmqueue* element_next)
{
        if (element == NULL)
                return -1;

        element->next = element_next;
        return 0;
}


static int create_mqueue_(int key, int msgflg)
{
        int msqid;
        msqid = msgget((key_t)key, msgflg);

        if (msqid < 0) {
                hlog_write(HLOG_ERROR, "create_mqueue_: Can't create/connect to message queue.\n");
                return -1;
        } else {
                hlog_write(HLOG_DEBUG, "create_mqueue_: Message queue created. key=");
                hlog_write_integer(key);
                hlog_write_text(". msqid=");
                hlog_write_integer(msqid);
                hlog_write_text(". msgflg=");
                hlog_write_integer(msgflg);
                hlog_write_text(".\n");
        }

        return msqid;
}


static int set_id_(struct hmqueue* element, int id)
{
        if (element == NULL)
                return -1;

        element->id = id;

        return 0;
}


static int set_key_(struct hmqueue* element, int key)
{
        if (element == NULL)
                return -1;

        element->key = key;

        return 0;
}


int hmqueue_size()
{
        return queue_cnt;
}


struct hmqueue* hmqueue_get_prev(struct hmqueue* element)
{
        if (element == NULL)
                return NULL;

        return (struct hmqueue*)(element->prev);
}


struct hmqueue* hmqueue_get_next(struct hmqueue* element)
{
        if (element == NULL)
                return NULL;

        return element->next;
}


int hmqueue_get_key(struct hmqueue* element)
{
        if (element == NULL)
                return -1;

        return element->key;
}


int hmqueue_get_id(struct hmqueue* element)
{
        if (element == NULL)
                return -1;

        return element->id;
}


struct hmqueue* hmqueue_get(int key)
{
        if (hmqueue_size() == 0)
                return NULL;

        struct hmqueue* queue;
        queue = hmqueue_get_first();
        if (hmqueue_get_key(queue) == key)
                return queue;

        while (hmqueue_get_next(queue) != NULL) {
                queue = hmqueue_get_next(queue);
                if (hmqueue_get_key(queue) == key)
                        return queue;
        }

        return NULL;
}

int hmqueue_get_id_by_key (int key)
{
        if(hmqueue_size() == 0)
                return -1;

        struct hmqueue *it;
        it = hmqueue_get_first();	
        if (hmqueue_get_key(it) == key)
                return hmqueue_get_id(it);
        while(1) {
                //last item
                if(hmqueue_get_next(it) == NULL)
                        return -1; 

                it = hmqueue_get_next(it);
                if (hmqueue_get_key(it) == key)
                        return hmqueue_get_id(it);	
        }	
}


struct hmqueue* hmqueue_add(int key, enum hmqueue_mode mode)
{
        // Remove any existing queues (should be none or just one).
        int nr_removed = 0;
        while (hmqueue_remove_by_key(key) == 0) {
                nr_removed += 1;
                if (nr_removed > 10)
                        return NULL;
        }

        int msgflg = -1;
        if (mode == HMQ_CREATE)
                msgflg = IPC_CREAT | 0666;
        else if (mode == HMQ_CONNECT)
                msgflg = 0666;
        else
                return NULL;

        struct hmqueue* queue = calloc(1, sizeof(*queue));
        int msqid = create_mqueue_(key, msgflg);

        set_key_(queue, key);
        set_id_(queue, msqid);
        set_next_(queue, NULL);
        set_prev_(queue, NULL);

        if(hmqueue_size() == 0) {
                // This is first element.
                firstq = queue;
                lastq  = queue; 
        } else {
                // There is element(s) in list already.
                set_next_(hmqueue_get_last(), queue);
                set_prev_(queue, hmqueue_get_last());
                lastq = queue;
        }

        queue_cnt++;

        return queue;
}


int hmqueue_remove(struct hmqueue* element)
{
        if (element == NULL)
                return -1;

        hlog_write(HLOG_DEBUG, "hmqueue_remove: Removing queue with id=");
        hlog_write_integer(hmqueue_get_id(element));
        hlog_write_text("\n");
        // Remove the actual message queue.
        if (hmqueue_get_id(element) > 0)
                msgctl(hmqueue_get_id(element), IPC_RMID, NULL);

        if (hmqueue_size() < 1) {
                //! List is invalid.
                hlog_write(HLOG_ERROR, "hmqueue_remove: Invalid list structure (-2).\n");
                return -2;
        } else if (hmqueue_size() == 1) {
                // Only one element.
                free(element);
                firstq = NULL;
                lastq = NULL;
                queue_cnt = 0;
        } else {
                // More than one element.
                struct hmqueue* element_prev = hmqueue_get_prev(element);
                struct hmqueue* element_next = hmqueue_get_next(element);
                if (element_prev == NULL && element_next != NULL) {
                        // This is first element.
                        firstq = element_next;
                        set_prev_( element_next, NULL );
                } else if (element_next == NULL && element_prev != NULL) {
                        // This is last element.
                        lastq = element_prev;
                        set_next_( element_prev, NULL);
                } else if (element_next != NULL && element_prev != NULL) {
                        // Element is in-between.

                        //! Connect next and previous element together.
                        set_prev_( element_next, element_prev );
                        set_next_( element_prev, element_next );

                } else {
                        // List is invalid.
                        hlog_write(HLOG_ERROR, "hmqueue_remove: Invalid list structure (-3).\n");
                        return -3;
                }

                free(element);
                queue_cnt--;
        }

        return 0;
}


int hmqueue_remove_by_key(int key)
{
        struct hmqueue* element = hmqueue_get(key);
        return hmqueue_remove(element); 
}


int hmqueue_remove_all()
{
        struct hmqueue* it = hmqueue_get_first();
        struct hmqueue* it_prev = NULL;

        int ret = 0;
        while (hmqueue_get_next(it) != NULL) {
                it = hmqueue_get_next(it);
                it_prev = hmqueue_get_prev(it);
                if (hmqueue_remove(it_prev) < 0)
                        ret = -1;
        }

        if (hmqueue_remove (it) < 0)
                ret = -1;

        return ret;
}


struct hmqueue* hmqueue_get_first()
{
        return firstq; 
}

struct hmqueue* hmqueue_get_last()
{
        return lastq; 
}

int hmqueue_get_daemonkey()
{
        // Unique key for main message queue.
        return 1234;
}




