/*! \file hmqueue.h
 *  \brief [Hardware OS Message Queue Interface] Functionality for keeping message queue identifiers.
 * 
 *  This is simply a double linked list of System V message queues.
 *  This module should probably be rewritten so it uses hlist internally.
 *
 *  Original author (2010): Vegard Endresen
 *
 *  Modified (2011): Sindre Hansen
 *      - Changed naming of functions.
 *      - Used more encapsulating.
 *
 */

#ifndef HMQUEUE_H
#define HMQUEUE_H

//! Types of modes when creating a message queue.
enum hmqueue_mode {
        HMQ_CREATE,
        HMQ_CONNECT
};

/*! \brief Get previous element in list.
 *
 *  @param element Pointer to element that has the previous field.
 *  @return Pointer to the previous element on success. NULL on failure.
 */
struct hmqueue* hmqueue_get_prev(struct hmqueue* element);

/*! \brief Get next element in list.
 *
 *  @param element Pointer to element that has the next field.
 *  @return Pointer to the next element on success. NULL on failure.
 */
struct hmqueue* hmqueue_get_next(struct hmqueue* element);

/*! \brief Get key of a given message queue.
 *
 *  @param element Pointer to list element that has the message queue.
 *  @return The key (positive integer) on success. Negative on failure.
 */
int hmqueue_get_key(struct hmqueue* element);

/*! \brief Get ID of message queue.
 *
 *  @param element Pointer to list element that has the message queue.
 *  @return ID of the message queue on success. Negative on failure.
 */
int hmqueue_get_id(struct hmqueue* element);

/*! \brief Get size of elements in message queue list.
 *
 *  @return Number of elements in list.
 */
int hmqueue_size();

/*! \brief Get ID of message queue identified by key.
 *
 *  @param key A key that identifies the message queue.
 *  @return ID of the message queue on success. Negative on failure.
 */
int hmqueue_get_id_by_key(int key);

/*! \brief Get message queue element identified by key.
 *
 *  @param key A key that identifies the message queue.
 *  @return List element on success. NULL on failure.
 */
struct hmqueue* hmqueue_get(int key);

/*! \brief Adds a message queue and returns the queue identifier.
 *
 *  Based on "mode" this function either connects to an exisiting
 *  message queue or creates an new message queue.
 *
 *  @param key Unique key for the new queue.
 *  @param mode Mode for adding a message queue to the list.
 *  @return The message queue struct for the new queue on success. NULL on failure.
 */
struct hmqueue* hmqueue_add(int key, enum hmqueue_mode mode);

/*! \brief Remove the given list element.
 *
 *  @param element Pointer to list element that has the message queue.
 *  @return 0 on success. Negative on failure.
 */
int hmqueue_remove(struct hmqueue* element);

/*! \brief Remove the given message queue by key.
 *
 *  @param key Key that identifies the message queue.
 *  @return 0 on success. Negative on failure.
 */
int hmqueue_remove_by_key(int key);

/*! \brief Remove all message queues in list.
 *
 *  @return 0 on success. Negative on failure.
 */
int hmqueue_remove_all();

/*! \brief Get first element in list.
 *
 *  @return The first list element.
 */
struct hmqueue* hmqueue_get_first();

/*! \brief Get last element in list.
 *
 *  @return The last list element.
 */
struct hmqueue* hmqueue_get_last();

/*! \brief Get the key of the message servers message queue. 
 *
 *  @return The unique key of the message server.
 */
int hmqueue_get_daemonkey();

#endif

