#include <stdlib.h>
#include <string.h>
#include <sys/msg.h>

#include "hmsg.h"
#include "hprocess.h"
#include "hlog.h"
#include "hstructures.h"
#include "hmqueue.h"


//! Private base message (used for primitive polymorphism/type checking).
struct hmsg_base {
        long type;                     //! Must always be first in struct (because of Message Queue interface).
        struct {
                int sender_id;         //! ID of sender.
        } mtext;
};

/*! These are some of the message structures used.
 *  If other types of arguments are needed, it is
 *  probably a good idea to create additional
 *  message structures and not bruteforce the
 *  exisiting ones.
 *
 *  All the structures can be received by both
 *  server and client.
 */

//! Message for generic control operations. Often used for simple control
//! commands to the HWOS and for returning response to client.
struct hmsg_ctrl {
        long type;                                      //! Must be HMT_CTRL.
        struct {
                int sender_id;                          //! Identifies what queue to use for response.
                enum hmsg_command command;              //! Specifies what action to take.
                int retval;                             //! Return value.
        } mtext;
};

//! Message for memory control operations.
struct hmsg_ctrlmem {
        long type;                                      //! Must be HMT_CTRLMEM.
        struct {
                int sender_id;                          //! Identifies what queue to use for response.
                enum hmsg_command command;              //! Specifies what action to take. Examples are read from or write to memory.
                int retval;
                int address;                            //! Base address for memory operations.
                int size;                               //! Size of the given data. Needed because data may not be null-terminated.
                char data[HMSG_DATA_SIZE];              //! Generic data field. String data should be null-terminated ('\0').
        } mtext;
};

//! Message for registering a process.
struct hmsg_regproc {
        long type;                                                      //! Must be HMT_REGPROC
        struct {
                int sender_id;                                          //! Identifies what queue to use for response.
                enum hmsg_command command;                              //! Must be HMC_REGPROC.
                int nice;                                               //! Nice value for the new process.
                char bitfilename[HPROCESS_MAX_FILENAME_SIZE];           //! Filename of bitfile.
        } mtext;
};

//! Specifies the maximal size of the mtext field.
const int mtext_max_size = sizeof(((struct hmsg_ctrlmem*)NULL)->mtext);


static int set_type_(void* msg, int type);
static const int get_mtext_max_size_();
static int get_struct_size_(int type);
static int get_mtext_size_(int type);


static int get_struct_size_(int type)
{
        switch (type) {
                case HMT_CTRL:
                        return sizeof(struct hmsg_ctrl);
                case HMT_CTRLMEM:
                        return sizeof(struct hmsg_ctrlmem);
                case HMT_REGPROC:
                        return sizeof(struct hmsg_regproc);
                default:
                        return -1;
        }
}


static int get_mtext_size_(int type)
{
        switch (type) {
                case HMT_CTRL:
                        return sizeof ((struct hmsg_ctrl*)NULL)->mtext;
                case HMT_CTRLMEM:
                        return sizeof ((struct hmsg_ctrlmem*)NULL)->mtext;
                case HMT_REGPROC:
                        return sizeof ((struct hmsg_regproc*)NULL)->mtext;
                default:
                        return -1;
        }
}


static int set_type_(void* msg, int type)
{
        if (msg == NULL)
                return -1;
        else if (type == HMT_CTRL || type == HMT_CTRLMEM || type == HMT_REGPROC) {
                ((struct hmsg_base*)msg)->type = type;
                return 0;
        }

        return -2;
}


static const int get_mtext_max_size_()
{
        return mtext_max_size;
}


char* hmsg_get_bitfilename(void* msg)
{
        if (hmsg_get_type(msg) == HMT_REGPROC)
                return (char*)(((struct hmsg_regproc*)msg)->mtext.bitfilename);
        else {
                hlog_write(HLOG_ERROR, "hmsg_get_bitfilename: Failed to return data. Unknown message type.\n");
                return NULL;
        }   
}


int hmsg_set_bitfilename(void* msg, char* bitfilename)
{
        if (hmsg_get_type(msg) == HMT_REGPROC) {
                strncpy(hmsg_get_bitfilename(msg), bitfilename, HPROCESS_MAX_FILENAME_SIZE);
                return 0;
        } else {
                hlog_write(HLOG_ERROR, "hmsg_set_bitfilename: Failed to set data. Unknown message type.\n");
                return -1;
        }
}


int hmsg_get_nice(void* msg)
{
        if (hmsg_get_type(msg) == HMT_REGPROC)
                return ((struct hmsg_regproc*)msg)->mtext.nice;
        else
                return -1;
}


int hmsg_set_nice(void* msg, int nice)
{
        if (hmsg_get_type(msg) == HMT_REGPROC) {
                ((struct hmsg_regproc*)msg)->mtext.nice = nice;
                return 0;
        } else
                return -1;
}


int hmsg_set_sender(void* msg, int id)
{
        if (msg == NULL)
                return -1;

        ((struct hmsg_base*)msg)->mtext.sender_id = id;

        return 0;
}


int hmsg_get_sender(void* msg)
{
        if (msg == NULL)
                return -1;

        return ((struct hmsg_base*)msg)->mtext.sender_id;
}


int hmsg_get_type(void* msg)
{
        if (msg == NULL)
                return -1;

        return ((struct hmsg_base*)msg)->type;
}


int hmsg_set_return(void* msg, enum hmsg_return retval)
{
        if (hmsg_get_type(msg) == HMT_CTRL || hmsg_get_type(msg) == HMT_CTRLMEM) {
                ((struct hmsg_ctrl*)msg)->mtext.retval = retval;
                return 0;
        } else
                return -1;
}


enum hmsg_return hmsg_get_return(void* msg)
{
        if (hmsg_get_type(msg) == HMT_CTRL || hmsg_get_type(msg) == HMT_CTRLMEM)
                return ((struct hmsg_ctrl*)msg)->mtext.retval;
        else
                return -1;
}


int hmsg_get_size(void* msg)
{
        if (hmsg_get_type(msg) != HMT_CTRLMEM)
                return -1;

        return ((struct hmsg_ctrlmem*)msg)->mtext.size;
}

int hmsg_set_size(void* msg, int size)
{
        if (hmsg_get_type(msg) != HMT_CTRLMEM)
                return -1;

        ((struct hmsg_ctrlmem*)msg)->mtext.size = size;

        return 0;
}

char* hmsg_get_data(void* msg)
{
        switch (hmsg_get_type(msg)) {
                case HMT_CTRLMEM:
                        return (char*)(((struct hmsg_ctrlmem*)msg)->mtext.data);
                default:
                        hlog_write(HLOG_ERROR, "hmsg_get_data: Failed to return data. Unknown message (type=");
                        hlog_write_integer(hmsg_get_type(msg));
                        hlog_write_text(").\n");
                        return NULL;
        }
}


int hmsg_set_data(void* msg, char* data)
{
        if (hmsg_get_type(msg) == HMT_CTRLMEM) {
                strncpy(hmsg_get_data(msg), data, HMSG_DATA_SIZE);
                return 0;
        } else
                return -1;
}

int hmsg_set_address(void* msg, int address)
{
        if (hmsg_get_type(msg) == HMT_CTRLMEM)
                ((struct hmsg_ctrlmem*)msg)->mtext.address = address;
        else {
                hlog_write(HLOG_ERROR, "hmsg_set_address: Failed to set address. Unknown message type.\n");
                return -1;
        }

        return 0;
}


int hmsg_get_address(void* msg)
{
        if (hmsg_get_type(msg) == HMT_CTRLMEM)
                return ((struct hmsg_ctrlmem*)msg)->mtext.address;
        else {
                hlog_write(HLOG_ERROR, "hmsg_set_address: Failed to set address. Unknown message type.\n");
                return -1;
        }
}

int hmsg_send(void* msg, struct hmqueue* msq)
{
        int msqid = hmqueue_get_id(msq);
        hlog_write(HLOG_DEBUG, "hmsg_send: Adding message to message queue. msqid=");
        hlog_write_integer(msqid);
        hlog_write_text(".\n");

        switch( hmsg_get_type(msg) ) {
                case HMT_CTRL:
                        return msgsnd(msqid, (struct hmsg_ctrl*)msg, get_mtext_size_(HMT_CTRL), 0);
                case HMT_CTRLMEM:
                        return msgsnd(msqid, (struct hmsg_ctrlmem*)msg, get_mtext_size_(HMT_CTRLMEM), 0);
                case HMT_REGPROC:
                        return msgsnd(msqid, (struct hmsg_regproc*)msg, get_mtext_size_(HMT_REGPROC), 0);
                default:
                        hlog_write(HLOG_ERROR, "hmsg_send: Failed to send message. Unknown type.\n");
                        return -2;
        }
}


enum hmsg_command hmsg_get_command(void* msg)
{
        switch( hmsg_get_type(msg) ) {
                case HMT_CTRL:
                        return ((struct hmsg_ctrl*)msg)->mtext.command;
                case HMT_CTRLMEM:
                        return ((struct hmsg_ctrlmem*)msg)->mtext.command;
                case HMT_REGPROC:
                        return ((struct hmsg_regproc*)msg)->mtext.command;
                default:
                        return -1;
        }
}


int hmsg_set_command(void* msg, enum hmsg_command command)
{
        switch( hmsg_get_type(msg) ) {
                case HMT_CTRL:
                        ((struct hmsg_ctrl*)msg)->mtext.command = command;
                        break;
                case HMT_CTRLMEM:
                        ((struct hmsg_ctrlmem*)msg)->mtext.command = command;
                        break;
                case HMT_REGPROC:
                        ((struct hmsg_regproc*)msg)->mtext.command = command;
                        break;
                default:
                        return -1;
        }

        return 0;
}


void* hmsg_create(int type)
{
        // Allocate memory based on message type.
        // Point msg object to the newly allocated memory afterwards.
        if (type == HMT_CTRL) {
                struct hmsg_ctrl* ctrl_msg;
                ctrl_msg = (struct hmsg_ctrl*)calloc(1, sizeof(struct hmsg_ctrl));
                if (set_type_(ctrl_msg, HMT_CTRL) < 0) {
                        free(ctrl_msg);
                        return NULL;
                }
                hmsg_set_command(ctrl_msg, -1);
                hmsg_set_size(ctrl_msg, -1);
                hmsg_set_return(ctrl_msg, -1);
                return ctrl_msg;
        } else if (type == HMT_CTRLMEM) {
                struct hmsg_ctrlmem* ctrlmem_msg;
                ctrlmem_msg = (struct hmsg_ctrlmem*)calloc(1, sizeof(struct hmsg_ctrlmem));
                if (set_type_(ctrlmem_msg, HMT_CTRLMEM) < 0) {
                        free(ctrlmem_msg);
                        return NULL;
                }
                hmsg_set_command(ctrlmem_msg, -1);
                hmsg_set_address(ctrlmem_msg, -1);
                hmsg_set_return(ctrlmem_msg, -1);
                hmsg_set_size(ctrlmem_msg, -1);
                hmsg_set_data(ctrlmem_msg, "");
                return ctrlmem_msg;
        } else if (type == HMT_REGPROC) {
                struct hmsg_regproc* regproc_msg;
                regproc_msg = (struct hmsg_regproc*)calloc(1, sizeof(struct hmsg_regproc));
                if (set_type_(regproc_msg, HMT_REGPROC) < 0) {
                        free(regproc_msg);
                        return NULL;
                }
                hmsg_set_nice(regproc_msg, -1);
                // Set command by default.
                hmsg_set_command(regproc_msg, HMC_REGPROC);
                hmsg_set_bitfilename(regproc_msg, "");
                return regproc_msg;
        } else {
                hlog_write(HLOG_ERROR, "hmsg_create: Failed to create message. Unknown type.\n");
                return NULL;
        }
}


int hmsg_remove(void* msg)
{
        if (msg == NULL)
                return -1;
        
        free(msg);

        return 0;
}


void* hmsg_receive(struct hmqueue* msq)
{
        void* msg = NULL;
        int msqid = hmqueue_get_id(msq);

        // Allocate a memory block large enough to store any kind of message.
        msg = calloc(1, get_mtext_max_size_() + sizeof(long));
        msgrcv(msqid, msg, get_mtext_max_size_(), 0, 0);
        // Reallocate a smaller memory block.
        msg = realloc(msg, get_struct_size_( hmsg_get_type(msg) ));

        return msg;
}


