/*! \file hmsg.h
 *  \brief [Hardware OS Message Interface] Functionality for messages and message queues.
 *
 *  Defines an interface for message passing between the HWOS and a client application.
 *  The message system is implemented using System V Message Queues.
 *
 *  Based on (2010): Vegard Endresen
 *
 *  Author (2011): Sindre Hansen
 *      - Some commands in enum hmsg_command taken from work done by Vegard Endresen (2010).
 */

#include "hstructures.h"

#ifndef HMSG_H
#define HMSG_H

//! Size of data for messages sent to HWOS.
#define HMSG_DATA_SIZE 128

//! Definition of message types.
enum hmsg_type {
        HMT_ALL
        ,HMT_CTRL
        ,HMT_CTRLMEM
        ,HMT_REGPROC
};


//! Definition of commands that can be sent through messages from a client application.
enum hmsg_command {
        HMC_ALLOC               //! Allocate BRAM memory.
        ,HMC_EXEC               //! Write program to instruction segment of BRAM.
        ,HMC_FREE               //! Free BRAM memory.
        ,HMC_LDDEV              //! Load device driver.
        ,HMC_READ               //! Read data from backend.
        ,HMC_RMDEV              //! Remove device driver.
        ,HMC_RMQUE              //! Register receive message queue to HWOS.
        ,HMC_UMQUE              //! Unregister receive message queue to HWOS.
        ,HMC_WRITE              //! Write data to backend.
        ,HMC_SET_BITFILE        //! Set filename for bitstream file (deprecated?)
        ,HMC_REGPROC            //! Register a process.
};


//! Definition of return values that can be sent back to a client application.
enum hmsg_return {
        HMR_OK=0                   //! No errors.
        ,HMR_ERROR=-1              //! Errors when processing command.
        ,HMR_NOPID=-2              //! No exists process with the given PID.
};


/*! \brief Filename for FPGA bitstream. 
 *
 *  @param msg Message of type HMT_REGPROC.
 *  @return The filename.
 */
char* hmsg_get_bitfilename(void* msg);

/*! \brief Set filename for FPGA bitstream. 
 *
 *  @param msg Message of type HMT_REGPROC.
 *  @param bitfilename The filename.
 *  @return 0 on success. Negative on failure.
 */
int hmsg_set_bitfilename(void* msg, char* bitfilename);

/*! \brief Get nice value (user defined priority). 
 *
 *  @param msg Message of type HMT_REGPROC.
 *  @return Positive nice value on success. Negative on failure.
 */
int hmsg_get_nice(void* msg);

/*! \brief Set nice value (user defined priority). 
 *
 *  @param msg Message of type HMT_REGPROC.
 *  @param nice The positive value.
 *  @return 0 on success. Negative on failure.
 */
int hmsg_set_nice(void* msg, int nice);

/*! \brief Get sender ID for the message.
 *
 *  In this version of the HWOS, this is
 *  always the message queue ID of the queue
 *  owned by the sender.
 *
 *  @param msg Message of any type.
 *  @return Positive message queue ID on success. Negative on failure.
 */
int hmsg_get_sender(void* msg);

/*! \brief Set sender ID for the message.
 *
 *  In this version of the HWOS, this is
 *  always the message queue ID of the queue
 *  owned by the sender.
 *
 *  @param msg Message of any type.
 *  @param id The message queue ID.
 *  @return 0 on success. Negative on failure.
 */
int hmsg_set_sender(void* msg, int id);

/*! \brief Get type of the message.
 *
 *  @param msg Message of any type.
 *  @return The type (positive integer) on success. Negative on failure.
 */
int hmsg_get_type(void* msg);

/*! \brief Get size of the data field.
 *
 *  @param msg Message of type HMT_CTRLMEM.
 *  @return The size (positive integer) on success. Negative on failure.
 */
int hmsg_get_size(void* msg);

/*! \brief Set size of the data field.
 *
 *  @param msg Message of type HMT_CTRLMEM.
 *  @param size The size of the data field.
 *  @return 0 on success. Negative on failure.
 */
int hmsg_set_size(void* msg, int size);

/*! \brief Set return value for the specified message. 
 *
 *  @param msg Pointer to a response message.
 *  @param retval Return value.
 *  @return Positive on success. Negative on failure.
 */
int hmsg_set_return(void* msg, enum hmsg_return retval);

/*! \brief Get return value for the specified message. 
 *
 *  @param msg Message of type HMT_CTRL or HMT_CTRLMEM.
 *  @return The return value (constrained by enum hmsg_return).
 */
enum hmsg_return hmsg_get_return(void* msg);

/*! \brief Get data (string of characters) for message.
 *
 *  @param msg Message of type HMT_CTRLMEM.
 *  @return Pointer to data on success. NULL on failure.
 */
char* hmsg_get_data(void* msg);

/*! \brief Set data (string of characters) for message.
 *
 *  @param msg Message of type HMT_CTRLMEM.
 *  @param data Pointer to the data. Size not larger than HMSG_DATA_SIZE.
 *  @return 0 on success. Negative on failure.
 */
int hmsg_set_data(void* msg, char* data);

/*! \brief Set address field for a message.
 *
 *      @param msg Message of type HMT_CTRLMEM.
 *      @return Positive on success. Negative on failure.
 */
int hmsg_set_address(void* msg, int address);

/*! \brief Get address field for a message.
 *
 *      @param msg Message of type HMT_CTRLMEM.
 *      @return Positive address on success. Negative on failure.
 */
int hmsg_get_address(void* msg);

/*! \brief Get command for message.
 *
 *  @param msg Message of any type.
 *  @return The command entry (positive integer) for the message. Negative on failure.
 */
enum hmsg_command hmsg_get_command(void* msg);

/*! \brief Set command for message.
 *
 *  @param msg Message of any type.
 *  @param The command entry (positive integer) for the message.
 *  @return 0 on success. Negative on failure.
 */
int hmsg_set_command(void* msg, enum hmsg_command command);

/*! \brief Create and allocate memory for a message.
 *
 *  @param type The type of the new message.
 *  @return Pointer to the memory of the new message on success. NULL on failure.
 */
void* hmsg_create(int type);

/*! \brief Destroy message.
 *
 *  Actually just runs free(msg) if msg != NULL.
 *
 *  @param type Message of any type.
 *  @return 0 on success. NULL on failure.
 */
int hmsg_remove(void* msg);

/*! \brief Send a message.
 *
 *  Puts a message in a message queue.
 *
 *  @param msg Message of any type.
 *  @param msq Message queue. Message is put in this queue.
 *  @return 0 on success. Negative value on failure.
 */
int hmsg_send(void* msg, struct hmqueue* msq);

/*! \brief Blocking wait for an incoming message.
 *
 *  @param msq The message queue where the message arrives.
 *  @return Pointer to the received message on success. NULL on failure.
 */
void* hmsg_receive(struct hmqueue* msq);


#endif

