/*! \file hplacer.h
 *  \brief [Hardware OS Placer Interface] Responsible for interrupting and replacing a running process on the FPGA.
 *
 *  Author (2011): Sindre Hansen
 *
 */

#ifndef HPLACER_H
#define HPLACER_H


#include "hstructures.h"


/*! \brief Set the full filename of the ICAP device.
 *
 *  Example: "/dev/icap"
 *
 *  @param device The full filename to the device.
 */
int hplacer_set_icapdevice(const char* device);

/*! \brief Set the path of the bitfiles.
 *
 *  Sets the path to the files where the process
 *  bitfiles are located.
 *
 *  @param path The path without the '/' at the end.
 */
int hplacer_set_bitfilespath(const char* path);

/*! \brief Create full bitfilename.
 *
 *  This will allocate a string containing the path of the bitfiles and
 *  the file name of the given process' bitfilename.
 *
 *  @param process Take bitfilename from this process.
 */
char* hplacer_create_full_bitfilename(struct hprocess* process);

/*! \brief Interrupt a process running on the FPGA.
 *
 *  This assumes that the process is already placed on the FPGA and is running.
 *
 */
int hplacer_interrupt_process(struct hprocess* process);

/*! \brief Start a process running on the FPGA.
 *
 *  This assumes that the process is already placed on the FPGA, but not running.
 *
 */
int hplacer_start_process(struct hprocess* process);

/*! \brief Communicate with the running FPGA-process and make it send the
 *         state information back to the HWOS.
 *
 *  This assumes that the process is already placed on the FPGA and is stopped.
 *  This version of the HWOS will always store the state information in a file.
 *
 */
int hplacer_save_state(struct hprocess* process);

/*! \brief Load state information from memory/disk into the running process
 *         on the FPGA.
 *
 *  This version of the HWOS will always fetch the state information from
 *  a file on disk.
 *
 */
int hplacer_load_state(struct hprocess* process);

/*! \brief Load process into FPGA area.
 *
 *  This will place the given process on the FPGA.
 *
 */
int hplacer_load_process(struct hprocess* process);


#endif

