#include <stdlib.h>
#include <stdio.h>
#include <sys/msg.h>
#include <string.h>

#include "hprocess.h"
#include "hlog.h"
#include "hsignal.h"
#include "hsqueue.h"
#include "hsqlist.h"


struct hprocess {
        //! The process id.
        int pid; 
        //! The state of the process.
        int state;
        //! Value to manipulate priority (set by user application).
        int nice;
        //! Actual priority for process.
        int priority;
        //! File where the FPGA-bitstream for the process is.
        char* bitfilename; 
        //! File where the state data for the process is.
        char* statefilename; 
        struct hprocess* next;
        struct hprocess* prev;
        struct hsqueue* parent_queue;
};


//! Counter for pid values.
//! This is NOT the number of processes in the system!
static int pid_counter = 0;


static int set_pid_(struct hprocess* process, int pid);
static int set_state_(struct hprocess* process, int state);
static int increment_pid_counter_();
static int set_next_(struct hprocess* process, struct hprocess* next);
static int set_prev_(struct hprocess* process, struct hprocess* prev);
static int set_parent_queue_(struct hprocess* process, struct hsqueue* queue);
static int enqueue_process_callback_(struct hprocess* process, struct hprocess* last_process);
static int remove_process_callback_(struct hprocess* process);
static int get_pid_count_();




static int enqueue_process_callback_(struct hprocess* process, struct hprocess* last_process)
{
        set_next_(last_process, process);
        set_prev_(process, last_process);
        set_next_(process, NULL);

        return 0;
}


static int remove_process_callback_(struct hprocess* process)
{
        if (process == NULL)
                return -1;

        // Update internal structure.
        struct hprocess* prev = hprocess_get_prev(process);
        struct hprocess* next = hprocess_get_next(process);
        set_prev_(next, prev);
        set_next_(prev, next);

        return 0;
}


static int set_pid_(struct hprocess* process, int pid)
{
        if (process == NULL)
                return -1;
        else if (!hprocess_is_valid_pid(pid))
                return -1;

        process->pid = pid;

        return 0;
}


static int set_state_(struct hprocess* process, int state)
{
        if (process == NULL)
                return -1;

        process->state = state;

        return 0;
}


static int increment_pid_counter_()
{
        pid_counter += 1;

        return pid_counter;
}


static int get_pid_count_()
{
        return pid_counter;
}


static int set_next_(struct hprocess* process, struct hprocess* next)
{
        if (process == NULL)
                return -1;

        process->next = next;

        return 0;
}


static int set_prev_(struct hprocess* process, struct hprocess* prev)
{
        if (process == NULL)
                return -1;

        process->prev = prev;

        return 0;
}


static int set_parent_queue_(struct hprocess* process, struct hsqueue* queue)
{
        if (process == NULL)
                return -1;
        
        process->parent_queue = queue;
        
        return 0;
}


const int hprocess_base_priority()
{
        return 49;
}


const int hprocess_max_processes()
{
        // Maximum number of processes (set to 50 for testing only).
        return 50;
}

const int hprocess_max_filename_size()
{
        return HPROCESS_MAX_FILENAME_SIZE;
}


struct hprocess* hprocess_get(int pid)
{
        if (pid < 0 || pid > get_pid_count_())
                return NULL;

        /*  This is probably a quite naive way of
         *  searching for the process with the given
         *  PID.
         *
         *  All processes in the scheduler are essentially linked
         *  together in one long list, making the search time
         *  O(n), where n are the number of processes.
         */

        int list_i = 0;
        struct hsqueue* queue = NULL;
        struct hprocess* process = NULL;
        // Walk through all lists.
        for (list_i = 0; list_i < HPS_NUMBER; list_i++) {
                queue = hsqlist_get_first_queue(list_i);
                // Walk through all queues.
                while (queue != NULL) {
                        process = hsqueue_get_first_process(queue);
                        // Walk through all processes and return the one with the given pid.
                        while (process != NULL) {
                                if (hprocess_get_pid(process) == pid)
                                        return process;
                                process = hprocess_get_next(process);        
                        }
                        queue = hsqueue_get_next(queue);
                }
        }

        return NULL;
}


struct hprocess* hprocess_get_next(struct hprocess* process)
{
        if (process == NULL)
                return NULL;

        return process->next;
}


struct hprocess* hprocess_get_prev(struct hprocess* process)
{
        if (process == NULL)
                return NULL;

        return process->prev;
}


struct hsqueue* hprocess_get_parent_queue(struct hprocess* process)
{
        if (process == NULL)
                return NULL;

        return process->parent_queue;
}


int hprocess_set_bitfilename(struct hprocess* process, char* filename)
{
        if (process == NULL)
                return -1;
        else if (strlen(filename) > hprocess_max_filename_size())
                return -1;
        
        // Just keep enough memory to store the new string and nothing more.
        if (process->bitfilename == NULL || strlen(process->bitfilename) < strlen(filename))
                process->bitfilename = realloc(process->bitfilename, strlen(filename) + 1);

        strcpy(process->bitfilename, filename);

        return 0;
}


char* hprocess_get_bitfilename(struct hprocess* process)
{
        if (process == NULL)
                return NULL;

        return process->bitfilename;
}


int hprocess_remove(struct hprocess* process)
{
        if (process == NULL)
                return -1;
       
        // Remove the process from the queue (and update internal structure).
        int ret = hsqueue_remove_process(hprocess_get_parent_queue(process), process);
        free(process);

        return ret;
}


int hprocess_get_nice(struct hprocess* process)
{
        if (process == NULL)
                return -1;

        return process->nice;
}


int hprocess_get_pid(struct hprocess* process)
{
        if (process == NULL)
                return -1;

        return process->pid;
}


int hprocess_is_valid_pid(int pid)
{
        if (pid < 0 || pid > (hprocess_max_processes() - 1))
                return 0;

        return 1;
}


struct hprocess* hprocess_create(int nice)
{
        struct hprocess* process = calloc(1, sizeof(*process));
        set_next_(process, NULL);
        set_prev_(process, NULL);
        set_state_(process, HPS_NEW);
        set_parent_queue_(process, NULL);
        set_pid_(process, increment_pid_counter_());
        hprocess_set_nice(process, nice);

        hsqueue_register_signal(HSIGNAL_ADD_PROCESS, enqueue_process_callback_);
        hsqueue_register_signal(HSIGNAL_REMOVE_PROCESS, remove_process_callback_);

        return process;
}


int hprocess_set_priority(struct hprocess* process, int priority)
{
        if (process == NULL)
                return -1;

        if (priority > 0)
                process->priority = priority;
        else {
                hlog_write(HLOG_ERROR, "hprocess_set_priority: Can't set negative priority.\n");
                return -1;
        }

        return priority;
}

int hprocess_get_priority(struct hprocess* process)
{
        if (process == NULL)
                return -1;

        return process->priority;
}

int hprocess_set_nice(struct hprocess* process, int nice)
{
        if (process == NULL)
                return -1;

        process->nice = nice;
        return 0;
}


int hprocess_get_state(struct hprocess* process)
{
        if (process == NULL)
                return -1;

        return process->state;
}


