/*! \file hprocess.h
 *  \brief [Hardware OS Process Interface] Interface for a process on the FPGA.
 *
 *  Author (2011): Sindre Hansen
 *
 */

#ifndef HPROCESS_H
#define HPROCESS_H

#include "hstructures.h"

//! Lowest priority possible for a process.
//#define HPROCESS_LOWEST_PRIORITY 0

#define HPROCESS_MAX_FILENAME_SIZE 100

//! Number of possible states.
#define HPS_NUMBER 5
//! Definition of states for a process in the HWOS.
enum hpstate {
        HPS_NEW                  //! Not running, just registered. An application has just asked to run this process.
        ,HPS_READY               //! Not running, but is scheduled for placement and runtime on the FPGA. Not placed on FPGA.
        ,HPS_BLOCKED             //! Blocked (waiting for an event or preempted). Not placed on FPGA.
        ,HPS_BLOCKED_PLACED      //! Blocked (waiting for an event or preempted). Placed on FPGA.
        ,HPS_RUNNING             //! Running. Process is placed and running on the FPGA.
};


/*! \brief Get base priority.
 *
 *  Returns the priority given to processes when
 *  they are first accepted to the system.
 *
 *  @return Base priority of processes.
 */
const int hprocess_base_priority();

/*! \brief Get maximum processes in system.
 *
 *  @return Maximum number of processes in the system.
 */
const int hprocess_max_processes();

/*! \brief Set bitfilename for the process.
 *
 *  @param filename Filename.
 *  @param filename Name of file where the FPGA bitstream is.
 *  @return Maximum number of processes in the system.
 */
int hprocess_set_bitfilename(struct hprocess* process, char* filename);

/*! \brief Get bitfilename for the process.
 *
 *  @return Name of file where the FPGA bitstream is.
 */
char* hprocess_get_bitfilename(struct hprocess* process);

/*! \brief Get process by PID.
 *
 *  Worst case time of this function is O(n),
 *  where n is the number of processes.
 *
 *  @param pid Process ID.
 *  @return Pointer to the process on success. NULL on failure.
 */
struct hprocess* hprocess_get(int pid);

/*! \brief Create process.
 *
 *  @param nice User defined value to define priority for process.
 *  @return Process pointer on success. NULL on failure.
 */
struct hprocess* hprocess_create(int nice);

/*! \brief Set priority for process.
 *
 *  @param pid Process ID for process.
 *  @param priority Priority for process.
 *  @return The new priority (larger than 0) on success. Negative on failure.
 */
int hprocess_set_priority(struct hprocess* process, int priority);

/*! \brief Get priority for process.
 *
 *  @return The priority for the process on success. Negative on failure.
 */
int hprocess_get_priority(struct hprocess* process);

/*! \brief Set nice value for process.
 *
 *  This is a user defined value that affects the priority of the process.
 *
 *  @param process Affected process.
 *  @param nice Nice value for process.
 *  @return The new nice value (larger than 0) on success. Negative on failure.
 */
int hprocess_set_nice(struct hprocess* process, int nice);

/*! \brief Get nice value for process.
 *
 *  @param process Pointer to process.
 *  @return The nice value for the process on success. Negative on failure.
 */
int hprocess_get_nice(struct hprocess* process);

/*! \brief Get PID for process.
 *
 *  @param process Pointer to process.
 *  @return The PID for the process on success. Negative on failure.
 */
int hprocess_get_pid(struct hprocess* process);

/*! \brief Get next process in the list of processes.
 *
 *  @param process Pointer to process that has the next field.
 *  @return Pointer to the next process. NULL on failure.
 */
struct hprocess* hprocess_get_next(struct hprocess* process);

/*! \brief Get previous process in the list of processes.
 *
 *  @param process Pointer to process that has the previous field.
 *  @return Pointer to the previous process. NULL on failure.
 */
struct hprocess* hprocess_get_prev(struct hprocess* process);

/*! \brief Get state of process.
 *
 *  @param process Pointer to process.
 *  @return State (positive integer) for process. Negative on failure.
 */
int hprocess_get_state(struct hprocess* process);

/*! \brief Get state of process.
 *
 *  @param process Pointer to process.
 *  @return State (positive integer) for process. Negative on failure.
 */
int hprocess_is_valid_pid(int pid);

/*! \brief Remove process.
 *
 *  @param process Pointer to process.
 *  @return 0 on success. Negative on failure.
 */
int hprocess_remove(struct hprocess* process);

/*! \brief Get pointer to the queue the process belongs to.
 *
 *  @param process Pointer to process.
 *  @return Pointer to the queue on success. NULL on failure.
 */
struct hsqueue* hprocess_get_parent_queue(struct hprocess* process);

#endif

