#include <stdio.h>
#include <stdlib.h>

#include "hsqlist.h"
#include "hsqueue.h"
#include "hlog.h"
#include "hprocess.h"
#include "hsignal.h"


/*!  A simple structure for a list of queues.
 *  One instance of this structure points to one list of queues.
 *
 * There are only as many instances of this structure as there are
 * different process states (defined by the constant HPS_NUMBER).
 */
struct hsqlist {
        //! Number of queues in this list.
        int queues_num;
        //! First queue element in this list.
        struct hsqueue* first_queue;
        //! Last queue element in this list.
        struct hsqueue* last_queue;
};

/*! Array of queue lists.
 *  There is one list for each possible state in the scheduler.
 *  Warning: This will result in a compiler error if the number of NULL's on the right
 *  side of the expression does not match HPS_NUMBER.
 */
static struct hsqlist* queue_lists[HPS_NUMBER] = {NULL, NULL, NULL, NULL, NULL};

/*! \brief Call queue module when queue structure has changed.
 *  
 *  This private function connects to a private function in the
 *  queue module through hsqlist_register_signal.
 *
 */
static int (*signal_add_queue_)();
static int (*signal_remove_queue_)();

static struct hsqlist* get_list_(int list);
static int set_list_(int list, struct hsqlist* new_list);
static int set_size_(struct hsqlist* list_element, int size);
static int increment_size_(struct hsqlist* list_element);
static int decrement_size_(struct hsqlist* list_element);
static int set_last_queue_(struct hsqlist* list_element, struct hsqueue* queue);
static int set_first_queue_(struct hsqlist* list_element, struct hsqueue* queue);


static struct hsqlist* get_list_(int list)
{
        if (!hsqlist_is_valid_range(list))
                return NULL;

        return queue_lists[list];
}


static int set_list_(int list, struct hsqlist* new_list)
{
        if (!hsqlist_is_valid_range(list))
                return -1;

        queue_lists[list] = new_list;

        return 0;
}


static int set_size_(struct hsqlist* list_element, int size)
{
        if (list_element == NULL)
                return -1;

        list_element->queues_num = size;
        
        return 0;
}


static int increment_size_(struct hsqlist* list_element)
{
        if (list_element == NULL)
                return -1;

        list_element->queues_num += 1;
        
        return 0;
}


static int decrement_size_(struct hsqlist* list_element)
{
        if (list_element == NULL)
                return -1;

        if (list_element->queues_num < 1) {
                list_element->queues_num = 0;
                return -2;
        }

        list_element->queues_num -= 1;
        
        return 0;
}


static int set_first_queue_(struct hsqlist* list_element, struct hsqueue* queue)
{
        if (list_element == NULL)
                return -1;

        list_element->first_queue = queue;
        
        return 0;
}


static int set_last_queue_(struct hsqlist* list_element, struct hsqueue* queue)
{
        if (list_element == NULL)
                return -1;

        list_element->last_queue = queue;
        
        return 0;
}


int hsqlist_register_signal(int signal, int (*function)())
{
        switch (signal) {
                case HSIGNAL_ADD_QUEUE:
                        signal_add_queue_ = function;
                        return 0;
                case HSIGNAL_REMOVE_QUEUE:
                        signal_remove_queue_ = function;
                        return 0;
                default:
                        return -1;
        }

        return -1;
}


int hsqlist_remove(int list)
{
        struct hsqlist* list_element = get_list_(list);

        if (list_element == NULL)
                return -1;
        
        free(list_element);
        set_list_(list, NULL);
        
        return 0;
}


int hsqlist_size(int list)
{
        struct hsqlist* list_element = get_list_(list);

        if (list_element == NULL)
                return -1;

        return list_element->queues_num;
}


int hsqlist_remove_queue(int list, struct hsqueue* queue)
{
        //! This function can be called by hsqueue_remove.
        //! This function must remove queue even though the list does not exist
        //! (orphan queues may exist).
        
        if (queue == NULL)
                return -1;
        
        struct hsqlist* list_element = get_list_(list);

        if (queue == hsqlist_get_first_queue(list)) {
                set_first_queue_(list_element, hsqueue_get_next(queue));
        }

        if (queue == hsqlist_get_last_queue(list))
                set_last_queue_(list_element, hsqueue_get_prev(queue));

        decrement_size_(list_element);

        // Ask queue module to clean up internal structure and free queue from memory.
        signal_remove_queue_(queue);

        return 0;
}


int hsqlist_create(int list)
{
        if (!hsqlist_is_valid_range(list))
                return -1;

        if (get_list_(list) != NULL)
                return -1;
        
        struct hsqlist* new_list = calloc(1, sizeof(*new_list));
        set_size_(new_list, 0);
        set_first_queue_(new_list, NULL);
        set_last_queue_(new_list, NULL);

        set_list_(list, new_list);

        return 0;
}

int hsqlist_is_valid_range(int list)
{
        if ((list < 0) || (list > HPS_NUMBER)) {
                hlog_write(HLOG_ERROR, "hsqlist_is_valid_range: Invalid list range.\n");
                return 0;
        }

        return 1;
}


struct hsqueue* hsqlist_get(int list, int priority)
{
        if (!hsqlist_is_valid_range(list))
                return NULL;

        if (hsqlist_size(list) == 0)
                return NULL;

        // Iterate through the queues and return the queue
        // that contains a process with the given priority.
        struct hsqueue* queue = hsqlist_get_first_queue(list);

        if (queue == NULL)
                return NULL;

        struct hprocess* process = NULL;

        while (queue != NULL) {
                process = hsqueue_get_first_process(queue);
                if (hprocess_get_priority(process) == priority)
                        return queue;
                queue = hsqueue_get_next(queue);
        }

        // If this point is reached: No queue was found.
        return NULL;
}


int hsqlist_add_queue(int list, struct hsqueue* queue)
{
        struct hsqlist* list_element = get_list_(list);

        // Signalize a change in the queue structure to the queue module.
        // Must be done before setting the last queue.
        signal_add_queue_(queue, hsqlist_get_last_queue(list), NULL, list);

        if (hsqlist_size(list) == 0)
                set_first_queue_(list_element, queue);

        set_last_queue_(list_element, queue);

        return increment_size_(list_element);
}


int hsqlist_insert_queue_before(int list, struct hsqueue* second_queue, struct hsqueue* queue)
{
        struct hsqlist* list_element = get_list_(list);

        // Signalize a change in the queue structure to the queue module.
        signal_add_queue_(queue, hsqueue_get_prev(second_queue), second_queue, list);

        if (hsqlist_size(list) == 0) {
                set_first_queue_(list_element, queue);
                set_last_queue_(list_element, queue);
        } else if (hsqlist_get_first_queue(list) == second_queue)
                set_first_queue_(list_element, queue);

        return increment_size_(list_element);
}


int hsqlist_get_state(int list)
{
        struct hsqueue* first_queue = hsqlist_get_first_queue(list);

        if (first_queue == NULL)
                return -1;

        return hsqueue_get_state(first_queue);
}


struct hsqueue* hsqlist_get_first_queue(int list)
{
        struct hsqlist* list_element = get_list_(list);

        if (list_element == NULL)
                return NULL;

        return list_element->first_queue;
}


struct hsqueue* hsqlist_get_last_queue(int list)
{
        struct hsqlist* list_element = get_list_(list);

        if (list_element == NULL)
                return NULL;

        return list_element->last_queue;
}


