/*! \file hsqlist.h
 *  \brief [Hardware OS Scheduler Interface] Defines a list of process queues.
 *
 *  The main purpose of this module is to keep lists of process queues.
 *  Since there is a small and constant number of lists (as many as there are
 *  process states), lists are referenced by a number from 0 to HPS_NUMBER.
 *
 *  Author (2011): Sindre Hansen
 *
 */

#ifndef HSQLIST_H
#define HSQLIST_H


#include "hstructures.h"

/*! \brief Get size of list.
 *
 *  @param list The list.
 *  @return Size of the list.
 */
int hsqlist_size(int list);

/*! \brief Create new list.
 *
 *  @param list The list.
 *  @return 0 on success. Negative on failure.
 */
int hsqlist_create(int list);

/*! \brief Get queue by priority in list.
 *
 *  @param list The list.
 *  @param priority The priority of the processes in the queue.
 *  @return Pointer to the queue on success. NULL on failure.
 */
struct hsqueue* hsqlist_get_queue(int list, int priority);

/*! \brief Add a queue to the list.
 *
 *  @param list The list.
 *  @param queue The queue to be added.
 *  @return 0 on success. Negative on failure.
 */
int hsqlist_add_queue(int list, struct hsqueue* queue);

/*! \brief Insert queue before another queue in the list.
 *
 *  @param list The list.
 *  @param second_queue The queue that is already in the list.
 *  @param queue The queue to be inserted.
 *  @return 0 on success. Negative on failure.
 */
int hsqlist_insert_queue_before(int list, struct hsqueue* second_queue, struct hsqueue* queue);

/*! \brief Get state of the processes in the list of queues.
 *
 *  This functions assumes that this list has a list of queues
 *  and that the all processes in the queues has the same state.
 *
 *  @param list The list.
 *  @return State (positive integer) of the processes. Negative on failure.
 */
int hsqlist_get_state(int list);

/*! \brief Get first queue in the list.
 *
 *  @param list The list.
 *  @return Pointer to queue on success. NULL on failure.
 */
struct hsqueue* hsqlist_get_first_queue(int list);

/*! \brief Get last queue in the list.
 *
 *  @param list The list.
 *  @return Pointer to queue on success. NULL on failure.
 */
struct hsqueue* hsqlist_get_last_queue(int list);

/*! \brief Checks if the given list value is in valid range.
 *
 *  @param list The list.
 *  @return 1 if list value is in valid range. 0 if not.
 */
int hsqlist_is_valid_range(int list);

/*! \brief Remove the given queue from the list.
 *
 *  @param list The list.
 *  @param queue Pointer to the queue.
 *  @return 0 on success. Negative on failure.
 */
int hsqlist_remove_queue(int list, struct hsqueue* queue);

/*! \brief Remove the given list.
 *
 *  @param list The list.
 *  @return 0 on success. Negative on failure.
 */
int hsqlist_remove(int list);

/*! \brief Register the given signal.
 *
 *  This functionality is part of an
 *  observer-observable pattern
 *  between the list module and the queue
 *  module (to keep correct
 *  encapsulation).
 *
 *  The queue module register a given signal
 *  with the list module.
 *
 *  After a signal is registered, changes done
 *  in the list will make sure internal structure
 *  in the affected queues are updated.
 *
 *  @param list The signal as defined by enum hsignal in hsignal.h.
 *  @param function Pointer to the callback function in the hqueue module.
 *  @return 0 on success. Negative on failure.
 */
int hsqlist_register_signal(int signal, int (*function)());

#endif

