#include <stdio.h>
#include <stdlib.h>

#include "hsqueue.h"
#include "hsqlist.h"
#include "hlog.h"
#include "hprocess.h"
#include "hsignal.h"


//! A double linked list structure for a process queue.
//! One instance of this structure points to one queue of processes.
//!
//! Several instances of this structure forms a list of queues.
//! All queues in the same list will have processes with the same state.
struct hsqueue {
        //! Number of processes in the queue.
        int processes_num;
        //! Points to the parent list, if any.
        int parent_list;
        //! First element in the queue.
        struct hprocess* first_process;
        //! Last element in the queue.
        struct hprocess* last_process;
        //! Points to the next queue in this list.
	struct hsqueue* next;
        //! Points to the previous queue in this list.
	struct hsqueue* prev;
};

static int set_first_process_(struct hsqueue* queue, struct hprocess* process);
static int set_last_process_(struct hsqueue* queue, struct hprocess* process);
static int increment_size_(struct hsqueue* queue);
static int decrement_size_(struct hsqueue* queue);
static int set_size_(struct hsqueue* queue, int number);
static int set_prev_(struct hsqueue* element, struct hsqueue* element_prev);
static int set_next_(struct hsqueue* element, struct hsqueue* element_next);

/*! \brief Called after a queue has been added to a list.
 *
 *  This callback function should be called after a queue has been
 *  added to a hsqlist_t structure. The function will set the pointers
 *  for the internal queue structure correct.
 *  This is done through callback to make sure the public interface to this
 *  module is correctly encapsulated.
 *
 *  @param queue The queue that should be added to the list.
 *  @param queue_prev The new queue's next queue in list.
 *  @param queue_next The new queue's previous queue in list.
 *  @return Positive value on success. Negative value on failure.
 *
 */
static int add_queue_callback_(struct hsqueue* queue, struct hsqueue* queue_prev, struct hsqueue* queue_next, int list);

static int remove_queue_callback_(struct hsqueue* queue);

static int set_parent_list_(struct hsqueue* queue, int list);

/*! \brief Function pointer to private function in process module.
 *  
 *  Makes it possible for this module to signalize a change in the internal
 *  structure of linked processes to the process module.
 *
 */
static int (*signal_enqueue_process_)();

static int (*signal_remove_process_)();


static int add_queue_callback_(struct hsqueue* queue, struct hsqueue* queue_prev, struct hsqueue* queue_next, int list)
{
        set_parent_list_(queue, list);
        set_prev_(queue, queue_prev);
        set_next_(queue, queue_next);
        set_prev_(queue_next, queue);
        set_next_(queue_prev, queue);

        return 0;
}


static int remove_queue_callback_(struct hsqueue* queue)
{
        if (queue == NULL)
                return -1;

        // Update internal structure and remove queue from memory.

        struct hsqueue* prev = hsqueue_get_prev(queue);
        struct hsqueue* next = hsqueue_get_next(queue);
        set_prev_(next, prev);
        set_next_(prev, next);

        free(queue);

        return 0;
}


static int set_prev_(struct hsqueue* element, struct hsqueue* element_prev)
{
        if (element == NULL)
                return -1;

        element->prev = element_prev;
        return 0;
}


static int set_next_(struct hsqueue* element, struct hsqueue* element_next)
{
        if (element == NULL)
                return -1;

        element->next = element_next;
        return 0;
}


static int set_first_process_(struct hsqueue* queue, struct hprocess* process)
{
        if (queue == NULL)
                return -1;

        queue->first_process = process;
        return 0;
}


static int set_last_process_(struct hsqueue* queue, struct hprocess* process)
{
        if (queue == NULL)
                return -1;

        queue->last_process = process;
        return 0;
}


static int increment_size_(struct hsqueue* queue)
{
        if (queue == NULL)
                return -1;
        
        queue->processes_num += 1;
        
        return queue->processes_num;
}


static int decrement_size_(struct hsqueue* queue)
{
        if (queue == NULL)
                return -1;

        if (queue->processes_num < 1) {
                queue->processes_num = 0;
                return 0;
        }
        
        queue->processes_num = queue->processes_num - 1;
        
        return queue->processes_num;
}


static int set_size_(struct hsqueue* queue, int number)
{
        if (queue == NULL)
                return -1;

        queue->processes_num = number;

        return queue->processes_num;
}


static int set_parent_list_(struct hsqueue* queue, int list)
{
        if (queue == NULL)
                return -1;

        queue->parent_list = list;

        return 0;
}


int hsqueue_register_signal(int signal, int (*function)())
{
        switch (signal) {
                case HSIGNAL_ADD_PROCESS:
                        signal_enqueue_process_ = function;
                        return 0;
                case HSIGNAL_REMOVE_PROCESS:
                        signal_remove_process_ = function;
                        return 0;
                default:
                        return -1;
        }

        return -1;
}


int hsqueue_get_parent_list(struct hsqueue* queue)
{
        if (queue == NULL)
                return -1;
        else if (queue->parent_list < 0)
                return -1;

        return queue->parent_list;
}


int hsqueue_size(struct hsqueue* element)
{
        if (element == NULL)
                return -1;

        return element->processes_num;
}


int hsqueue_get_state(struct hsqueue* queue)
{
        return hprocess_get_state( hsqueue_get_first_process(queue) );
}


struct hsqueue* hsqueue_get_prev(struct hsqueue* element)
{
        if (element == NULL)
                return NULL;

        return (struct hsqueue*)(element->prev);
}


struct hsqueue* hsqueue_get_next(struct hsqueue* element)
{
        if (element == NULL)
                return NULL;

        return element->next;
}


struct hsqueue* hsqueue_create()
{
        struct hsqueue* queue = calloc(1, sizeof(*queue));
        set_next_(queue, NULL);
        set_prev_(queue, NULL);
        set_first_process_(queue, NULL);
        set_last_process_(queue, NULL);
        set_size_(queue, 0);
        set_parent_list_(queue, -1);

        hsqlist_register_signal(HSIGNAL_ADD_QUEUE, add_queue_callback_);
        hsqlist_register_signal(HSIGNAL_REMOVE_QUEUE, remove_queue_callback_);

        return queue;
}


int hsqueue_enqueue_process(struct hsqueue* queue, struct hprocess* process)
{
        if (process == NULL)
                return -1;

        struct hprocess* last_process = hsqueue_get_last_process(queue);

        // Signalize change in process structure to process module.
        signal_enqueue_process_(process, last_process);

        if (last_process == NULL) {
               set_first_process_(queue, process);
        }

        set_last_process_(queue, process);

        return increment_size_(queue);
}


int hsqueue_remove_process(struct hsqueue* queue, struct hprocess* process)
{
        //! This function can be called by hprocess_remove.
        //! This function must remove process even though the queue does not exist
        //! (orphan processes may exist).
        
        if (process == NULL)
                return -1;
        
        if (process == hsqueue_get_first_process(queue))
                set_first_process_(queue, hprocess_get_next(process));

        if (process == hsqueue_get_last_process(queue))
                set_last_process_(queue, hprocess_get_prev(process));

        decrement_size_(queue);

        // Ask process module to clean up internal structure (but not remove process from memory).
        signal_remove_process_(process);

        return 0;
}

struct hprocess* hsqueue_dequeue_process(struct hsqueue* queue)
{
        if (queue == NULL)
                return NULL;

        //! Dequeue is the same as removing the first process in the queue.
        struct hprocess* process = hsqueue_get_first_process(queue);
        if (hsqueue_remove_process(queue, process) == 0)
                return process;
        else
                return NULL;
}

int hsqueue_remove(struct hsqueue* queue)
{
        if (queue == NULL)
                return -1;

        // Ask list module to remove this queue.
        return hsqlist_remove_queue(hsqueue_get_parent_list(queue), queue);
}


struct hprocess* hsqueue_get_first_process(struct hsqueue* queue)
{
        if (queue == NULL)
                return NULL;

        return queue->first_process; 
}


struct hprocess* hsqueue_get_last_process(struct hsqueue* queue)
{
        if (queue == NULL)
                return NULL;

        return queue->last_process;
}

