/*! \file hsqueue.h
 *  \brief [Hardware OS Scheduler Interface] Functionality for process queues.
 *
 *  The main purpose of this module is to keep track of queues of processes.
 *  There is also functionality for linking several queues together in a double linked list.
 *
 *  Each queue contains processes of the same priority. Each list of queues contains queues
 *  of different priorities, but with the same state.
 *
 *  Author (2011): Sindre Hansen
 *
 */

#ifndef HSQUEUE_H
#define HSQUEUE_H

#include "hstructures.h"

//! Defines the max amount of hsqueues in the system.
#define HSQUEUE_MAX_QUEUES 1000


/*! \brief Get size of queue.
 *
 *  @param element The queue.
 *  @return Size of the queue.
 */
int hsqueue_size(struct hsqueue* element);

/*! \brief Get state of the processes in the queue.
 *
 *  This functions assumes that all processes in the
 *  queue has the same state.
 *
 *  @param queue Pointer to the queue.
 *  @return State (positive integer) of the processes. Negative on failure.
 */
int hsqueue_get_state(struct hsqueue* queue);

/*! \brief Get previous queue in list.
 *
 *  @param element Queue element that has the previous field.
 *  @return Pointer to queue on success. NULL on failure.
 */
struct hsqueue* hsqueue_get_prev(struct hsqueue* element);

/*! \brief Get next queue in list.
 *
 *  @param element Queue element that has the next field.
 *  @return Pointer to queue on success. NULL on failure.
 */
struct hsqueue* hsqueue_get_next(struct hsqueue* element);

/*! \brief Create a new queue.
 *
 *  @return Pointer to queue on success. NULL on failure.
 */
struct hsqueue* hsqueue_create();

/*! \brief Remove a given queue.
 *
 *  @param element Queue element.
 *  @return 0 on success. Negative on failure.
 */
int hsqueue_remove(struct hsqueue* element);

/*! \brief Get first process in the queue.
 *
 *  @param element Queue element.
 *  @return Pointer to the process on success. NULL on failure.
 */
struct hprocess* hsqueue_get_first_process(struct hsqueue* queue);

/*! \brief Get last process in the queue.
 *
 *  @param element Queue element.
 *  @return Pointer to the process on success. NULL on failure.
 */
struct hprocess* hsqueue_get_last_process(struct hsqueue* queue);

/*! \brief Enqueue the given process.
 *
 *  @param queue Queue element.
 *  @param process Process element.
 *  @return 0 on success. Negative on failure.
 */
int hsqueue_enqueue_process(struct hsqueue* queue, struct hprocess* process);

/*! \brief Dequeue the given process.
 *
 *  @param queue Queue element.
 *  @return Pointer to the dequeued process on success. NULL on failure.
 */
struct hprocess* hsqueue_dequeue_process(struct hsqueue* queue);

/*! \brief Removes the given process from queue.
 *
 *  This function does NOT free the given process from
 *  memory.
 *
 *  @param queue Queue element.
 *  @param process Process element.
 *  @return 0 on success. Negative on failure.
 */
int hsqueue_remove_process(struct hsqueue* queue, struct hprocess* process);

/*! \brief Get parent list of the queue.
 *
 *  @param queue Queue element.
 *  @return Positive integer (corresponding to a process state) describing the list on success. Negative on failure.
 */
int hsqueue_get_parent_list(struct hsqueue* queue);

/*! \brief Register the given signal.
 *
 *  This functionality is part of an
 *  observer-observable pattern
 *  between the queue module and the process
 *  module (to keep correct
 *  encapsulation).
 *
 *  The process module register a given signal
 *  with the queue module.
 *
 *  After a signal is registered, changes done
 *  in the queue will make sure the internal
 *  structure in the affected processes are updated.
 *
 *  @param list The signal as defined by enum hsignal in hsignal.h.
 *  @param function Pointer to the callback function in the hprocess module.
 *  @return 0 on success. Negative on failure.
 */
int hsqueue_register_signal(int signal, int (*function)());

#endif

