#include <stdio.h>
#include <stdlib.h>

#include "hvmem.h"
#include "hvmemptr.h"
#include "hlist.h"

// TODO: Should not be defined here?
#define INSTR_BUF_SIZE 10
#define POOL_SIZE 16
#define BACKEND "/var/id_backend"

//! The size of the physical BRAM memory (in bytes) including any reserved memory.
#define BRAM_SIZE 10

//! Get double linked list of all physically allocated BRAM entries.
//! This entries are always sorted in ascending order.
static struct hlist* get_bram_entries_();
//! Create double linked list of all physically allocated BRAM entries.
static int create_bram_entries_();
//! Physical BRAM address to the first byte in BRAM (is other than 0 if some memory is reserved).
static const int get_first_bram_();
//! Get physical address of the given entry.
static struct hlorphan* create_entry_(int address);

static struct hlist* bram_entries = NULL;
static struct hlist* get_bram_entries_()
{
        return bram_entries;
}

static int create_bram_entries_()
{
        bram_entries = hlist_create();

        return 0;
}

static const int first_bram = 0;
static const int get_first_bram_()
{
        return first_bram;
}


static int get_entry_address_(struct hlelement* element)
{
        if (element == NULL)
                return -1;

        return *(int*)hlorphan_get_data( hlelement_get_data(element) );
}


static struct hlorphan* create_entry_(int address)
{
        int* entry = malloc(sizeof(int));
        *entry = address;
        return hlorphan_create(entry);
}


int hvmem_get_free_count()
{
        struct hlelement* bram_i = hlist_get_first(get_bram_entries_());
        int bytes_free = 0;
        int last_alloc_bram = get_first_bram_() - 1;
        while (bram_i != NULL) {
                bytes_free += get_entry_address_(bram_i) - last_alloc_bram - 1;
                last_alloc_bram = get_entry_address_(bram_i);
                bram_i = hlelement_get_next(bram_i);
        }

        if ( get_entry_address_(hlist_get_last(get_bram_entries_())) < BRAM_SIZE - 1 )
                bytes_free += BRAM_SIZE - get_entry_address_(hlist_get_last(get_bram_entries_())) - 1;

        return bytes_free;
}


void hvmem_print_entries()
{
        struct hlelement* bram_i = hlist_get_first(get_bram_entries_());
        while (bram_i != NULL) {
                //printf("bram_i, data: %d\n", get_entry_address_(bram_i) );
                bram_i = hlelement_get_next(bram_i);
        }
}


struct hvmemptr* hvmem_allocate(struct hprocess* process, int size)
{
        if (process == NULL || size < 1)
                return NULL;

        if (get_bram_entries_() == NULL)
                create_bram_entries_();

        struct hlelement* bram_i = hlist_get_first(get_bram_entries_());
        struct hlelement* new_entry = NULL;
        struct hlorphan* bram_entry = NULL;
        int bytes_allocated = 0;

        if (bram_i == NULL) {
                bram_entry = create_entry_(get_first_bram_());
                bytes_allocated += 1;
                bram_i = hlist_enqueue(get_bram_entries_(), bram_entry);
                new_entry = bram_i;
                //printf("hvmem_allocate: adding first=%d\n", get_entry_address_(bram_i));
        }

        struct hlelement* bram_j = bram_i;

        // Insert elements before first element if possible.
        while ( get_entry_address_(bram_i) > get_first_bram_() && bytes_allocated < size) {
                bram_entry = create_entry_(get_entry_address_(bram_i) - 1);
                bytes_allocated += 1;
                bram_i = hlist_insert_before(get_bram_entries_(), bram_i, bram_entry);
                //printf("hvmem_allocate: inserting before, data=%d\n", get_entry_address_(bram_i));
        }

        if (bytes_allocated > 0)
                new_entry = bram_i;

        struct hlelement* bram_k = NULL; 

        int bram_j_next_addr = -1;
        while (bram_j != NULL && bytes_allocated < size) {
                bram_k = bram_j;
                if (hlelement_get_next(bram_j) == NULL)
                        bram_j_next_addr = BRAM_SIZE;
                else
                        bram_j_next_addr = get_entry_address_(hlelement_get_next(bram_j));

                if ( get_entry_address_(bram_j) + (size - bytes_allocated) + 1 < bram_j_next_addr)
                        bram_j_next_addr = get_entry_address_(bram_j) + (size - bytes_allocated) + 1;

                //printf("hvmem_allocate: bram_j_next_addr=%d\n", bram_j_next_addr);
                while ( bram_j_next_addr > get_entry_address_(bram_k) + 1) {
                        bram_entry = create_entry_(get_entry_address_(bram_k) + 1);
                        bram_k = hlist_insert_after(get_bram_entries_(), bram_k, bram_entry);
                        if (bytes_allocated == 0)
                                new_entry = bram_k;
                        bytes_allocated += 1;
                        //printf("hvmem_allocate: adding after, data=%d\n", get_entry_address_(bram_k));
                }
                bram_j = hlelement_get_next(bram_j);
        }
       

        if (new_entry != NULL && bytes_allocated != size) {
                free(new_entry);
                return NULL;
        } else if (new_entry != NULL) {
                return hvmemptr_create(new_entry, size);
        } else
                return NULL;
}


int hvmem_free(struct hprocess* process, struct hvmemptr* pointer)
{
        if (process == NULL || pointer == NULL)
                return -1;

        struct hlelement* bram_i = hvmemptr_get_bramentry(pointer);
        int size = hvmemptr_get_size(pointer);
        int bytes_removed = 0;
        struct hlelement* bram_next = NULL;
        while (bram_i != NULL && bytes_removed < size) {
                bram_next = hlelement_get_next(bram_i);
                hlelement_remove(bram_i, get_bram_entries_());
                bram_i = bram_next;
                bytes_removed += 1;
        }

        if (bytes_removed != size)
                return -2;      // Fatal error (removed some bytes, but not all the allocated ones).
        else
                return bytes_removed;
}


