/*! \file hvmem.h
 *  \brief [Hardware OS Virtual Memory Interface] Keeps track of each process' BRAM memory.
 *  
 *  Based on known concepts in OS-theory, this interface for each process to allocate,
 *  read, write and free BRAM memory has been made.
 *  The memory is called virtual because allocated memory may or may not be contiguous
 *  in BRAM. There may be possible to swap memory to disk in future versions.
 *
 *  The underlying implementation is based on double linked lists.
 *
 *  NOTE: The implementation of hvmem was just started. It may be possible to use
 *  some of this code when doing further development of a virtual memory implementation.
 *
 *  Author (2011): Sindre Hansen
 */


#ifndef HVMEM_H
#define HVMEM_H

#include "hstructures.h"


void hvmem_print_entries();

/*! \brief Allocate virtual memory of the given size.
 *  
 *  Note that the allocated memory may or may not be
 *  stored contiguous in BRAM.
 *
 *  @param process Process that should own the memory.
 *  @param size Allocate memory of the given size.
 *  @return Virtual pointer to the allocated memory. NULL on failure.
 */
struct hvmemptr* hvmem_allocate(struct hprocess* process, int size);

/*! \brief Return total size of free memory.
 *  
 *  Note that this is a quite expensive method as
 *  it will have to walk through all the allocated
 *  memory.
 *
 *  @return Total number of free bytes in memory.
 */
int hvmem_get_free_count();

/*! \brief Free virtual memory pointed to by the virtual pointer.
 *
 *  @param process Process that owns the memory.
 *  @param size Allocate memory of the given size.
 *  @return Physical address to first byte in BRAM. Negative on failure.
 */
int hvmem_free(struct hprocess* process, struct hvmemptr* pointer);

/*! \brief Write data to the virtual memory.
 *
 *  @param process Process that owns the memory.
 *  @param pointer Virtual pointer to the memory.
 *  @param offset Start writing at this offset in memory.
 *  @param bytes How many bytes should be written.
 *  @param data Pointer to physical data to be written.
 *  @return 0 on success. Negative on failure.
 */
int hvmem_write(struct hprocess* process, struct hvmemptr* pointer, int offset, int bytes, char* data);

/*! \brief Read data from the virtual memory.
 *
 *  @param process Process that owns the memory.
 *  @param pointer Virtual pointer to the memory.
 *  @param offset Start writing at this offset in memory.
 *  @param bytes How many bytes should be written.
 *  @return 0 on success. Negative on failure.
 */
int hvmem_read(struct hprocess* process, struct hvmemptr* pointer, int offset, int bytes);

#endif

