#include <stdlib.h>
#include <stdio.h>
#include <check.h>

#include <hmqueue.h>
#include <hlog.h>
#include <hdev.h>
#include "test_hwos.h"

static void create_enqueued_hdevs();
static void setup (void);
static void teardown (void);
static Suite* hdev_suite (void);
void remove_enqueued_hdevs (void);


//! Enqueued message queues. First enqueued is on index 0. Next is on index 1 and so on. Last enqueued is the last index.
static struct hdev* enqueued_hdevs[TEST_HDEV_SIZE];


static void setup(void)
{
        create_enqueued_hdevs();
}

void create_enqueued_hdevs (void)
{
        int i;
        // Setup all devices.
        char *teststr = calloc(1, sizeof(char) * HDEV_PATH_SIZE);
        for (i=0; i < TEST_HDEV_SIZE; i++) {
                sprintf(teststr, "/dev/dev%d\0", i);
                enqueued_hdevs[i] = hdev_add(i, teststr);
        }
}

void remove_enqueued_hdevs (void)
{
        int i;
        // Remove all devices.
        for (i=0; i < TEST_HDEV_SIZE; i++) {
                hdev_remove(enqueued_hdevs[i]);
                enqueued_hdevs[i] = NULL;
        }
}


static void teardown (void)
{
        remove_enqueued_hdevs();
}


START_TEST (test_queue_create)
{
        fail_unless (hdev_size() == TEST_HDEV_SIZE, "Queue size is not correct!");
        fail_unless (hdev_get_first() == enqueued_hdevs[0], "First failure!");
        fail_unless (hdev_get_last() == enqueued_hdevs[TEST_HDEV_SIZE - 1], "Last failure!");
}
END_TEST


START_TEST (test_enqueue)
{
        int i;

        fail_unless (hdev_size() == 0, "Incorrect queue size!");
        fail_unless (hdev_get_first() == NULL, "First failure!");
        fail_unless (hdev_get_last() == NULL, "First failure!");


        char *teststr = calloc(1, sizeof(char) * HDEV_PATH_SIZE);

        // i goes from 0 to TEST_HDEV_SIZE - 1.
        for (i = 0; i < TEST_HMQUEUE_SIZE; i++)
        {
                sprintf(teststr, "/dev/dev%d\0", i);
                // Enqueue one more device.
                enqueued_hdevs[i] = hdev_add(i, teststr);

                fail_unless (hdev_size() == i + 1, "Incorrect queue size!");

                fail_unless (hdev_get_first() == enqueued_hdevs[0], "First failure! i=%d", i);
                fail_unless (hdev_get_last() == enqueued_hdevs[i], "Last failure!");
                fail_unless (hdev_get_next(enqueued_hdevs[i]) == NULL, "Next field for last should be NULL!");
                fail_unless (hdev_get_prev(enqueued_hdevs[0]) == NULL, "Prev field for first should be NULL!");
                if (i > 0) {
                        fail_unless (hdev_get_prev(enqueued_hdevs[i]) == enqueued_hdevs[i - 1], "Prev field failure!");
                        fail_unless (hdev_get_next(enqueued_hdevs[i - 1]) == enqueued_hdevs[i], "Next field failure!");
                }
                else if (i == 0)
                        fail_unless (hdev_get_prev(enqueued_hdevs[i]) == NULL, "Prev field failure!");
        }
}
END_TEST


START_TEST (test_dequeue)
{
        int i, size_now;

        // i goes from 0 to TEST_HDEV_SIZE - 1.
        for (i = 0; i < TEST_HDEV_SIZE; i++)
        {
                // size_now goes from TEST_HDEV_SIZE to 1.
                size_now = TEST_HDEV_SIZE - i;

                fail_unless (hdev_size() == size_now, "Incorrect queue size!");

                fail_unless (hdev_get_first() == enqueued_hdevs[i], "First failure!");
                fail_unless (hdev_get_last() == enqueued_hdevs[TEST_HDEV_SIZE - 1], "Last failure!");
                fail_unless (hdev_get_prev(enqueued_hdevs[i]) == NULL, "Prev field for first should be NULL! i=%d", i);
                fail_unless (hdev_get_next(enqueued_hdevs[TEST_HDEV_SIZE - 1]) == NULL, "Next field for last should be NULL!");
              
                if (i < TEST_HDEV_SIZE - 1) {
                        fail_unless (hdev_get_next(enqueued_hdevs[i]) == enqueued_hdevs[i + 1], "Next process failure!");
                        fail_unless (hdev_get_prev(enqueued_hdevs[i + 1]) == enqueued_hdevs[i], "Prev process failure!");
                }

                // Remove one more queue.
                hdev_remove(enqueued_hdevs[i]);
        }

        // All hdevs have been removed.
        fail_unless (hdev_size() == 0, "Incorrect queue size!");
        fail_unless (hdev_get_first() == NULL, "First failure!");
        fail_unless (hdev_get_last() == NULL, "Last failure!");
}
END_TEST



static Suite* hdev_suite (void)
{
        Suite *s = suite_create ("hdev");

        TCase *tc_core = tcase_create ("Core");
        tcase_add_checked_fixture (tc_core, setup, teardown);
        tcase_add_test (tc_core, test_queue_create);
        suite_add_tcase (s, tc_core);

        TCase *tc_enq = tcase_create ("Enqueue");
        tcase_add_checked_fixture (tc_enq, NULL, teardown);
        tcase_add_test (tc_enq, test_enqueue);
        suite_add_tcase (s, tc_enq);

        TCase *tc_deq  = tcase_create ("Dequeue");
        tcase_add_checked_fixture (tc_deq, setup, NULL);
        tcase_add_test (tc_deq, test_dequeue);
        suite_add_tcase (s, tc_deq);
        

        return s;
}

int test_hdev (void)
{
        int number_failed;
        Suite *s = hdev_suite ();
        SRunner *sr = srunner_create (s);
        srunner_set_log (sr, TEST_LOG_HDEV);
        srunner_run_all (sr, CK_VERBOSE);

        number_failed = srunner_ntests_failed (sr);
        srunner_free (sr);
        return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

