#include <stdlib.h>
#include <stdio.h>
#include <check.h>

#include <hlist.h>
#include <hprocess.h>
#include <hlog.h>
#include "test_hwos.h"


static struct hlist* queue0 = NULL;

static void create_enqueued_processes(void);
static void remove_enqueued_processes(void);
static void setup(void);
static void setup_queue(void);
static void teardown(void);
static void teardown_queue(void);
static Suite* hlist_suite (void);


//! Enqueued processes. First enqueued is on index 0. Next is on index 1 and so on. Last enqueued is the last index.
static struct hprocess* enqueued_processes[TEST_HQUEUE_SIZE];


static void setup(void)
{
        create_enqueued_processes();
        queue0 = hlist_create();
}

static void create_enqueued_processes (void)
{
        int i;
        // Setup all processes. Give them nice value i.
        for (i=0; i < TEST_HQUEUE_SIZE; i++)
                enqueued_processes[i] = hprocess_create(i);
}

static void remove_enqueued_processes (void)
{
        int i;
        // Remove all processes from memory.
        for (i=0; i < TEST_HQUEUE_SIZE; i++) {
                hprocess_remove( enqueued_processes[i] );
        }
}


static void setup_queue (void)
{
        setup();
        int i;
        // Add all processes to queue.
        for (i=0; i < TEST_HQUEUE_SIZE; i++)
                hlist_enqueue(queue0, enqueued_processes[i]);
}


static void teardown (void)
{
        remove_enqueued_processes();
        teardown_queue();
}

static void teardown_queue(void)
{
        hlist_remove(queue0);
        queue0 = NULL;
}

START_TEST (test_queue_create)
{
        fail_unless (queue0 != NULL, "Queue is NULL!");
        fail_unless (hlist_size(queue0) == 0, "Queue size is not 0!");
        fail_unless (hlist_get_first(queue0) == NULL, "First process failure!");
        fail_unless (hlist_get_last(queue0) == NULL, "Last process failure!");
}
END_TEST


START_TEST (test_enqueue)
{
        int i;

        fail_unless (hlist_size(queue0) == 0, "Incorrect queue size!");
        fail_unless (hlist_get_first(queue0) == NULL, "First process failure!");
        fail_unless (hlist_get_last(queue0) == NULL, "First process failure!");

        struct hlelement* element = NULL;
        // i goes from 1 to TEST_HQUEUE_SIZE - 1.
        for (i = 0; i < TEST_HQUEUE_SIZE; i++)
        {
                // Enqueue one more process.
                element = hlist_enqueue(queue0, enqueued_processes[i]);

                fail_unless (hlist_size(queue0) == i + 1, "Incorrect queue size!");

                fail_unless (hlelement_get_data(element) == enqueued_processes[i], "First process failure!");
                fail_unless (hlelement_get_data(hlist_get_first(queue0)) == enqueued_processes[0], "First process failure!");
                fail_unless (hlelement_get_data(hlist_get_last(queue0)) == enqueued_processes[i], "Last process failure!");
                fail_unless (hlelement_get_next(hlist_get_last(queue0)) == NULL, "Next field for last element should be NULL!");
                fail_unless (hlelement_get_prev(hlist_get_first(queue0)) == NULL, "Prev field for first element should be NULL!");
                if (i > 0) {
                        fail_unless (hlelement_get_data( hlelement_get_prev(element) ) == enqueued_processes[i - 1], "Prev field failure!");
                        fail_unless (hlelement_get_data( hlelement_get_next( hlelement_get_prev(element) ) ) == enqueued_processes[i], "Next field failure!");
                }
                else if (i == 0) {
                        fail_unless (hlelement_get_prev(element) == NULL, "Prev field failure!");
                        fail_unless (hlelement_get_next(element) == NULL, "Next field failure!");
                }
        }
}
END_TEST


START_TEST (test_insert_inbetween)
{
        // TODO: This method can be rewritten to the same form as test_enqueue.

        fail_unless (hlist_size(queue0) == 0, "Incorrect queue size!");
        fail_unless (hlist_get_first(queue0) == NULL, "First process failure!");
        fail_unless (hlist_get_last(queue0) == NULL, "First process failure!");

        struct hlelement* elements[9];
        elements[0] = hlist_enqueue(queue0, enqueued_processes[0]);
        elements[1] = hlist_enqueue(queue0, enqueued_processes[1]);
        elements[2] = hlist_enqueue(queue0, enqueued_processes[2]);
        // 0, 1, 2

        elements[3] = hlist_insert_after(queue0, elements[2], enqueued_processes[3]);
        // 0, 1, 2, 3
        fail_unless (hlelement_get_prev(elements[0]) == NULL, "Prev field failure!");
        fail_unless (hlelement_get_next(elements[3]) == NULL, "Next field failure!");
        fail_unless (hlelement_get_next(elements[2]) == elements[3], "Next field failure!");
        fail_unless (hlelement_get_prev(elements[3]) == elements[2], "Prev field failure!");
        elements[4] = hlist_insert_after(queue0, elements[0], enqueued_processes[4]);
        // 0, 4, 1, 2, 3
        fail_unless (hlelement_get_prev(elements[0]) == NULL, "Prev field failure!");
        fail_unless (hlelement_get_next(elements[0]) == elements[4], "Next field failure!");
        fail_unless (hlelement_get_next(elements[4]) == elements[1], "Next field failure!");
        fail_unless (hlelement_get_prev(elements[4]) == elements[0], "Prev field failure!");
        elements[5] = hlist_insert_after(queue0, elements[2], enqueued_processes[4]);
        // 0, 4, 1, 2, 5, 3
        fail_unless (hlelement_get_prev(elements[2]) == elements[1], "Prev field failure!");
        fail_unless (hlelement_get_next(elements[2]) == elements[5], "Next field failure!");
        fail_unless (hlelement_get_next(elements[5]) == elements[3], "Next field failure!");
        fail_unless (hlelement_get_prev(elements[5]) == elements[2], "Prev field failure!");
        elements[6] = hlist_insert_before(queue0, elements[0], enqueued_processes[5]);
        // 6, 0, 4, 1, 2, 5, 3
        fail_unless (hlelement_get_prev(elements[6]) == NULL, "Prev field failure!");
        fail_unless (hlelement_get_next(elements[6]) == elements[0], "Next field failure!");
        fail_unless (hlelement_get_next(elements[0]) == elements[4], "Next field failure!");
        fail_unless (hlelement_get_prev(elements[0]) == elements[6], "Prev field failure!");
        fail_unless (hlist_get_first(queue0) == elements[6], "First field failure!");
        elements[7] = hlist_insert_before(queue0, elements[6], enqueued_processes[5]);
        // 7, 6, 0, 4, 1, 2, 5, 3
        fail_unless (hlelement_get_prev(elements[7]) == NULL, "Prev field failure!");
        fail_unless (hlelement_get_next(elements[7]) == elements[6], "Next field failure!");
        fail_unless (hlelement_get_next(elements[6]) == elements[0], "Next field failure!");
        fail_unless (hlelement_get_prev(elements[6]) == elements[7], "Prev field failure!");
        elements[8] = hlist_insert_before(queue0, elements[1], enqueued_processes[5]);
        // 7, 6, 0, 4, 8, 1, 2, 5, 3
        fail_unless (hlelement_get_prev(elements[8]) == elements[4], "Prev field failure!");
        fail_unless (hlelement_get_next(elements[8]) == elements[1], "Next field failure!");
        fail_unless (hlelement_get_next(elements[4]) == elements[8], "Next field failure!");
        fail_unless (hlelement_get_prev(elements[4]) == elements[0], "Prev field failure!");
        fail_unless (hlelement_get_next(elements[1]) == elements[2], "Next field failure!");
        fail_unless (hlelement_get_prev(elements[1]) == elements[8], "Prev field failure!");
}
END_TEST


START_TEST (test_dequeue)
{
        int i, size_now;

        // i goes from 0 to TEST_HQUEUE_SIZE - 1.
        for (i = 0; i < TEST_HQUEUE_SIZE; i++)
        {
                // size_now goes from TEST_HQUEUE_SIZE to 1.
                size_now = TEST_HQUEUE_SIZE - i;

                fail_unless (hlist_size(queue0) == size_now, "Incorrect queue size!");

                fail_unless (hlelement_get_data(hlist_get_first(queue0)) == enqueued_processes[i], "First process failure!");
                fail_unless (hlelement_get_data(hlist_get_last(queue0)) == enqueued_processes[TEST_HQUEUE_SIZE - 1], "Last process failure!");
                fail_unless (hlelement_get_prev(hlist_get_first(queue0)) == NULL, "Prev field for first process should be NULL! i=%d", i);
                fail_unless (hlelement_get_next(hlist_get_last(queue0)) == NULL, "Next field for last process should be NULL!");
              
                if (i < TEST_HQUEUE_SIZE - 1) {
                        fail_unless (hlelement_get_data( hlelement_get_next( hlist_get_first(queue0) ) ) == enqueued_processes[i + 1], "Next process in queue failure!");
                        fail_unless (hlelement_get_data( hlelement_get_prev( hlelement_get_next( hlist_get_first(queue0) ) ) ) == enqueued_processes[i], "Prev process in queue failure!");
                }

                // Remove one more process.
                hlist_dequeue(queue0);
        }

        // All processes have been removed.
        fail_unless (hlist_size(queue0) == 0, "Incorrect queue size!");
        fail_unless (hlist_get_first(queue0) == NULL, "First process failure!");
        fail_unless (hlist_get_last(queue0) == NULL, "Last process failure!");
}
END_TEST




static Suite* hlist_suite (void)
{
        Suite *s = suite_create ("hlist");

        TCase *tc_core = tcase_create ("Core");
        tcase_add_checked_fixture (tc_core, setup, teardown);
        tcase_add_test (tc_core, test_queue_create);
        tcase_add_test (tc_core, test_enqueue);
        tcase_add_test (tc_core, test_insert_inbetween);
        suite_add_tcase (s, tc_core);

        TCase *tc_deq = tcase_create ("Queue");
        tcase_add_checked_fixture (tc_deq, setup_queue, teardown_queue);
        tcase_add_test (tc_deq, test_dequeue);
        suite_add_tcase (s, tc_deq);

        return s;
}

int test_hlist (void)
{
        int number_failed;
        Suite *s = hlist_suite ();
        SRunner *sr = srunner_create (s);
        srunner_set_log (sr, TEST_LOG_HQUEUE);
        srunner_run_all (sr, CK_VERBOSE);

        number_failed = srunner_ntests_failed (sr);
        srunner_free (sr);
        return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

