#include <stdlib.h>
#include <stdio.h>
#include <check.h>

#include <hmqueue.h>
#include <hlog.h>
#include "test_hwos.h"

static void create_enqueued_hmqueues();
static void setup (void);
static void teardown (void);
static Suite* hmqueue_suite (void);
void remove_enqueued_hmqueues (void);


//! Enqueued message queues. First enqueued is on index 0. Next is on index 1 and so on. Last enqueued is the last index.
static struct hmqueue* enqueued_hmqueues[TEST_HMQUEUE_SIZE];


static void setup(void)
{
        create_enqueued_hmqueues();
}

void create_enqueued_hmqueues (void)
{
        int i;
        // Setup all queues. Give them key i.
        for (i=0; i < TEST_HMQUEUE_SIZE; i++)
                enqueued_hmqueues[i] = hmqueue_add(i, HMQ_CREATE);
}

void remove_enqueued_hmqueues (void)
{
        int i;
        // Remove all queues from memory.
        for (i=0; i < TEST_HMQUEUE_SIZE; i++) {
                hmqueue_remove(enqueued_hmqueues[i]);
                enqueued_hmqueues[i] = NULL;
        }
}


static void teardown (void)
{
        remove_enqueued_hmqueues();
}


START_TEST (test_queue_create)
{
        fail_unless (hmqueue_size() == TEST_HMQUEUE_SIZE, "Queue size is not correct!");
        fail_unless (hmqueue_get_first() == enqueued_hmqueues[0], "First failure!");
        fail_unless (hmqueue_get_last() == enqueued_hmqueues[TEST_HMQUEUE_SIZE - 1], "Last failure!");
}
END_TEST


START_TEST (test_enqueue)
{
        int i;

        fail_unless (hmqueue_size() == 0, "Incorrect queue size!");
        fail_unless (hmqueue_get_first() == NULL, "First failure!");
        fail_unless (hmqueue_get_last() == NULL, "First failure!");

        // i goes from 0 to TEST_HMQUEUE_SIZE - 1.
        for (i = 0; i < TEST_HMQUEUE_SIZE; i++)
        {
                // Enqueue one more queue.
                enqueued_hmqueues[i] = hmqueue_add(i, HMQ_CREATE);

                fail_unless (hmqueue_size() == i + 1, "Incorrect queue size!");

                fail_unless (hmqueue_get_first() == enqueued_hmqueues[0], "First failure! i=%d", i);
                fail_unless (hmqueue_get_last() == enqueued_hmqueues[i], "Last failure!");
                fail_unless (hmqueue_get_next(enqueued_hmqueues[i]) == NULL, "Next field for last queue should be NULL!");
                fail_unless (hmqueue_get_prev(enqueued_hmqueues[0]) == NULL, "Prev field for first queue should be NULL!");
                if (i > 0) {
                        fail_unless (hmqueue_get_prev(enqueued_hmqueues[i]) == enqueued_hmqueues[i - 1], "Prev field failure!");
                        fail_unless (hmqueue_get_next(enqueued_hmqueues[i - 1]) == enqueued_hmqueues[i], "Next field failure!");
                }
                else if (i == 0)
                        fail_unless (hmqueue_get_prev(enqueued_hmqueues[i]) == NULL, "Prev field failure!");
        }
}
END_TEST


START_TEST (test_dequeue)
{
        int i, size_now;

        // i goes from 0 to TEST_HMQUEUE_SIZE - 1.
        for (i = 0; i < TEST_HMQUEUE_SIZE; i++)
        {
                // size_now goes from TEST_HMQUEUE_SIZE to 1.
                size_now = TEST_HMQUEUE_SIZE - i;

                fail_unless (hmqueue_size() == size_now, "Incorrect queue size!");

                fail_unless (hmqueue_get_first() == enqueued_hmqueues[i], "First failure!");
                fail_unless (hmqueue_get_last() == enqueued_hmqueues[TEST_HSQUEUE_SIZE - 1], "Last failure!");
                fail_unless (hmqueue_get_prev(enqueued_hmqueues[i]) == NULL, "Prev field for first queue should be NULL! i=%d", i);
                fail_unless (hmqueue_get_next(enqueued_hmqueues[TEST_HMQUEUE_SIZE - 1]) == NULL, "Next field for last queue should be NULL!");
              
                if (i < TEST_HMQUEUE_SIZE - 1) {
                        fail_unless (hmqueue_get_next(enqueued_hmqueues[i]) == enqueued_hmqueues[i + 1], "Next process in queue failure!");
                        fail_unless (hmqueue_get_prev(enqueued_hmqueues[i + 1]) == enqueued_hmqueues[i], "Prev process in queue failure!");
                }

                // Remove one more queue.
                hmqueue_remove(enqueued_hmqueues[i]);
        }

        // All hmqueues have been removed.
        fail_unless (hmqueue_size() == 0, "Incorrect queue size!");
        fail_unless (hmqueue_get_first() == NULL, "First failure!");
        fail_unless (hmqueue_get_last() == NULL, "Last failure!");
}
END_TEST



static Suite* hmqueue_suite (void)
{
        Suite *s = suite_create ("hmqueue");

        TCase *tc_core = tcase_create ("Core");
        tcase_add_checked_fixture (tc_core, setup, teardown);
        tcase_add_test (tc_core, test_queue_create);
        suite_add_tcase (s, tc_core);

        TCase *tc_enq = tcase_create ("Enqueue");
        tcase_add_checked_fixture (tc_enq, NULL, teardown);
        tcase_add_test (tc_enq, test_enqueue);
        suite_add_tcase (s, tc_enq);

        TCase *tc_deq  = tcase_create ("Dequeue");
        tcase_add_checked_fixture (tc_deq, setup, NULL);
        tcase_add_test (tc_deq, test_dequeue);
        suite_add_tcase (s, tc_deq);
        

        return s;
}

int test_hmqueue (void)
{
        int number_failed;
        Suite *s = hmqueue_suite ();
        SRunner *sr = srunner_create (s);
        srunner_set_log (sr, TEST_LOG_HMQUEUE);
        srunner_run_all (sr, CK_VERBOSE);

        number_failed = srunner_ntests_failed (sr);
        srunner_free (sr);
        return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

