#include <stdlib.h>
#include <stdio.h>
#include <check.h>

#include <hsqlist.h>
#include <hsqueue.h>
#include <hprocess.h>
#include "test_hwos.h"


static void setup (void);
static void setup_list (void);
static void teardown (void);
static void teardown_list(void);
static void create_listed_queues (void);
static void remove_listed_queues (void);


static int list0 = HPS_NEW;

//! Listed processes. First added is on index 0. Next is on index 1 and so on. Last added is the last index.
static struct hsqueue* listed_queues[TEST_HSQLIST_SIZE];


static void setup(void)
{
        create_listed_queues();
        hsqlist_create(list0);
}

void create_listed_queues (void)
{
        int i;
        // Setup all queues. Give them nice value i.
        for (i=0; i < TEST_HSQLIST_SIZE; i++)
                listed_queues[i] = hsqueue_create();
}

static void remove_listed_queues (void)
{
        int i;
        // Remove all queues from memory.
        for (i=0; i < TEST_HSQLIST_SIZE; i++) {
                hsqueue_remove(listed_queues[i]);
                listed_queues[i] = NULL;
        }
}


static void setup_list (void)
{
        setup();
        int i;
        // Add all queues to list.
        for (i=0; i < TEST_HSQLIST_SIZE; i++)
                hsqlist_add_queue(list0, listed_queues[i]);
}


static void teardown (void)
{
        remove_listed_queues();
        teardown_list();
}


static void teardown_list(void)
{
        hsqlist_remove(list0);
        list0 = -1;
}


START_TEST (test_list_create)
{
        fail_unless (hsqlist_size(list0) == 0, "List size is not 0!");
        fail_unless (hsqlist_get_first_queue(list0) == NULL, "First queue failure!");
        fail_unless (hsqlist_get_last_queue(list0) == NULL, "Last queue failure!");
}
END_TEST


START_TEST (test_add_queues)
{
        int i;

        fail_unless (hsqlist_size(list0) == 0, "Incorrect list size!");
        fail_unless (hsqlist_get_first_queue(list0) == NULL, "First queue failure!");
        fail_unless (hsqlist_get_last_queue(list0) == NULL, "First queue failure!");

        // i goes from 1 to TEST_HSQLIST_SIZE - 1.
        for (i = 0; i < TEST_HSQLIST_SIZE; i++)
        {
                // Add one more queue.
                hsqlist_add_queue(list0, listed_queues[i]);

                fail_unless (hsqlist_size(list0) == i + 1, "Incorrect list size!");

                fail_unless (hsqlist_get_first_queue(list0) == listed_queues[0], "First queue failure!");
                fail_unless (hsqlist_get_last_queue(list0) == listed_queues[i], "Last queue failure!");
                fail_unless (hsqueue_get_next(listed_queues[i]) == NULL, "Next field for last queue should be NULL!");
                fail_unless (hsqueue_get_prev(listed_queues[0]) == NULL, "Prev field for first queue should be NULL! i=%d.", i);
                if (i > 0) {
                        fail_unless (hsqueue_get_prev(listed_queues[i]) == listed_queues[i - 1], "Prev field failure!");
                        fail_unless (hsqueue_get_next(listed_queues[i - 1]) == listed_queues[i], "Next field failure!");
                }
                else if (i == 0)
                        fail_unless (hsqueue_get_prev(listed_queues[i]) == NULL, "Prev field failure!");
        }
}
END_TEST


START_TEST (test_insert_queues_before)
{
        int i;

        fail_unless (hsqlist_size(list0) == 0, "Incorrect list size!");
        fail_unless (hsqlist_get_first_queue(list0) == NULL, "First queue failure!");
        fail_unless (hsqlist_get_last_queue(list0) == NULL, "First queue failure!");

        // i goes from 1 to TEST_HSQLIST_SIZE - 1.
        for (i = 0; i < TEST_HSQLIST_SIZE; i++)
        {
                // Add one more queue.
                hsqlist_insert_queue_before(list0, hsqlist_get_first_queue(list0), listed_queues[i]);

                fail_unless (hsqlist_size(list0) == i + 1, "Incorrect list size!");

                fail_unless (hsqlist_get_first_queue(list0) == listed_queues[i], "First queue failure!");
                fail_unless (hsqlist_get_last_queue(list0) == listed_queues[0], "Last queue failure!");
                fail_unless (hsqueue_get_next(listed_queues[0]) == NULL, "Next field for last queue should be NULL!");
                fail_unless (hsqueue_get_prev(listed_queues[i]) == NULL, "Prev field for first queue should be NULL! i=%d.", i);
                if (i > 0) {
                        fail_unless (hsqueue_get_next(listed_queues[i]) == listed_queues[i - 1], "Next field failure! i=%d", i);
                        fail_unless (hsqueue_get_prev(listed_queues[i - 1]) == listed_queues[i], "Prev field failure!");
                }
                else if (i == 0)
                        fail_unless (hsqueue_get_next(listed_queues[i]) == NULL, "Next field failure!");
        }
}
END_TEST


START_TEST (test_remove_queues)
{
        int i, size_now;

        // i goes from 0 to TEST_HSQLIST_SIZE - 1.
        for (i = 0; i < TEST_HSQLIST_SIZE; i++)
        {
                // size_now goes from TEST_HSQLIST_SIZE to 1.
                size_now = TEST_HSQLIST_SIZE - i;

                fail_unless (hsqlist_size(list0) == size_now, "Incorrect list size!");

                fail_unless (hsqlist_get_first_queue(list0) == listed_queues[i], "First queue failure!");
                fail_unless (hsqlist_get_last_queue(list0) == listed_queues[TEST_HSQLIST_SIZE - 1], "Last queue failure! i=%d.", i);
                fail_unless (hsqueue_get_prev(listed_queues[i]) == NULL, "Prev field for first queue should be NULL!");
                fail_unless (hsqueue_get_next(listed_queues[TEST_HSQLIST_SIZE - 1]) == NULL, "Next field for last queue should be NULL!");
              

                if (i < TEST_HSQLIST_SIZE - 1) {
                        fail_unless (hsqueue_get_next(listed_queues[i]) == listed_queues[i + 1], "Next queue failure!");
                        fail_unless (hsqueue_get_prev(listed_queues[i + 1]) == listed_queues[i], "Prev queue failure!");
                }

                // Remove one more queue.
                hsqlist_remove_queue(list0, listed_queues[i]);
        }

        // All queuees have been removed.
        fail_unless (hsqlist_size(list0) == 0, "Incorrect queue size!");
        fail_unless (hsqlist_get_first_queue(list0) == NULL, "First queue failure!");
        fail_unless (hsqlist_get_last_queue(list0) == NULL, "Last queue failure!");
}
END_TEST




static Suite* hsqlist_suite (void)
{
        Suite *s = suite_create ("hsqlist");

        TCase *tc_core = tcase_create ("Core");
        tcase_add_checked_fixture (tc_core, setup, teardown);
        tcase_add_test (tc_core, test_list_create);
        tcase_add_test (tc_core, test_add_queues);
        tcase_add_test (tc_core, test_insert_queues_before);
        suite_add_tcase (s, tc_core);

        TCase *tc_deq = tcase_create ("Remove");
        tcase_add_checked_fixture (tc_deq, setup_list, teardown_list);
        tcase_add_test (tc_deq, test_remove_queues);
        suite_add_tcase (s, tc_deq);

        return s;
}

int test_hsqlist (void)
{
        int number_failed;
        Suite *s = hsqlist_suite ();
        SRunner *sr = srunner_create (s);
        srunner_set_log (sr, TEST_LOG_HSQLIST);
        srunner_run_all (sr, CK_VERBOSE);

        number_failed = srunner_ntests_failed (sr);
        srunner_free (sr);
        return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}



