#include <stdlib.h>
#include <stdio.h>
#include <check.h>

//! Queue interface.
#include <hsqueue.h>
#include <hprocess.h>
#include <hlog.h>
#include "test_hwos.h"


static struct hsqueue* queue0 = NULL;
static struct hprocess* process0 = NULL;

static void create_enqueued_processes (void);
static void remove_enqueued_processes (void);
static void setup (void);
static void setup_queue (void);
static void teardown (void);
static void teardown_queue(void);
static Suite* hsqueue_suite (void);


//! Enqueued processes. First enqueued is on index 0. Next is on index 1 and so on. Last enqueued is the last index.
static struct hprocess* enqueued_processes[TEST_HSQUEUE_SIZE];


static void setup(void)
{
        create_enqueued_processes();
        queue0 = hsqueue_create();
}

static void create_enqueued_processes (void)
{
        int i;
        // Setup all processes. Give them nice value i.
        for (i=0; i < TEST_HSQUEUE_SIZE; i++)
                enqueued_processes[i] = hprocess_create(i);
        process0 = hprocess_create(10);
}

static void remove_enqueued_processes (void)
{
        int i;
        // Remove all processes from memory.
        for (i=0; i < TEST_HSQUEUE_SIZE; i++) {
                hsqueue_remove_process(queue0, enqueued_processes[i]);
                enqueued_processes[i] = NULL;
        }
        hsqueue_remove_process(queue0, process0);
        process0 = NULL;
}


static void setup_queue (void)
{
        setup();
        int i;
        // Add all processes to queue.
        for (i=0; i < TEST_HSQUEUE_SIZE; i++)
                hsqueue_enqueue_process(queue0, enqueued_processes[i]);
}


static void teardown (void)
{
        remove_enqueued_processes();
        teardown_queue();
}

static void teardown_queue(void)
{
        hsqueue_remove(queue0);
        queue0 = NULL;
}

START_TEST (test_queue_create)
{
        fail_unless (queue0 != NULL, "Queue is NULL!");
        fail_unless (hsqueue_size(queue0) == 0, "Queue size is not 0!");
        fail_unless (hsqueue_get_first_process(queue0) == NULL, "First process failure!");
        fail_unless (hsqueue_get_last_process(queue0) == NULL, "Last process failure!");
}
END_TEST


START_TEST (test_enqueue)
{
        int i;

        fail_unless (hsqueue_size(queue0) == 0, "Incorrect queue size!");
        fail_unless (hsqueue_get_first_process(queue0) == NULL, "First process failure!");
        fail_unless (hsqueue_get_last_process(queue0) == NULL, "First process failure!");

        // i goes from 1 to TEST_HSQUEUE_SIZE - 1.
        for (i = 0; i < TEST_HSQUEUE_SIZE; i++)
        {
                // Enqueue one more process.
                hsqueue_enqueue_process(queue0, enqueued_processes[i]);

                fail_unless (hsqueue_size(queue0) == i + 1, "Incorrect queue size!");

                fail_unless (hsqueue_get_first_process(queue0) == enqueued_processes[0], "First process failure!");
                fail_unless (hsqueue_get_last_process(queue0) == enqueued_processes[i], "Last process failure!");
                fail_unless (hprocess_get_next(enqueued_processes[i]) == NULL, "Next field for last process should be NULL!");
                fail_unless (hprocess_get_prev(enqueued_processes[0]) == NULL, "Prev field for first process should be NULL!");
                if (i > 0) {
                        fail_unless (hprocess_get_prev(enqueued_processes[i]) == enqueued_processes[i - 1], "Prev field failure!");
                        fail_unless (hprocess_get_next(enqueued_processes[i - 1]) == enqueued_processes[i], "Next field failure!");
                }
                else if (i == 0)
                        fail_unless (hprocess_get_prev(enqueued_processes[i]) == NULL, "Prev field failure!");
        }
}
END_TEST


START_TEST (test_dequeue)
{
        int i, size_now;

        // i goes from 0 to TEST_HSQUEUE_SIZE - 1.
        for (i = 0; i < TEST_HSQUEUE_SIZE; i++)
        {
                // size_now goes from TEST_HSQUEUE_SIZE to 1.
                size_now = TEST_HSQUEUE_SIZE - i;

                fail_unless (hsqueue_size(queue0) == size_now, "Incorrect queue size!");

                fail_unless (hsqueue_get_first_process(queue0) == enqueued_processes[i], "First process failure!");
                fail_unless (hsqueue_get_last_process(queue0) == enqueued_processes[TEST_HSQUEUE_SIZE - 1], "Last process failure!");
                fail_unless (hprocess_get_prev(enqueued_processes[i]) == NULL, "Prev field for first process should be NULL! i=%d", i);
                fail_unless (hprocess_get_next(enqueued_processes[TEST_HSQUEUE_SIZE - 1]) == NULL, "Next field for last process should be NULL!");
              
                if (i < TEST_HSQUEUE_SIZE - 1) {
                        fail_unless (hprocess_get_next(enqueued_processes[i]) == enqueued_processes[i + 1], "Next process in queue failure!");
                        fail_unless (hprocess_get_prev(enqueued_processes[i + 1]) == enqueued_processes[i], "Prev process in queue failure!");
                }

                // Remove one more process.
                hsqueue_dequeue_process(queue0);
        }

        // All processes have been removed.
        fail_unless (hsqueue_size(queue0) == 0, "Incorrect queue size!");
        fail_unless (hsqueue_get_first_process(queue0) == NULL, "First process failure!");
        fail_unless (hsqueue_get_last_process(queue0) == NULL, "Last process failure!");
}
END_TEST




static Suite* hsqueue_suite (void)
{
        Suite *s = suite_create ("hsqueue");

        TCase *tc_core = tcase_create ("Core");
        tcase_add_checked_fixture (tc_core, setup, teardown);
        tcase_add_test (tc_core, test_queue_create);
        tcase_add_test (tc_core, test_enqueue);
        suite_add_tcase (s, tc_core);

        TCase *tc_deq = tcase_create ("Queue");
        tcase_add_checked_fixture (tc_deq, setup_queue, teardown_queue);
        tcase_add_test (tc_deq, test_dequeue);
        suite_add_tcase (s, tc_deq);

        return s;
}

int test_hsqueue (void)
{
        int number_failed;
        Suite *s = hsqueue_suite ();
        SRunner *sr = srunner_create (s);
        srunner_set_log (sr, TEST_LOG_HSQUEUE);
        srunner_run_all (sr, CK_VERBOSE);

        number_failed = srunner_ntests_failed (sr);
        srunner_free (sr);
        return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

