-------------------------------------------------------------------------------
--
-- Title       : BRAM
-- Design      : BRAM
-- Author      : Vegard Endresen
-- Company     : NTNU
-- Description : 
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity bram_packed is
 generic
  (
    -- ADD USER GENERICS BELOW THIS LINE ---------------
    --USER generics added here
    -- ADD USER GENERICS ABOVE THIS LINE ---------------

    -- DO NOT EDIT BELOW THIS LINE ---------------------
    -- Bus protocol parameters, do not add to or delete
    C_SLV_AWIDTH                   : integer              := 32;
    C_SLV_DWIDTH                   : integer              := 32;
    C_NUM_MEM                      : integer              := 1
    -- DO NOT EDIT ABOVE THIS LINE ---------------------
  );
  port
  (
    -- ADD USER PORTS BELOW THIS LINE ------------------
	 hw_clk		: in std_logic;
	 hw_en		: in std_logic;
	 hw_we   	: in std_logic;
	 hw_addr		: in std_logic_vector(0 to 7);
	 hw_read_ack: out std_logic;
	 hw_write_ack: out std_logic; 
	 hw_data_i 	: in std_logic_vector(0 to C_SLV_DWIDTH-1);
	 hw_data_o 	: out std_logic_vector(0 to C_SLV_DWIDTH-1);
	 -- ADD USER PORTS ABOVE THIS LINE ------------------

    -- DO NOT EDIT BELOW THIS LINE ---------------------
    -- Bus protocol ports, do not add to or delete
    Bus2IP_Clk                     : in  std_logic;
    Bus2IP_Reset                   : in  std_logic;
    Bus2IP_Addr                    : in  std_logic_vector(0 to C_SLV_AWIDTH-1);
    Bus2IP_CS                      : in  std_logic_vector(0 to C_NUM_MEM-1);
    Bus2IP_RNW                     : in  std_logic;
    Bus2IP_Data                    : in  std_logic_vector(0 to C_SLV_DWIDTH-1);
    Bus2IP_BE                      : in  std_logic_vector(0 to C_SLV_DWIDTH/8-1);
    IP2Bus_Data                    : out std_logic_vector(0 to C_SLV_DWIDTH-1);
    IP2Bus_RdAck                   : out std_logic;
    IP2Bus_WrAck                   : out std_logic;
    IP2Bus_Error                   : out std_logic
    -- DO NOT EDIT ABOVE THIS LINE ---------------------
  );

  attribute SIGIS : string;
  attribute SIGIS of Bus2IP_Clk    : signal is "CLK";
  attribute SIGIS of Bus2IP_Reset  : signal is "RST";
  
end entity bram_packed;

architecture IMP of bram_packed is

  ------------------------------------------
  -- Signals for user logic memory space example
  ------------------------------------------
  type WORD_RAM_TYPE is array (0 to 255) of std_logic_vector(0 to 31);
  type DO_TYPE is array (0 to C_NUM_MEM-1) of std_logic_vector(0 to C_SLV_DWIDTH-1);
  
  
  signal mem_data_out                   : DO_TYPE;
  signal mem_address                    : std_logic_vector(0 to 7);
  signal mem_select                     : std_logic_vector(0 to 0);
  signal mem_read_enable                : std_logic;
  signal mem_read_enable_dly1				 : std_logic;
  signal mem_read_req                   : std_logic;
  signal mem_ip2bus_data                : std_logic_vector(0 to C_SLV_DWIDTH-1);
  signal mem_read_ack_dly1              : std_logic;
  signal mem_read_ack                   : std_logic;
  signal mem_write_ack                  : std_logic;

  
 shared variable ram    : WORD_RAM_TYPE;
 
 
 type  STATE is (IDLE,DELAYED);
signal cstate,nstate : STATE;

 
 
signal write_enable  : std_logic;
 

begin

  --USER logic implementation added here

  ------------------------------------------
  -- Example code to access user logic memory region
  -- 
  -- Note:
  -- The example code presented here is to show you one way of using
  -- the user logic memory space features. The Bus2IP_Addr, Bus2IP_CS,
  -- and Bus2IP_RNW IPIC signals are dedicated to these user logic
  -- memory spaces. Each user logic memory space has its own address
  -- range and is allocated one bit on the Bus2IP_CS signal to indicated
  -- selection of that memory space. Typically these user logic memory
  -- spaces are used to implement memory controller type cores, but it
  -- can also be used in cores that need to access additional address space
  -- (non C_BASEADDR based), s.t. bridges. This code snippet infers
  -- 1 256x32-bit (byte accessible) single-port Block RAM by XST.
  ------------------------------------------
  mem_select      <= Bus2IP_CS;
  mem_read_enable <= (( Bus2IP_CS(0) ) and Bus2IP_RNW);
  mem_read_ack    <= mem_read_ack_dly1;
  mem_write_ack   <= ( Bus2IP_CS(0) ) and not(Bus2IP_RNW);
  mem_address     <= Bus2IP_Addr(C_SLV_AWIDTH-10 to C_SLV_AWIDTH-3);
  mem_read_req    <= mem_read_enable and not mem_read_enable_dly1;
  

  -- this process generates the read acknowledge 1 clock after read enable
  -- is presented to the BRAM block. The BRAM block has a 1 clock delay
  -- from read enable to data out.
  
  -- MODIFIED to have no delay due to dual port trouble
  
  BRAM_READ_ENABLE_PROC : process(Bus2IP_Reset,Bus2IP_Clk) is
  begin
	  if(Bus2IP_Reset = '1') then
			mem_read_enable_dly1 <='0';
		elsif Bus2IP_Clk'event and Bus2IP_Clk ='1' then
			mem_read_enable_dly1 <= mem_read_enable;
	  end if;
  end process BRAM_READ_ENABLE_PROC;
  
  
  BRAM_RD_ACK_PROC : process(Bus2IP_Reset,Bus2IP_Clk) is
  begin
      if ( Bus2IP_Reset = '1' ) then
        mem_read_ack_dly1 <= '0';
      elsif Bus2IP_Clk'event and Bus2IP_Clk ='1' then
        mem_read_ack_dly1 <= mem_read_req;
		end if;
  end process BRAM_RD_ACK_PROC;
  
  
  
    NSTATE_PROC : process(cstate,hw_en,hw_we) is
	begin
		case cstate is
			when IDLE => 
				if hw_en ='1' and not hw_we='1' then
					nstate <= DELAYED;
				else
					nstate <= IDLE;
				end if;			   
			when DELAYED =>
				nstate <= IDLE;
		end case;
	end process NSTATE_PROC;   
	
	CSTATE_PROC : process(Bus2IP_Clk) is
	begin
		if Bus2IP_Clk'event and Bus2IP_Clk='1' then
			cstate <= nstate;
		end if;
	end process CSTATE_PROC;
	
	
	READ_ACK_DLY_PROC : process(cstate,hw_en) is
	begin
		if cstate = DELAYED then
			hw_read_ack <= hw_en;
		else
			hw_read_ack <='0';
		end if;
	end process READ_ACK_DLY_PROC;	
	
	hw_write_ack <= hw_en and hw_we;
		
		

      write_enable <= not( Bus2IP_RNW) and
                       Bus2IP_cs(0);


		---------------------------------------------------------------------
		-- BUS process

      BUS_PROC : process( Bus2IP_Clk ) is
      begin
        if ( Bus2IP_Clk'event and Bus2IP_Clk = '1' ) then
          -- WRITING TO RAM
			 if( Bus2IP_cs(0)='1') then
				if ( write_enable = '1' ) then
            	ram(CONV_INTEGER(mem_address)) :=  Bus2IP_Data;
				end if;
					mem_data_out(0) <= ram(CONV_INTEGER(mem_address));
			end if;
		 end if;
		end process BUS_PROC;
      ---------------------------------------------------------------------

		---------------------------------------------------------------------
		-- Dual port extension
		process (hw_clk)
		begin
			if hw_clk'event and hw_clk = '1' then
				if hw_en = '1' then
					if hw_we = '1' then
						ram(conv_integer(hw_addr)) := hw_data_i;
					end if;
						hw_data_o <= ram(conv_integer(hw_addr));
				end if;
			end if;
		end process;
		---------------------------------------------------------------------


  -- implement Block RAM read mux
  MEM_IP2BUS_DATA_PROC : process( mem_data_out, mem_select ) is
  begin

    case mem_select is
      when "1" => mem_ip2bus_data <= mem_data_out(0);
      when others => mem_ip2bus_data <= (others => '0');
    end case;
  end process MEM_IP2BUS_DATA_PROC;
  
  ------------------------------------------
  -- Example code to drive IP to Bus signals
  ------------------------------------------
  IP2Bus_Data  <= mem_ip2bus_data when mem_read_ack = '1' else (others => '0');
  IP2Bus_WrAck <= mem_write_ack;
  IP2Bus_RdAck <= mem_read_ack;
  IP2Bus_Error <= '0';
    
end IMP;
