-------------------------------------------------------------------------------
--
-- Title       : instr_controller
-- Design      : cpu_interface
-- Author      : Vegard
-- Company     : NTNU
-- Description : 
-------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;

entity instr_controller is						  
	generic(			
		-- Instruction size, modification requires redefining instructions
		INSTR_SIZE : natural :=16
	);
	port(
		clk				: in std_logic;
		instruction 	: in std_logic_vector(INSTR_SIZE-1 downto 0);
		read_ack 		: in std_logic;	
		write_ack		: in std_logic;
		hw_complete		: in std_logic; 
		next_instr		: in std_logic;
		cs				: in std_logic_vector(0 to 0);
		address			: out std_logic_vector(7 downto 0);
		shift			: out std_logic;
		register_write	: out std_logic;
		hw_en			: out std_logic;
		hw_we			: out std_logic;
		shift_mask		: out std_logic;
		zero_mask		: out std_logic;
		hw_run			: inout std_logic;
		halt			: inout std_logic;
		instr_complete  : inout std_logic
	    );

end instr_controller;


architecture instr_controller of instr_controller is

type state is (IDLE,SHIFTING,RUNNING,WAITING);
signal cstate,nstate : STATE;	

signal shift_i  : std_logic; 
signal shift_ii : std_logic;
signal count 	: std_logic_vector(1 downto 0);
signal shift_complete : std_logic;
signal hw_run_i : std_logic;
begin	 
	
	
	NSTATE_PROC : process(shift_i,hw_run,hw_complete,shift_complete,next_instr,cstate) is
	begin
		case cstate is 
			when IDLE =>	 
				if shift_i ='1' then 
					nstate <= SHIFTING;
				elsif hw_run ='1' then
					nstate <= RUNNING;
				else
					nstate <= IDLE;
				end if;
			when SHIFTING	=>
				if shift_complete ='1' then
					nstate <= WAITING;
				else
					nstate <= SHIFTING;
				end if;	  			 
			when RUNNING =>
				if hw_complete ='1' then
					nstate <= WAITING;
				else
					nstate <= RUNNING;
				end if;
			when WAITING =>		
				if next_instr ='1' then
					nstate <= IDLE;	 
				else
					nstate <= WAITING;
				end if;
		end case;
	end process NSTATE_PROC;
	
	CSTATE_PROC : process(clk) is
	begin
		if clk'event and clk='1' then
			cstate <= nstate;	
		end if;
	end process CSTATE_PROC;
	
	
	SHIFT_PROC : process(cstate,shift_i) is
	begin							   
		if cstate = IDLE or cstate = SHIFTING then
			shift_ii <= shift_i;
		else
			shift_ii <= '0';
		end if;
	end process SHIFT_PROC;

	
	
	INSTRUCTION_DECODER : entity instr_decoder
		generic map(
			INSTR_SIZE => INSTR_SIZE
		)
		port map(
			opcode => instruction(INSTR_SIZE-1 downto INSTR_SIZE-2),
			operand => instruction(INSTR_SIZE-3 downto 0),
			read_ack => read_ack,
			write_ack => write_ack,
			shift_complete => shift_complete,
			hw_complete => hw_complete,
			address => address,
			cs => cs,
			register_write => register_write,
			hw_en => hw_en,
			hw_we => hw_we, 
			count => count,
			shift_mask => shift_mask,
			zero_mask => zero_mask,
			hw_run  => hw_run_i,
			halt => halt,
			shift_o => shift_i,
			instr_complete => instr_complete
		);
	
		SHIFT_CONTROLLER : entity shift_alt
		generic map(C_COUNT_SIZE => 2)
		port map(
			clk 			=> clk,	 
			shift_i			=> shift_ii,
			count			=> count,
			shift_o			=> shift,
			shift_complete	=> shift_complete
		); 
	
		HW_RUN_PROC : process(cstate,halt,hw_run_i) 
		begin
			if halt ='1' or cstate = WAITING then
				hw_run <='0';
			else
				hw_run <= hw_run_i;
			end if;
		end process HW_RUN_PROC;
		

end instr_controller;
