-- Top-level entity for design:
-- Two DFFs. One for initiliazing the clock signal to the module and one
-- for occupying the lowest clock wire.
--
-- Sindre Hansen (2011).

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

library unisim;
use unisim.vcomponents.all; 

entity reconf is
		port (
	 clk_in	  				: in   std_logic;
	 in_16    				: in   std_logic_vector(0 to 15);
	 dummy_input			: in   std_logic;
	 dummy_reset			: in   std_logic;
	 out_16   				: out  std_logic_vector(0 to 15);
	 dummy_output			: out   std_logic
	);
end entity reconf;

------------------------------------------------------------------------------
-- Architecture section
------------------------------------------------------------------------------

architecture IMP of reconf is

signal bm_l2r_in : std_logic_vector(0 to 15);
signal bm_l2r_out : std_logic_vector(0 to 15);

signal bm_r2l_in : std_logic_vector(0 to 15);
signal bm_r2l_out : std_logic_vector(0 to 15);

signal clk_buffer_out : std_logic;
signal clk_buffer_out_dummy : std_logic;
signal clk_buffer_ce : std_logic;
signal dummy_register_bm : std_logic;
signal dummy_output_bm : std_logic;
signal dummy_input_bm : std_logic;
signal dummy_reset_bm : std_logic;
 ------------------------------------------------
 --Bus macros
 ------------------------------------------------
component busmacro_xc4v_l2r_async_wide is
    port (
		input0 : in std_logic;
		input1 : in std_logic;
		input2 : in std_logic;
		input3 : in std_logic;
		input4 : in std_logic;
		input5 : in std_logic;
		input6 : in std_logic;
		input7 : in std_logic;
		output0 : out std_logic;
		output1 : out std_logic;
		output2 : out std_logic;
		output3 : out std_logic;
		output4 : out std_logic;
		output5 : out std_logic;
		output6 : out std_logic;
		output7 : out std_logic
    );
	 end component;
	 
component busmacro_xc4v_r2l_async_wide is
    port (
		input0 : in std_logic;
		input1 : in std_logic;
		input2 : in std_logic;
		input3 : in std_logic;
		input4 : in std_logic;
		input5 : in std_logic;
		input6 : in std_logic;
		input7 : in std_logic;
		output0 : out std_logic;
		output1 : out std_logic;
		output2 : out std_logic;
		output3 : out std_logic;
		output4 : out std_logic;
		output5 : out std_logic;
		output6 : out std_logic;
		output7 : out std_logic
    );
end component;


begin

	-- Extra flip-flop to create reconfigurable clock interface.
	DUMMY_PROC : process(clk_buffer_out_dummy, dummy_reset_bm)
	begin
		if (dummy_reset_bm = '1') then
			dummy_register_bm <= '0';
		elsif (clk_buffer_out_dummy'EVENT and clk_buffer_out_dummy='1') then
			dummy_register_bm <= dummy_input_bm;
		end if;
	end process DUMMY_PROC;
	

	-- Connect the outputs & inputs to the bus macro outputs & inputs 
	out_16 <= bm_r2l_out;
	bm_l2r_in <= in_16;

	R : entity reconfigurable_module
    	port map
    	(
		clk_in => clk_buffer_out,
		in_16 => bm_l2r_out,
		out_16 => bm_r2l_in
	);

	-- Connect one of the outputs from the HWSW-register to clk_buffer_ce.
	-- This means the clock buffer's CE-input can be controlled by a Linux-app.
	clk_buffer_ce <= bm_l2r_in(2);

	-- Instantiate global clock buffer for reconfigurable module explicitely.
	BUFGCE_inst : BUFGCE
	port map (
						 O => clk_buffer_out,		-- Clock buffer output (connected to reconf. module)
						 CE => clk_buffer_ce,		-- Clock enable input
						 I => clk_in						-- Clock buffer input
					 );
	
	-- Instantiate global clock buffer for dummy register explicitely.
	BUFGCE_dummy : BUFGCE
	port map (
						 O => clk_buffer_out_dummy,		-- Clock buffer output (connected to reconf. module)
						 CE => clk_buffer_ce,		-- Clock enable input
						 I => clk_in						-- Clock buffer input
					 );
	
	bm_dummy : component busmacro_xc4v_l2r_async_wide
		port map(
				--
			input0 =>dummy_input,
			input1 =>dummy_reset,
			input2 =>dummy_register_bm,
			input3 =>bm_l2r_in(3),
			input4 =>bm_l2r_in(4),
			input5 =>bm_l2r_in(5),
			input6 =>bm_l2r_in(6),
			input7 =>bm_l2r_in(7),
			output0 =>dummy_input_bm,
			output1 =>dummy_reset_bm,
			output2 =>dummy_output,
			output3 =>open,
			output4 =>open,
			output5 =>open,
			output6 =>open,
			output7 =>open
			
		);


	bm_l2r_0 : component busmacro_xc4v_l2r_async_wide
		port map(
				--
			input0 =>bm_l2r_in(0),
			input1 =>bm_l2r_in(1),
			input2 =>bm_l2r_in(2),
			input3 =>bm_l2r_in(3),
			input4 =>bm_l2r_in(4),
			input5 =>bm_l2r_in(5),
			input6 =>bm_l2r_in(6),
			input7 =>bm_l2r_in(7),
			output0 =>bm_l2r_out(0),
			output1 =>bm_l2r_out(1),
			output2 =>bm_l2r_out(2),
			output3 =>bm_l2r_out(3),
			output4 =>bm_l2r_out(4),
			output5 =>bm_l2r_out(5),
			output6 =>bm_l2r_out(6),
			output7 =>bm_l2r_out(7)
			
		);
		
		bm_l2r_1 : component busmacro_xc4v_l2r_async_wide
		port map(
				
			input0 =>bm_l2r_in(8),
			input1 =>bm_l2r_in(9),
			input2 =>bm_l2r_in(10),
			input3 =>bm_l2r_in(11),
			input4 =>bm_l2r_in(12),
			input5 =>bm_l2r_in(13),
			input6 =>bm_l2r_in(14),
			input7 =>bm_l2r_in(15),
			output0 =>bm_l2r_out(8),
			output1 =>bm_l2r_out(9),
			output2 =>bm_l2r_out(10),
			output3 =>bm_l2r_out(11),
			output4 =>bm_l2r_out(12),
			output5 =>bm_l2r_out(13),
			output6 =>bm_l2r_out(14),
			output7 =>bm_l2r_out(15)
			
		);
		
		bm_r2l_0 : component busmacro_xc4v_r2l_async_wide
		port map(
				
			input0 =>bm_r2l_in(0),
			input1 =>bm_r2l_in(1),
			input2 =>bm_r2l_in(2),
			input3 =>bm_r2l_in(3),
			input4 =>bm_r2l_in(4),
			input5 =>bm_r2l_in(5),
			input6 =>bm_r2l_in(6),
			input7 =>bm_r2l_in(7),
			output0 =>bm_r2l_out(0),
			output1 =>bm_r2l_out(1),
			output2 =>bm_r2l_out(2),
			output3 =>bm_r2l_out(3),
			output4 =>bm_r2l_out(4),
			output5 =>bm_r2l_out(5),
			output6 =>bm_r2l_out(6),
			output7 =>bm_r2l_out(7)
			
		);
		
	
	
		bm_r2l_1 : component busmacro_xc4v_r2l_async_wide
		port map(
				
			input0 =>bm_r2l_in(8),
			input1 =>bm_r2l_in(9),
			input2 =>bm_r2l_in(10),
			input3 =>bm_r2l_in(11),
			input4 =>bm_r2l_in(12),
			input5 =>bm_r2l_in(13),
			input6 =>bm_r2l_in(14),
			input7 =>bm_r2l_in(15),
			output0 =>bm_r2l_out(8),
			output1 =>bm_r2l_out(9),
			output2 =>bm_r2l_out(10),
			output3 =>bm_r2l_out(11),
			output4 =>bm_r2l_out(12),
			output5 =>bm_r2l_out(13),
			output6 =>bm_r2l_out(14),
			output7 =>bm_r2l_out(15)
			
		);

end IMP;
