-------------------------------------------------------------------------------
--
-- Title       : bram_id_cache
-- Design      : cpu_interface
-- Author      : Vegard
-- Company     : NTNU
-- Description : 

-- Modified (2011) by: Sindre Hansen
-- Modifications:
-- * Isolated the multiplier from the rest of the base design
--   and made it ready for ISE.
-- * A BUFGCE has been set up for the clock signal.
-------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

library unisim;
use unisim.vcomponents.all;

entity bram_id_cache is
 generic(
	
  C_SLV_AWIDTH     : integer := 32;
  C_SLV_DWIDTH     : integer := 32;
  C_NUM_REG			 : integer  := 2;
  C_NUM_MEM        : integer := 1;
  SHIFT_SIZE		 : integer :=8;
  INSTR_SIZE		: natural := 16; 
  INSTR_ADDR_BITS : natural :=5
 
	);
 port(
	  Bus2IP_Clk   : in  std_logic;
	  Bus2IP_Reset : in std_logic;
	  Bus2IP_Addr  : in	std_logic_vector(0 to C_SLV_AWIDTH-1);
	  Bus2IP_CS    : in std_logic_vector(0 to C_NUM_MEM-1);
	  Bus2IP_RNW   : in std_logic;
	  Bus2IP_Data  : in std_logic_vector(0 to C_SLV_DWIDTH-1);
	  Bus2IP_BE    : in std_logic_vector(0 to C_SLV_DWIDTH/8-1);
	  Bus2IP_RdCE  : in std_logic_vector(0 to C_NUM_REG-1);
	  Bus2IP_WrCE  : in std_logic_vector(0 to C_NUM_REG-1);
	  IP2Bus_Data  : out std_logic_vector(0 to C_SLV_DWIDTH-1);
	  IP2Bus_RdAck : out std_logic;
	  IP2Bus_WrAck : out std_logic;
	  IP2Bus_Error : out std_logic;
	  hw_complete  : out std_logic;
	  shift_o      : out std_logic_vector(0 to 7)
	);

end bram_id_cache;


architecture bram_id_cache of bram_id_cache is
 
 ------------------------------------------------
 --Bus macros
 ------------------------------------------------
 component busmacro_xc4v_l2r_async_wide is
    port (
		input0 : in std_logic;
		input1 : in std_logic;
		input2 : in std_logic;
		input3 : in std_logic;
		input4 : in std_logic;
		input5 : in std_logic;
		input6 : in std_logic;
		input7 : in std_logic;
		output0 : out std_logic;
		output1 : out std_logic;
		output2 : out std_logic;
		output3 : out std_logic;
		output4 : out std_logic;
		output5 : out std_logic;
		output6 : out std_logic;
		output7 : out std_logic
    );
	 end component;
component busmacro_xc4v_r2l_async_wide is
    port (
		input0 : in std_logic;
		input1 : in std_logic;
		input2 : in std_logic;
		input3 : in std_logic;
		input4 : in std_logic;
		input5 : in std_logic;
		input6 : in std_logic;
		input7 : in std_logic;
		output0 : out std_logic;
		output1 : out std_logic;
		output2 : out std_logic;
		output3 : out std_logic;
		output4 : out std_logic;
		output5 : out std_logic;
		output6 : out std_logic;
		output7 : out std_logic
    );
end component;
 
-----------------------------------------------------------------
-- Bus macro outputs
-----------------------------------------------------------------
signal bm_o_shift_i 	: std_logic_vector(SHIFT_SIZE-1 downto 0);
signal bm_o_run			: std_logic; 
signal bm_o_addr		: std_logic;
signal bm_o_zerofill	: std_logic;
signal bm_o_shift_g		: std_logic;
signal bm_o_reset		: std_logic;

-----------------------------------------------------------------
-- Bus macro inputs
----------------------------------------------------------------- 
signal bm_i_shift_o 	: std_logic_vector(SHIFT_SIZE-1 downto 0);
signal bm_i_finish		: std_logic;  

---------------------------------------------------------------------
 -- HW - BRAM signals
---------------------------------------------------------------------
signal bram_hw_en			:  std_logic;
signal bram_hw_we   		:  std_logic;
signal bram_hw_addr			:  std_logic_vector(7 downto 0);
signal bram_hw_data_i 		:  std_logic_vector(C_SLV_DWIDTH-1 downto 0);
signal bram_hw_data_o 		:  std_logic_vector(C_SLV_DWIDTH-1 downto 0);
signal bram_hw_read_ack 	: std_logic;
signal bram_hw_write_ack 	: std_logic;

---------------------------------------------------------------------		
  
 ------------------------------------------
  -- Signals for user logic slave model s/w accessible register example
  ------------------------------------------

  signal slv_reg0                       : std_logic_vector(0 to C_SLV_DWIDTH-1); -- sync/control
  signal slv_reg1                       : std_logic_vector(0 to C_SLV_DWIDTH-1); -- reg0
  signal slv_reg2                       : std_logic_vector(0 to C_SLV_DWIDTH-1); --reg1
  signal slv_reg_write_sel              : std_logic_vector(0 to 1);
  signal slv_reg_read_sel               : std_logic_vector(0 to 1);
  signal slv_ip2bus_data                : std_logic_vector(0 to C_SLV_DWIDTH-1);
  signal slv_read_ack                   : std_logic;
  signal slv_write_ack                  : std_logic;
  
  

  signal execute : std_logic;
  
  ---------------------------------------------------------------------
  -- Shift controller signals
 
  signal shift			: std_logic;
  ---------------------------------------------------------------------		
  
  
  ---------------------------------------------------------------------
  -- Instruction controller signals
  
  
  signal id_en : std_logic;
  signal id_we : std_logic;
  signal id_read_ack : std_logic;
  
	signal instruction 		: std_logic_vector(INSTR_SIZE-1 downto 0);
	signal id_addr		    : std_logic_vector(7 downto 0);
	signal register_write	: std_logic;

	signal shift_mask		: std_logic;
	signal zero_mask		: std_logic;	
	signal instr_complete   : std_logic;
  ---------------------------------------------------------------------
  
  ---------------------------------------------------------------------
  -- Instruction fetcher signals
  
    signal if_en		  	:  std_logic;
  	signal if_read_ack 		:  std_logic;	
	signal if_instr_write	:  std_logic;
	signal if_instr_shift	:  std_logic;
	signal if_addr	:  std_logic_vector(INSTR_ADDR_BITS-1 downto 0)	;
  ---------------------------------------------------------------------
  
  
  
  ---------------------------------------------------------------------
  -- REG0 signals
  
  signal reg0_shift_o : std_logic_vector(0 to 7);
  signal reg0_data_o : std_logic_vector(31 downto 0);
  ---------------------------------------------------------------------
  
  
  ---------------------------------------------------------------------
  -- Multiplier signals
  signal hw_run : std_logic;  
  signal shift_i :std_logic_vector(7 downto 0);
  ---------------------------------------------------------------------
  
  	
  
 
signal zeros : std_logic_vector(31 downto 0); 
  
signal halt : std_logic;
  

signal mem_ip2bus_data : std_logic_vector(0 to 31);
signal mem_read_ack : std_logic;
signal mem_write_ack : std_logic;
signal mem_error : std_logic;


signal next_instr : std_logic;
												
signal instr_depth : std_logic_vector(7 downto 0); 

signal if_stall : std_logic;

--! Output from clock buffer.
signal clk_buffer_out : std_logic;
--! Clock Enable input to clock buffer.
signal clk_buffer_ce : std_logic := '1';

begin
	
		-- Data left to right
	bm_l2r_data: component busmacro_xc4v_l2r_async_wide
	port map(
			
		input0 => shift_i(0),
		input1 => shift_i(1),
		input2 => shift_i(2),
		input3 => shift_i(3),
		input4 => shift_i(4),
		input5 => shift_i(5),
		input6 => shift_i(6),
		input7 => shift_i(7),
		
		output0 =>  bm_o_shift_i(0),
		output1 =>  bm_o_shift_i(1),
		output2 =>  bm_o_shift_i(2),
		output3 =>  bm_o_shift_i(3),
		output4 =>  bm_o_shift_i(4),
		output5 =>  bm_o_shift_i(5),
		output6 =>  bm_o_shift_i(6),
		output7 =>  bm_o_shift_i(7)
	);										 
	

	-- Control signals left to right
	bm_l2r_ctrl: component busmacro_xc4v_l2r_async_wide
	port map(
			
		input0 => Bus2IP_Reset,
		input1 => hw_run,
		input2 => id_addr(0),
		input3 => zero_mask,
		input4 => shift,
		input5 => zeros(0),
		input6 => zeros(0),
		input7 => zeros(0),
		
		output0 =>  bm_o_reset,
		output1 =>  bm_o_run,
		output2 =>  bm_o_addr,
		output3 =>  bm_o_zerofill,
		output4 =>  bm_o_shift_g,
		output5 =>  open,
		output6 =>  open ,
		output7 =>  open
	);
	
	
	
	-- Data right to left
	bm_r2l_data: component busmacro_xc4v_r2l_async_wide
	port map(
			
		input0 => bm_i_shift_o(0),
		input1 => bm_i_shift_o(1),
		input2 => bm_i_shift_o(2),
		input3 => bm_i_shift_o(3),
		input4 => bm_i_shift_o(4),
		input5 => bm_i_shift_o(5),
		input6 => bm_i_shift_o(6),
		input7 => bm_i_shift_o(7),
		
		output0 =>  shift_o(0),
		output1 =>  shift_o(1),
		output2 =>  shift_o(2),
		output3 =>  shift_o(3),
		output4 =>  shift_o(4),
		output5 =>  shift_o(5),
		output6 =>  shift_o(6),
		output7 =>  shift_o(7)
	);		
	
	-- Control signals right to left
	bm_r2l_ctrl: component busmacro_xc4v_r2l_async_wide
	port map(
			
		input0 => bm_i_finish,
		input1 => bm_i_finish,
		input2 => bm_i_finish,
		input3 => bm_i_finish,
		input4 => bm_i_finish,
		input5 => bm_i_finish,
		input6 => bm_i_finish,
		input7 => bm_i_finish,
		
		output0 =>  hw_complete,
		output1 =>  open,
		output2 =>  open,
		output3 =>  open,
		output4 =>  open,
		output5 =>  open,
		output6 =>  open,
		output7 =>  open
	);		
	
	
	-- Instantiate global clock buffer explicitely to:
	-- * Enable/disable clock
	-- * Make it possible to generate Directed Routing Constraint for clock signal.
	BUFGCE_inst : BUFGCE
	port map (
		O => clk_buffer_out,		-- Clock buffer output (connected to reconf. module)
		CE => clk_buffer_ce,		-- Clock enable input
		I => Bus2IP_Clk				-- Clock buffer input
	);
	
		
	MULT : entity multiplier
	generic map(
		OPERAND_SIZE => 32,
		LOG_OPERAND_SIZE => 5
	)
	port map(
		clk =>  clk_buffer_out,		
		reset => bm_o_reset,
		addr => bm_o_addr,
		zerofill => bm_o_zerofill,
		shift_i => bm_o_shift_i,
		shift_o => bm_i_shift_o,
		shift_g => bm_o_shift_g,
		run 	=> bm_o_run,
		finish  => bm_i_finish			
	);
		
	
end bram_id_cache;
