-------------------------------------------------------------------------------
--
-- Title       : multiplier
-- Design      : reconfiguration
-- Author      : Vegard
-- Company     : NTNU
-- Description : Parameterized sequential multiplicator
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.std_logic_unsigned.all;


entity multiplier is
generic(
		OPERAND_SIZE      	: natural	:=32;
		SHIFT_SIZE			: natural   :=8;
		LOG_OPERAND_SIZE	: natural   :=5
		);

	port(
		clk			: in 	std_logic;
		reset		: in    std_logic;
		run			: in 	std_logic; 
		addr		: in    std_logic;
		zerofill	: in    std_logic;
		shift_i 	: in 	std_logic_vector(SHIFT_SIZE-1 downto 0);
		shift_g		: in 	std_logic;
		shift_o 	: out 	std_logic_vector(SHIFT_SIZE-1 downto 0);
		finish		: out	std_logic 
	);

end multiplier;


architecture arch of multiplier is

--STATE
----------------------------------------------------------------------
signal cstate,nstate : std_logic_vector(1 downto 0);

constant IDLE : std_logic_vector(1 downto 0) :="00";
constant LOAD : std_logic_vector(1 downto 0)	:="01";
constant SHIFTING : std_logic_vector(1 downto 0) :="10";
constant COMPLETE : std_logic_vector(1 downto 0) :="11";
----------------------------------------------------------------------


	   
signal cnt  	 		 : std_logic_vector(LOG_OPERAND_SIZE-1 downto 0);
signal partial_sum		 : std_logic_vector(OPERAND_SIZE  downto 0);
signal partial_operand	 : std_logic_vector(2*OPERAND_SIZE-1  downto 0);


signal CNT_REG_count : std_logic ;
signal STATE_REG_write : std_logic;
signal REM_data_o : std_logic_vector(0 downto 0);


signal OPERAND_A_REG_shift,OPERAND_A_REG_shift_l: std_logic; 
signal OPERAND_A_REG_shift_o : std_logic_vector(SHIFT_SIZE-1 downto 0);
signal OPERAND_B_REG_par_o : std_logic_vector(OPERAND_SIZE-1 downto 0);	
signal OPERAND_B_REG_shift_o : std_logic_vector(SHIFT_SIZE-1 downto 0);
signal RESULT_REG_shift_l,RESULT_REG_write : std_logic; 


signal addend_b : std_logic_vector(OPERAND_SIZE-1 downto 0);

signal sel : std_logic_vector(1 downto 0);


signal outputs : std_logic_vector(2*SHIFT_SIZE-1 downto 0);

signal RESULT_REG_shift_i : std_logic_vector(SHIFT_SIZE-1 downto 0);

signal shift_chain0, shift_chain1 : std_logic;

signal shift_chain0_o, shift_chain1_o : std_logic_vector(SHIFT_SIZE-1 downto 0);

signal sig_zeros : std_logic_vector(OPERAND_SIZE-1 downto 0);

begin
	
	
	sig_zeros <= (others =>'0');
	
	shift_chain0 <= shift_g and sel(0);
	shift_chain1 <= shift_g and sel(1);
	outputs <= shift_chain1_o & shift_chain0_o;
	shift_chain1_o <=  cnt & cstate & REM_data_o;
	
	OPERAND_A_REG_shift <= OPERAND_A_REG_shift_l or shift_g;
	
	DECODER: entity decoder_custom 
	port map(
		addr => addr,
		output => sel
	);
	
	MUX : entity mux_custom
	port map(
		input => outputs,
		addr  => addr,
		output => shift_o	
	);
	
	OPERAND_A_REG : entity register_macro_so
	   	generic map(
			C_SIZE => SHIFT_SIZE, 
			C_NUM_REG => OPERAND_SIZE/SHIFT_SIZE
		)
		port map(
			clk		=> clk,
			reset   => reset,
			shift_i => shift_i,
			shift_p	=> shift_chain0,
			shift_s => OPERAND_A_REG_shift_l,
			shift_o	=> OPERAND_A_REG_shift_o
		);
		

	OPERAND_B_REG : entity register_macro
	generic map(C_SIZE => SHIFT_SIZE, C_NUM_REG => OPERAND_SIZE/SHIFT_SIZE)
	port map(
		clk 	=> clk,
		data_i	=> sig_zeros,
		shift_i => OPERAND_A_REG_shift_o,
		write 	=> sig_zeros(0),
		shift 	=> shift_chain0,
		data_o 	=> OPERAND_B_REG_par_o,
		shift_o => OPERAND_B_REG_shift_o
	); 		  
	
	CNT_REG : entity counter_reg
	generic map(C_SIZE => LOG_OPERAND_SIZE)
	port map(
		clk 	=> clk,
		reset	=> reset,
		write   => shift_chain1,
		count   => CNT_REG_count,
		data_i  => shift_i(SHIFT_SIZE-1 downto SHIFT_SIZE - LOG_OPERAND_SIZE),
		data_o 	=> cnt
	); 		  
	
	STATE_REG : entity register_macro
	generic map(
		C_SIZE =>2,
		C_NUM_REG =>1
	)
	port map(	   
		clk		=>clk,
		data_i	 => nstate, 
		shift_i  =>	shift_i(SHIFT_SIZE - LOG_OPERAND_SIZE-1 downto SHIFT_SIZE - LOG_OPERAND_SIZE-2),
		write	=> STATE_REG_write,
		shift	=> shift_chain1,
		data_o	=> cstate
	);	 
	
	REM_BUFFER : entity paralell_shift
	generic map(C_SIZE => 1)
	port map(	   
		clk		=>clk,
		write   => shift_chain1, 
		data_i  => shift_i(0 downto 0),
		data_o	=> REM_data_o
	);
	
	--PARTIAL PRODUCT / RESULT
	RESULT_REG : entity register_macroPIO
	generic map(					  
		C_SIZE => SHIFT_SIZE,
		C_INPUT_SIZE => OPERAND_SIZE+1,
		C_NUM_REG => 2*OPERAND_SIZE/SHIFT_SIZE
	)
	port map(
		clk 	=> clk,
		reset   => reset,
		data_i 	=> partial_sum,
    	data_o  => partial_operand,
		shift_p	=> shift_chain0,
		shift_s => RESULT_REG_shift_l,
		shift_i => RESULT_REG_shift_i,
		shift_o => shift_chain0_o,
		write => RESULT_REG_write
	);			 
	

	
	ADDER : entity cla_adder
	generic map(OPERAND_SIZE => OPERAND_SIZE)
							
	port map(
		a_i => partial_operand(2*(OPERAND_SIZE-1)downto OPERAND_SIZE-1),
		b_i => addend_b,
		s_o => partial_sum
	);
	
	NSTATE_PROC : process(cstate,run,cnt)
	begin	
		if run ='0' then
			nstate <= IDLE; 
			STATE_REG_write <='0';
		else
			STATE_REG_write <='1';
			case cstate is
				when IDLE =>
					nstate <= LOAD;
				when LOAD =>
					nstate <= SHIFTING;
				when SHIFTING =>
					if cnt = OPERAND_SIZE-1 then
						nstate <= COMPLETE;
					else 
						nstate <= LOAD;
					end if;	 
				when COMPLETE =>
					nstate <= IDLE;	
				when others =>
			end case;
		end if;
	end process NSTATE_PROC;
	
	
	CNT_PROC : process(run,cstate) is 
	begin	
		if (run ='1' and cstate = LOAD) or cstate = COMPLETE then
			CNT_REG_count <= '1';
		else
			CNT_REG_count <='0';
		end if;
	end process CNT_PROC;
	
	
	WRITE_PROC : process(cstate,run) is 
	begin
		if (cstate = LOAD or cstate = COMPLETE) and run ='1' then
			RESULT_REG_write <='1';
		else
			RESULT_REG_write <='0';
		end if;
	end process WRITE_PROC;
	
	SHIFT_PROC : process(cstate,run) is
	begin
		if cstate = SHIFTING and run ='1' then
			RESULT_REG_shift_l 	<='1';
			OPERAND_A_REG_shift_l <='1';
		else
			OPERAND_A_REG_shift_l <='0'; 
			RESULT_REG_shift_l 	<='0';
		end if;
	end process SHIFT_PROC;
	
	ADDEND_B_PROC : process(OPERAND_A_REG_shift_o,OPERAND_B_REG_par_o)is
	begin
		if OPERAND_A_REG_shift_o(0) ='1' then
			addend_b <= OPERAND_B_REG_par_o;
		else
			addend_b <= (others =>'0');
		end if;
			
		
	end process ADDEND_B_PROC;

	
	FINISH_PROC : process(clk,cstate)
	begin
		if clk'event and clk ='1' then
			if cstate = COMPLETE then
				finish <='1';
			else
				finish <='0';
			end if;	   
		end if;
		
	end process FINISH_PROC;
	
	SREG_I_PROC : process(zerofill,OPERAND_B_REG_shift_o) is
	begin	
		-- should case selec as well
		if zerofill ='1' then
			RESULT_REG_shift_i <= (others =>'0');
		else
			RESULT_REG_shift_i <= OPERAND_B_REG_shift_o;
		end if;
	end process SREG_I_PROC;
	
end arch;
